/*
 * _palm.cpp - General Palm layers.
 */

// vim: set noet:			// Hard tabs.
// vim: set ts=8:			// Tab width 8.
// vim: set sw=4:			// Shift (indent) width 4.
// vim: set cino={0s:			// Braces flush with outer statement.
// vim: set cino=^1s:			// First indent 8 (add 1 shift width).

#include "precomp.h"

#ifndef OS_PALM

#include "palmlayer.h"

/*
 * StrCompare and StrCaselessCompare on Windows Mobile.
 *
 * Sorting 8-bit text on Windows Mobile is a challenge.  Palm OS PDB files are
 * in codepage 1252, which is 8-bit text.  The strcmp family of functions does
 * not sort international characters correctly on Windows Mobile, and there is
 * no lstrcmpA or CompareStringA on Windows Mobile.
 *
 * We could convert everything to Unicode in memory and use native lstrcmpW
 * (etc) APIs, but that would either defer realization of lossy conversions
 * (e.g. when entering new text) or incur a nasty performance hit.
 *
 * Instead, I build a sort order table by first converting each 8-bit
 * character to Unicode and then sorting the converted characters.  The
 * resulting lookup table can be used to mimic the sort order that would be
 * used by the Unicode functions.
 *
 * I build the table dynamically at runtime so that the sorting behavior is
 * guaranteed to match the current configuration of the user's device.  This
 * also means that any sorting problems should be the fault of the OS/etc.
 *
 * The notable caveat is that this approach does not work for multibyte
 * encodings.  But it should be about the same brokenness as strcmp for
 * multibyte encodings.
 */

#ifdef OS_WINDOWSMOBILE

struct AmbiChar
{
	unsigned char ch;
	WCHAR wz[2];

	static int cmp( void const* pv1, void const* pv2 );
	static int cmpi( void const* pv1, void const* pv2 );
	static int cmpiSort( void const* pv1, void const* pv2 );
};

int
AmbiChar::cmp( void const* pv1, void const* pv2 )
{
	AmbiChar const* p1 = reinterpret_cast<AmbiChar const*>( pv1 );
	AmbiChar const* p2 = reinterpret_cast<AmbiChar const*>( pv2 );
	return lstrcmpW( p1->wz, p2->wz );
}

int
AmbiChar::cmpi( void const* pv1, void const* pv2 )
{
	AmbiChar const* p1 = reinterpret_cast<AmbiChar const*>( pv1 );
	AmbiChar const* p2 = reinterpret_cast<AmbiChar const*>( pv2 );
	return lstrcmpiW( p1->wz, p2->wz );
}

int
AmbiChar::cmpiSort( void const* pv1, void const* pv2 )
{
	AmbiChar const* p1 = reinterpret_cast<AmbiChar const*>( pv1 );
	AmbiChar const* p2 = reinterpret_cast<AmbiChar const*>( pv2 );
	int const n = lstrcmpiW( p1->wz, p2->wz );
	if( n )
	    return n;
	return lstrcmpW( p1->wz, p2->wz );
}

class CharsInOrder
{
	enum { ORDER_STRICT, ORDER_CASELESS };
    public:
				CharsInOrder();
	unsigned char const*	GetOrder() const { return m_rgCharsInOrder[ORDER_STRICT]; }
	unsigned char const*	GetCaselessOrder() const { return m_rgCharsInOrder[ORDER_CASELESS]; }
    private:
	unsigned char		m_rgCharsInOrder[2][256];
};

CharsInOrder::CharsInOrder()
{
	AmbiChar rg[256];

	for( unsigned iStyle = 2; iStyle--; )
	{
	    unsigned char* const prgCharsInOrder = m_rgCharsInOrder[iStyle];

	    for( unsigned ii = 0; ii < DimensionOf( rg ); ii++ )
	    {
		rg[ii].ch = ii;
		if( ii > 0 && ii < 0xfe )
		{
		    MultiByteToWideChar( CP_ACP, 0, reinterpret_cast<char const*>( &rg[ii].ch ), 1, RgC( rg[ii].wz ) );
		    rg[ii].wz[1] = 0;
		}
	    }

	    if( iStyle == ORDER_STRICT )
		qsort( rg + 1, DimensionOf( rg ) - 3, sizeof( rg[0] ), AmbiChar::cmp );
	    else
		qsort( rg + 1, DimensionOf( rg ) - 3, sizeof( rg[0] ), AmbiChar::cmpiSort );

	    unsigned iEquivalent = 0;
	    for( unsigned ii = 0; ii < DimensionOf( rg ); ii++ )
	    {
		if( iStyle == ORDER_STRICT )
		    iEquivalent = ii;
		else if( AmbiChar::cmpi( &rg[iEquivalent], &rg[ii] ) )
		    iEquivalent = ii;

		prgCharsInOrder[rg[ii].ch] = iEquivalent;
	    }
	}
}

static CharsInOrder s_CharsInOrder;

int
StrCompare( Char const* s1, Char const* s2 )
{
	unsigned char const* const prgOrder = s_CharsInOrder.GetOrder();
	while( true )
	{
	    int const n = ( int( prgOrder[static_cast<unsigned char>( *s1 )] ) -
			    int( prgOrder[static_cast<unsigned char>( *s2 )] ) );
	    if( n )
		return n;
	    if( !*s1 )
		return 0;
	    s1++;
	    s2++;
	}
}

int
StrCaselessCompare( Char const* s1, Char const* s2 )
{
	unsigned char const* const prgOrder = s_CharsInOrder.GetCaselessOrder();
	while( true )
	{
	    int const n = ( int( prgOrder[static_cast<unsigned char>( *s1 )] ) -
			    int( prgOrder[static_cast<unsigned char>( *s2 )] ) );
	    if( n )
		return n;
	    if( !*s1 )
		return 0;
	    s1++;
	    s2++;
	}
}

int
StrCaselessCompareCount( Char const* s1, Char const* s2, size_t cch )
{
	unsigned char const* const prgOrder = s_CharsInOrder.GetCaselessOrder();
	while( cch-- )
	{
	    int const n = ( int( prgOrder[static_cast<unsigned char>( *s1 )] ) -
			    int( prgOrder[static_cast<unsigned char>( *s2 )] ) );
	    if( n )
		return n;
	    if( !*s1 )
		return 0;
	    s1++;
	    s2++;
	}
	return 0;
}

int
StrCaselessCompareSort( Char const* s1, Char const* s2 )
{
	unsigned char const* const prgOrder = s_CharsInOrder.GetCaselessOrder();
	unsigned char const* const prgStrict = s_CharsInOrder.GetOrder();
	int nStrict = 0;
	while( true )
	{
	    int const n = ( int( prgOrder[static_cast<unsigned char>( *s1 )] ) -
			    int( prgOrder[static_cast<unsigned char>( *s2 )] ) );
	    if( n )
		return n;
	    if( !*s1 )
		return nStrict;
	    if( !nStrict )
		nStrict = ( int( prgStrict[static_cast<unsigned char>( *s1 )] ) -
			    int( prgStrict[static_cast<unsigned char>( *s2 )] ) );
	    s1++;
	    s2++;
	}
}

#endif

/*
 * SysUtils.
 */

static Int32 s_other;
static CmpFuncPtr s_comparF;

static int
cmpQsort( void const* pv1, void const* pv2 )
{
	return s_comparF( ( void* )pv1, ( void* )pv2, s_other );
}

void
SysInsertionSort( void* baseP, Int16 numOfElements, Int16 width,
		  CmpFuncPtr const comparF, Int32 const other )
{
	s_comparF = comparF;
	s_other = other;

	// Who said it had to really do an insertion sort?  What ultimately
	// matters in an emulation layer is that it sorts.

	qsort( baseP, numOfElements, width, cmpQsort );
}

#endif // !OS_PALM

