/*
 * BitTable.cpp - Fast reference table for filtered records.
 */

// vim: set noet:			// Hard tabs.
// vim: set ts=8:			// Tab width 8.
// vim: set sw=4:			// Shift (indent) width 4.
// vim: set cino={0s:			// Braces flush with outer statement.
// vim: set cino=^1s:			// First indent 8 (add 1 shift width).

#include "precomp.h"

#include "dbopen.h"
#include "bittable.h"
#include "progress.h"

BitTable::BitTable( DbOpen* pdb )
{
	m_pdb = pdb;
	m_pbits = 0;
	m_cAlloc = 0;

	Clear();
}

void
BitTable::Clear()
{
	Assert( this );

	m_fInitialized = false;

	if( m_pbits && m_cRecords )
	{
	    Assert( FImplies( m_cRecords, m_cAlloc ) );
	    ZeroMemory( m_pbits, ( RecordIndex( m_cAlloc - 1 ) + 1 ) * sizeof( *m_pbits ) );
	}

	m_cRecords = 0;
	m_cInView = 0;
}

void
BitTable::Free()
{
	MemFree( m_pbits );
	m_pbits = 0;

	Clear();
}

bool
BitTable::Alloc( unsigned cRecords )
{
	const unsigned c_cGrowPowerOfTwo = 9;
	const unsigned c_cGrow = DWORD( 1 ) << c_cGrowPowerOfTwo;

	Assert( !( m_cAlloc % c_cGrow ) );

	if( cRecords > m_cAlloc )
	{
	    unsigned cAlloc = ( cRecords + c_cGrow - 1 ) & ~( c_cGrow - 1 );
	    unsigned cbAlloc = ( RecordIndex( cAlloc - 1 ) + 1 ) * sizeof( *m_pbits );
	    DWORD* p;

	    p = ( DWORD* )( MemReAlloc( m_pbits, cbAlloc _MEM_NOSIZECHECK ) );
	    if( !p )
		return false;

	    m_pbits = p;

	    // Zero out the newly allocated section.

	    ZeroMemory( m_pbits + RecordIndex( m_cAlloc ),
			( RecordIndex( cAlloc ) - RecordIndex( m_cAlloc ) ) * sizeof( *m_pbits ) );

	    m_cAlloc = cAlloc;
	}

	m_cRecords = cRecords;

	return true;
}

bool
BitTable::InView( unsigned record )
{
	Assert( m_pdb );

	EnsureScanned();

	if( record >= m_cRecords )
	    return false;

	return !!( m_pbits[RecordIndex( record )] & RecordMask( record ) );
}

unsigned
BitTable::IndexInView( unsigned record )
{
	Assert( m_pdb );

	EnsureScanned();

	unsigned c = 0;

	Assert3( InView( record ), "record %u is not in view!\nRecordIndex=%u\nRecordMask=%lx",
		 record, RecordIndex( record ), RecordMask( record ) );

	while( FindNextRecord( &record, false ) )
	    c++;

	return c;
}

unsigned
BitTable::CountInView()
{
	Assert( m_pdb );

	EnsureScanned();

	return m_cInView;
}

bool
BitTable::FindNextRecord( unsigned* precord, bool fNext )
{
	Assert( m_pdb );

	EnsureScanned();

	if( !m_cRecords || *precord >= m_cRecords )
	    return false;

	DWORD mask = RecordMask( *precord );
	const DWORD* p = m_pbits + RecordIndex( *precord );
	register unsigned record = *precord;

	if( fNext )
	{
	    for( ;; )
	    {
		record++;

		if( record >= m_cRecords )
		    return false;

		mask <<= 1;

		if( !mask )
		{
		    p++;
		    mask = 1;

		    // Extra optimization for skipping thru empty regions
		    // quickly.

		    while( !*p )
		    {
			record += 32;
			if( record >= m_cRecords )
			    return false;
			p++;
		    }
		}

		if( ( *p ) & mask )
		{
		    *precord = record;
		    return true;
		}
	    }
	}
	else
	{
	    while( record )
	    {
		record--;
		mask >>= 1;

		if( !mask )
		{
		    p--;
		    mask = DWORD( 1 ) << 31;

		    // Extra optimization for skipping thru empty regions
		    // quickly.

		    while( !*p )
		    {
			if( record < 32 )
			    return false;

			Assert( record % 32 == 32 - 1 );
			Assert( p > m_pbits );

			record -= 32;
			p--;
		    }
		}

		if( ( *p ) & mask )
		{
		    *precord = record;
		    return true;
		}
	    }
	}

	return false;
}

bool
BitTable::EnsureScanned()
{
	Assert( m_pdb );

	if( m_fInitialized )
	    return true;

	return FullScan();
}

bool
BitTable::FullScan()
{
	ScopedProgressHolder sph;

	Clear();

	if( !m_pdb /*|| !m_pdb->IsOpen()*/ )
	{
	    m_fInitialized = true;
	    return true;
	}

	if( !Alloc( m_pdb->Count() ) )
	    return false;

	unsigned record;
	DWORD mask;
	DWORD* p;

	sph.Create();
	sph.Init( m_cRecords, PROGRESS_NOWAIT|PROGRESS_PREDICTIVEWAIT|PROGRESS_NODISPATCH, CAPTION_SCANNING );

	mask = 1;
	p = m_pbits;

	for( record = 0; record < m_cRecords; )
	{
	    if( !m_pdb->IsRecordFiltered( record ) )
	    {
		m_cInView++;
		*p |= mask;
	    }

	    mask <<= 1;
	    if( !mask )
	    {
		mask = 1;
		p++;
	    }

	    record++;

	    sph.Update( record );
	}

	m_fInitialized = true;

	DbgMsg(( 5, TEXT( "FullScan:\n%d records\n%d in view" ), m_cRecords, m_cInView ));

	return true;
}

bool
BitTable::RefineScan()
{
	if( !m_fInitialized )
	    return FullScan();

	if( !m_pdb )
	    return true;

	unsigned record;
	unsigned cProgress = 0;
	DWORD mask;
	DWORD* p;

	ScopedProgressHolder sph;
	sph.Create();
	sph.Init( CountInView(), PROGRESS_NOWAIT|PROGRESS_PREDICTIVEWAIT, CAPTION_SCANNING );

	mask = 1;
	p = m_pbits;

	for( record = 0; record < m_cRecords; )
	{
	    if( *p & mask )
	    {
		cProgress++;

		if( m_pdb->IsRecordFiltered( record ) )
		{
		    m_cInView--;
		    *p &= ~mask;
		}
	    }

	    mask <<= 1;
	    if( !mask )
	    {
		mask = 1;
		p++;
	    }

	    record++;

	    sph.Update( cProgress );
	}

	DbgMsg(( 5, TEXT( "RefineScan:\n%d records\n%d in view" ), m_cRecords, m_cInView ));

	return true;
}

bool
BitTable::NewRecord( unsigned record, bool fInView )
{
	if( m_pdb && m_fInitialized )
	{
	    if( !Alloc( m_cRecords + 1 ) )
		return false;

	    MoveRecord( m_cRecords - 1, record );
	    MarkRecord( record, fInView );
	}

	return true;
}

void
BitTable::RemoveRecord( unsigned record )
{
	if( m_pdb && m_fInitialized )
	{
	    MoveRecord( record, m_cRecords - 1 );
	    MarkRecord( m_cRecords - 1, false );

	    // Afterwards, in case anything uses it during.

	    m_cRecords--;
	}
}

void
BitTable::MoveRecord( unsigned recordFrom, unsigned recordTo )
{
	if( !m_pdb || !m_fInitialized )
	    return;

	DWORD* p;
	DWORD mask;
	DWORD tmp;
	DWORD carry;
	bool fThis = InView( recordFrom );

	if( recordFrom < recordTo )
	{
	    // Shifting backward...

	    // Check for simple shift.

	    if( RecordIndex( recordFrom ) == RecordIndex( recordTo ) )
	    {
		/*
		 *	from	= 3	[00000000000000000000000000001000]
		 *	to	= 8	[00000000000000000000000100000000]
		 *
		 *	want to shift	[00000000000000000000000111110000] >> 1
		 *	want to zero	[00000000000000000000000111111000]
		 *
		 */

		p = m_pbits + RecordIndex( recordFrom );
		mask = ~( ( RecordMask( recordFrom ) << 1 ) - 1 ); // get middle+upper mask
		mask &= ( RecordMask( recordTo ) << 1 ) - 1; // get intersection with middle+lower mask

		tmp = ( *p ) & mask;			// bits to shift
		( *p ) &= ~( mask | ( mask >> 1 ) );	// clear bits
		( *p ) |= tmp >> 1;			// shift bits

		if( fThis )
		    ( *p ) |= RecordMask( recordTo );	// 'move' the record

		return;
	    }

	    // Shift partial From set.

	    p = m_pbits + RecordIndex( recordFrom );
	    mask = RecordMask( recordFrom );
	    tmp = ( *p ) & ~( ( mask << 1 ) - 1 );	// extract upper bits
	    *p &= ( mask - 1 );				// zero upper bits
	    *p |= tmp >> 1;				// shift the upper bits (highest is set later)

	    // Shift partial To set.

	    p = m_pbits + RecordIndex( recordTo );
	    mask = RecordMask( recordTo ) << 1;
	    tmp = ( *p ) & ( mask - 1 );		// extract lower bits
	    *p &= ~( mask - 1 );			// zero lower bits
	    *p |= tmp >> 1;				// shift lower bits
	    carry = tmp << 31;				// carry lowest bit

	    if( fThis )
		( *p ) |= RecordMask( recordTo );

	    // Shift backward with carry.

	    const DWORD* pBegin = m_pbits + RecordIndex( recordFrom );

	    while( --p > pBegin )
	    {
		tmp = ( *p ) << 31;			// remember low bit for carry
		( *p ) >>= 1;				// shift downward
		( *p ) |= carry;			// carry from previous
		carry = tmp;				// update carry for next time
	    }

	    // Last carry.

	    Assert( p == pBegin );
	    *p |= carry;				// set the original leftover bit
	}
	else if( recordFrom > recordTo )
	{
	    // Shifting forward...

	    // Check for simple shift.

	    if( RecordIndex( recordFrom ) == RecordIndex( recordTo ) )
	    {
		/*
		 *	from	= 9	[00000000000000000000001000000000]
		 *	to	= 5	[00000000000000000000000000100000]
		 *
		 *	want to shift	[00000000000000000000000111100000] << 1
		 *	want to zero	[00000000000000000000001111100000]
		 *
		 */

		p = m_pbits + RecordIndex( recordFrom );
		mask = ~( RecordMask( recordTo ) - 1 );	// get middle+upper mask
		mask &= RecordMask( recordFrom ) - 1;	// get intersection with middle+lower mask

		tmp = ( *p ) & mask;			// bits to shift
		( *p ) &= ~( mask | ( mask << 1 ) );	// clear bits
		( *p ) |= tmp << 1;			// shift bits

		if( fThis )
		    ( *p ) |= RecordMask( recordTo );	// 'move' the record

		return;
	    }

	    // Shift partial From set.

	    p = m_pbits + RecordIndex( recordFrom );
	    mask = RecordMask( recordFrom );
	    tmp = ( *p ) & ( mask - 1 );		// extract lower bits
	    *p &= ~( ( mask << 1 ) - 1 );		// zero lower bits
	    *p |= tmp << 1;				// shift lower bits (lowest is set later)

	    // Shift partial To set.

	    p = m_pbits + RecordIndex( recordTo );
	    mask = RecordMask( recordTo );
	    tmp = ( *p ) & ~( mask - 1 );		// extract upper bits
	    *p &= ( mask - 1 );				// zero upper bits
	    *p |= tmp << 1;				// shift upper bits
	    carry = tmp >> 31;				// carry highest bit

	    if( fThis )
		( *p ) |= mask;				// 'move' the record

	    // Shift forward with carry.

	    const DWORD* pEnd = m_pbits + RecordIndex( recordFrom );

	    while( ++p < pEnd )
	    {
		tmp = ( *p ) >> 31;			// remember high bit for carry
		( *p ) <<= 1;				// shift upward
		( *p ) |= carry;			// carry from previous
		carry = tmp;				// update carry for next time
	    }

	    // Last carry.

	    Assert( p == pEnd );
	    *p |= carry;				// set the original leftover bit
	}
}

void
BitTable::MarkRecord( unsigned record, bool fInView )
{
	if( m_pdb && m_fInitialized )
	{
	    Assert( record < m_cRecords );

	    DWORD* p = &m_pbits[RecordIndex( record )];
	    DWORD mask = RecordMask( record );

	    if( ( ( *p ) & mask ) && !fInView )
		m_cInView--;
	    else if( !( ( *p ) & mask ) && fInView )
		m_cInView++;

	    if( fInView )
		*p |= mask;
	    else
		*p &= ~mask;
	}
}

void
BitTable::RefreshRecord( unsigned record )
{
	// Only refreshing when the BitTable is already initialized can allow
	// certain iterating actions to avoid an unnecessary performance hit
	// during PackRecord.

	if( m_pdb && m_fInitialized )
	{
	    Assert( record < m_cRecords );

	    MarkRecord( record, !m_pdb->IsRecordFiltered( record ) );
	}
}

