/*
 * Converter.cpp - Conversion tool for HandyShopper database files.
 */

// vim: set noet:			// Hard tabs.
// vim: set ts=8:			// Tab width 8.
// vim: set sw=4:			// Shift (indent) width 4.
// vim: set cino={0s:			// Braces flush with outer statement.
// vim: set cino=^1s:			// First indent 8 (add 1 shift width).

#include "precomp.h"

#ifdef OS_PALM
#error The converter is not designed for the Palm OS.
#endif

#include <stdio.h>

#include "palmlayer.h"

#include "converter.h"
#include "arrays.h"
#include "str.h"
#include "storelist.h"
#include "item.h"
#include "dbopen.h"
#include "bldnum.h"

/*
 * Usage info.
 */

static char const* const c_pszUsageShort = "Use 'hs3convert -?' for help.";
static char const* const c_pszUsageLong =
"\n"
"    hs3convert -- Convert HandyShopper databases to/from other formats.\n"
"\n"
"    hs3convert [ options ] fromFile [ toFile ]\n"
"\n"
"	Hs3convert converts HandyShopper 2 and HandyShopper 3 databases\n"
"	to/from CVS format.  Hs3convert requires at least HandyShopper 2.1\n"
"	(build 3409).  To ensure compatibility, always upgrade Hs3convert\n"
"	and HandyShopper (on the handheld) together.\n"
"\n"
"    Options:\n"
"\n"
"	-?		Show this help message.\n"
"	-? examples	Show examples.\n"
"	-c		Output in CSV format (as two files, .txt and .csv).\n"
"	-d		Dump raw PDB data to stdout.\n"
"	-f		Force overwriting existing file, if any.\n"
#if 0 // Suppress the information about -h until it is officially supported.
"	-h		Output in HSDB format.\n"
#endif
"	-i		Use store Ids instead of store Names (only applies\n"
"			to CSV input and output.\n"
"	-n		Suppress logo information.\n"
"	-p		Output in PDB format.\n"
"	-t		Output freeform Text (default to stdout).\n"
"	-x		Output in XML format (default to stdout).\n"
"	-C		Use comma (,) for the decimal separator character.\n"
"	-X		When creating a CSV file, this prefixes each text\n"
"			string column with a single quote, to help various\n"
"			spreadsheet programs recognize them as text strings\n"
"			instead of formulas.\n"
"	-V		Show version information for Hs3convert.\n"
"\n"
"    Help:\n"
"\n"
"	For help with this tool or HandyShopper in general, please visit the\n"
"	HandyShopper User Group at http://groups.yahoo.com/group/handyshopper\n"
;
static char const* const c_pszUsageExamples =
"\n"
"    Examples:\n"
"\n"
"	hs3convert -c HS2_MyDatabase.pdb mydb\n"
"	    Converts the HS2_MyDatabase.pdb file into CSV format,\n"
"	    generating mydb.csv (items) and mydb.txt (non-item data).\n"
"\n"
"	hs3convert -p mydb.csv HS2_MyDatabase\n"
"	    Converts the mydb.csv (and mydb.txt) files into PDB format,\n"
"	    generating HS2_MyDatabase.pdb.  Afterwards you can install the\n"
"	    database onto your handheld device (consult the user manual for\n"
"	    your device for details).\n"
#ifdef INCLUDE_HSDB_SUPPORT
"\n"
"	hs3convert -c MyPocketDatabase.hsdb mydb\n"
"	    Converts the MyPocketDatabase.hsdb file into CSV format,\n"
"	    generating mydb.csv (items) and mydb.txt (non-item data).\n"
"\n"
"	hs3convert -h mydb.csv MyPocketDatabase\n"
"	    Converts the mydb.csv (and mydb.txt) files into HSDB format,\n"
"	    generating MyPocketDatabase.hsdb.  Afterwards you can install the\n"
"	    database onto your handheld device (consult the user manual for\n"
"	    details).\n"
"\n"
"	hs3convert MyDatabase.hsdb MyDatabase.pdb\n"
"	    Converts the MyDatabase.hsdb file (from HSDB format) into\n"
"	    MyDatabase.pdb (to PDB format).\n"
"\n"
"	hs3convert MyDatabase.pdb MyDatabase.hsdb\n"
"	    Converts the MyDatabase.pdb file (from PDB format) into\n"
"	    MyDatabase.hsdb (to HSDB format).\n"
#endif
;
static char const* const c_pszLogo =
"HS3CONVERT -- HandyShopper database conversion tool.\n"
"No warranties are granted; use this tool at your own risk.\n"
"\n"
;
static char const* const c_pszVersion =
"HS3CONVERT -- build " STR_BLDNUMFULL " -- " STR_BLDDATE ".\n"
"Copyright (c) 1998-2010 by Christopher Antos.\n"
;

/*
 * General helpers.
 */

void
Abort( TCHAR const* const psz )
{
	fprintf( stderr, "%S\n", psz );
	exit( 1 );
}

/*
 * Glue.
 */

void
CreateProgressObject( IProgress** ppProgress )
{
	*ppProgress = 0;
}

/*
 * Options parsing.
 */

enum { c_cMaxOptions = 10 };

enum OptFlag
{
	// Bitwise selectors.

	OPT_ONE		= 0x01,		// Exactly one.
	OPT_TWO		= 0x02,		// Exactly two.
	OPT_THREE	= 0x04,		// Exactly three.
	OPT_MORE	= 0x10,		// More than three.
	OPT_NONE	= 0x20,		// Require none.

	// Combos of the above.

	OPT_OPT		= OPT_NONE|OPT_ONE,
	OPT_ANY		= OPT_NONE|OPT_ONE|OPT_TWO|OPT_THREE|OPT_MORE,
	OPT_SOME	= OPT_ONE|OPT_TWO|OPT_THREE|OPT_MORE,
};

class Options
{
    public:
			Options() : m_cOpts( 0 ), m_pszError( 0 ) {}
			~Options() { MemFree( m_pszError ); }

	bool		Parse( int& argc, char const**& argv, char const* pszOpts,
			       int flag, char const* pszUsage );
	char const*	GetErrorString() const { Assert( m_pszError ); return m_pszError; }

	char const*	GetValue( char chOpt, int iSubOpt ) const;
	char const*	operator[]( char chOpt ) const { return GetValue( chOpt, 0 ); }

    protected:
	void		ClearError() { MemFree( m_pszError ); m_pszError = 0; }
	void		SetError( char const* pszFormat, char const* pszUsage, ... );

    private:
	int		m_cOpts;
	char		m_rgchFlags[c_cMaxOptions];
	char const*	m_rgpszOpts[c_cMaxOptions];

	char*		m_pszError;
};

static char const*
GetArg( char const* psz, int& argc, char const**& argv )
{
	if( *psz )
	    return psz;

	if( !argc )
	    return 0;

	argc--;
	argv++;
	return argv[0];
}

bool
Options::Parse( int& argc, char const**& argv, char const* pszOpts,
		int const flag, char const* const pszUsage )
{
	Assert( pszOpts );
	Assert( pszUsage );

	ClearError();

	// If the options string begins with a slash, then recognize both '/'
	// and '-' and flag characters.

	bool const fSlash = ( *pszOpts == '/' );

	if( fSlash )
	    pszOpts++;

	// Parse flags.

	while( argc )
	{
	    // If this is not a flag, then we're done parsing.

	    if( argv[0][0] != '-' && ( !fSlash || argv[0][0] != '/' ) )
		break;

	    // '--' is special and means that subsequent arguments should not
	    // be treated as flags even if they being with '-'.

	    if( argv[0][1] == '-' )
	    {
		argc--;
		argv++;
		break;
	    }

	    char const* pszArg = argv[0];

	    while( true )
	    {
		pszArg++;		// Skip the '-' or option character.
		if( !*pszArg )
		    break;

#ifdef DEBUG
		if( *pszArg == '!' )
		{
		    __DbgBreak();
		    continue;
		}
#endif

		char const* psz = pszOpts;
		while( *psz && *psz != *pszArg )
		    psz++;

		if( !*psz )
		{
		    SetError( "Invalid option: '%c'.\n", pszUsage, *pszArg );
		    return false;
		}

		if( m_cOpts >= c_cMaxOptions )
		{
		    SetError( "Too many options.\n", pszUsage );
		    return false;
		}

		m_rgchFlags[m_cOpts] = *pszArg;
		m_rgpszOpts[m_cOpts] = "true";

		if( psz[1] == '.' )
		{
		    m_rgpszOpts[m_cOpts++] = pszArg + 1;
		    break;
		}

		if( psz[1] == ':' )
		{
		    pszArg++;
		    while( isspace( *pszArg ) )
			pszArg++;
		    if( !*pszArg )
			pszArg = GetArg( pszArg, argc, argv );
		    if( !pszArg )
		    {
			SetError( "Option '%c' missing required argument.\n", pszUsage, *psz );
			return false;
		    }
		    m_rgpszOpts[m_cOpts++] = pszArg;
		    break;
		}

		m_cOpts++;
	    }

	    argc--;
	    argv++;
	}

	// Check number of arguments.

	if( !( ( argc == 0 && ( flag & OPT_NONE ) ) ||
	       ( argc == 1 && ( flag & OPT_ONE ) ) ||
	       ( argc == 2 && ( flag & OPT_TWO ) ) ||
	       ( argc == 3 && ( flag & OPT_THREE ) ) ||
	       ( argc > 3 && ( flag & OPT_MORE ) ) ) )
	{
	    SetError( "Missing/wrong number of arguments.\n", pszUsage );
	    return false;
	}

	return true;
}

void
Options::SetError( char const* const pszFormat, char const* const pszUsage, ... )
{
	va_list args;
	va_start( args, pszUsage );

	ClearError();

	size_t const cchAlloc = 1024;
	m_pszError = static_cast<char*>( MemAlloc( cchAlloc * sizeof( *m_pszError ) ) );
	char* psz = m_pszError;
	char const* const pszEnd = m_pszError + cchAlloc;
	psz += sprintf_s( psz, pszEnd - psz, "Usage: %s\n", pszUsage );
	psz += vsprintf_s( psz, pszEnd - psz, pszFormat, args );

	va_end( args );
}

char const*
Options::GetValue( char const chOpt, int iSubOpt ) const
{
	for( int ii = m_cOpts; ii--; )
	    if( chOpt == m_rgchFlags[ii] )
	    {
		if( iSubOpt )
		{
		    iSubOpt--;
		    continue;
		}

		return m_rgpszOpts[ii];
	    }

	return 0;
}

/*
 * Dump raw PDB file.
 */

void
DumpBytes( FILE* const pout, char const* const pszTag, void const* const pv, unsigned cb )
{
	int const c_cBytesPerLine = 12;
	BYTE const* p = reinterpret_cast<BYTE const*>( pv );

	fprintf( pout, "%-13.13s = ", pszTag );

	while( cb )
	{
	    char sz[c_cBytesPerLine + 8];
	    unsigned jj;

	    for( jj = 0; jj < c_cBytesPerLine; jj++ )
	    {
		if( jj < cb )
		{
		    fprintf( pout, "%02X ", p[jj] );
		    sz[jj] = isprint( p[jj] ) ? p[jj] : '.';
		}
		else
		{
		    fprintf( pout, "   " );
		    sz[jj] = ( jj == cb ) ? '*' : ' ';
		}
	    }
	    sz[jj] = ( jj <= cb ) ? '*' : ' ';
	    sz[jj + 1] = 0;

	    fprintf( pout, " *%s", sz );

	    if( jj < cb )
	    {
		fprintf( pout, "\n                " );
		cb -= jj;
		p += jj;
	    }
	    else
		cb = 0;
	}

	fprintf( pout, "\n" );
}

static int
DoDumpRawPDB( TCHAR const* const pszFile )
{
	SPNew<Database> spdb = new Database;

	if( spdb->Read( pszFile ) < 0 )
	{
	    fprintf( stderr, "Unable to read input file." );
	    return 1;
	}

	// Dump header.

	UInt32 const type = MacDWord( spdb->GetHeader()->GetType() );
	UInt32 const creator = MacDWord( spdb->GetHeader()->GetCreator() );

	fprintf( stdout, "DATABASE NAME '%s'\n", spdb->GetDbName() );
	fprintf( stdout, "attr          = 0x%04x\n", spdb->GetHeader()->GetAttr() );
	fprintf( stdout, "ver           = %u\n", spdb->GetHeader()->GetVersion() );
	fprintf( stdout, "records       = %u\n", spdb->GetHeader()->NumRecords() );
	fprintf( stdout, "appinfo       = %u\n", spdb->GetHeader()->GetOfsAppInfo() );
	fprintf( stdout, "sortinfo      = %u\n", spdb->GetHeader()->GetOfsSortInfo() );
	fprintf( stdout, "type          = %.4s\n", &type );
	fprintf( stdout, "creator       = %.4s\n", &creator );

	// Dump AppInfo block.

	MemPtr const pvAppInfo = MemLocalIDToLockedPtr( spdb->GetAppInfo(), 0 );

	if( pvAppInfo )
	{
	    DumpBytes( stdout, "appinfo", pvAppInfo, MemPtrSize( pvAppInfo ) );

	    MemPtrUnlock( pvAppInfo );
	}

	// Dump the records.

	for( unsigned record = 0; record < spdb->DmNumRecords(); record++ )
	{
	    RecordEntry const* const p = spdb->GetEntry( record );
	    UInt32 const cb = p->GetSize();
	    MemPtr const pv = MemHandleLock( p->m_hData );

	    if( pv )
	    {
		fprintf( stdout, "\nRECORD %u (size: %u, attr: 0x%04x, uid: 0x%06x)\n", record, cb, p->GetAttr(), p->GetUid() );
		fprintf( stdout, "offset        = %u\n", p->GetOfsData() );

		DumpBytes( stdout, "data", pv, cb );

		MemPtrUnlock(pv);
	    }
	}

	spdb->Close();
	return 0;
}

/*
 * Convert database file.
 */

enum ConversionFormatType
{
	CFT_UNKNOWN,
	CFT_PALM,
	CFT_MOBILE,
	CFT_CSV,
};

static ConversionFormatType
ConversionFormatTypeFromExtension( TCHAR const* const pszFile )
{
	DbFormatType const dbft = DbFormatTypeFromExtension( pszFile );

	if( dbft == DBFT_PALM )
	    return CFT_PALM;
	if( dbft == DBFT_MOBILE )
	    return CFT_MOBILE;

	Assert( dbft == DBFT_UNKNOWN );

	TCHAR const* const pszExt = GetExtension( pszFile );

	if( pszExt )
	{
	    if( !_tcsicmp( pszExt, TEXT( ".csv" ) ) )
		return CFT_CSV;
	}

	return CFT_UNKNOWN;

}

static int
DoDbConvert( TCHAR const* const pszFrom, TCHAR const* const pszTo, bool const fForceOverwrite )
{
	Assert( DbFormatTypeFromExtension( pszFrom ) != DBFT_UNKNOWN );
	Assert( DbFormatTypeFromExtension( pszTo ) != DBFT_UNKNOWN );

	// Open the 'from' database.

	SPNew<DbOpen> spdbFrom;
	SysError e = OpenDatabase( pszFrom, &spdbFrom );

	if( e )
	{
	    fprintf( stderr, "Unable to open file '%S'.\n", pszFrom );
	    return 1;
	}

	// If the -f flag was used (force overwrite), then delete the database
	// before trying to create it.

	if( fForceOverwrite )
	    __DeleteFile( pszTo );

	// Create a new database.

	StrT sNewName;
	SPNew<DbOpen> spdbTo;

	e = NewDatabase( pszTo, spdbFrom->Name(), &spdbTo );

	if( e )
	{
	    fprintf( stderr, "Unable to create new file '%S'.\n", pszTo );
	    return 1;
	}

	// Convert (copy) the database.

	e = spdbTo->ConvertFrom( spdbFrom );

	if( e )
	{
	    fprintf( stderr, "Unable to convert the database.\n" );
	    return 1;
	}

	if( spdbTo->CanFlush() )
	{
	    e = spdbTo->Flush();

	    if( e )
	    {
		fprintf( stderr, "Unable to flush the converted database.\n" );
		return 1;
	    }
	}

	return 0;
}

static int
DoOutputDatabase( DbOpen const* const pdb, OutputInterface* const pout )
{
	Assert( pdb );
	Assert( pout );

	// Begin.

	int n = 0;

	n = pout->Begin( pdb );

	if( n < 0 )
	    return n;

	// Categories.

	n = pout->OutputCategory( pdb, pdb->GetCategoryNameById( catUnfiledCategory ), catUnfiledCategory );

	if( n < 0 )
	    return n;

	for( unsigned ii = 0; ii < pdb->GetCategoriesCount(); ii++ )
	{
	    WORD const category = pdb->GetCategoryIdByIndex( ii );

	    n = pout->OutputCategory( pdb, pdb->GetCategoryNameById( category ), category );

	    if( n < 0 )
		return n;
	}

	n = pout->OutputCategory( pdb, 0, 0 );

	if( n < 0 )
	    return n;

	// Preferences.

	n = pout->OutputPreferences( pdb );

	if( n < 0 )
	    return n;

	// Stores.

	for( unsigned ii = 0; ii < pdb->GetStoresCount(); ii++ )
	{
	    Store const store = pdb->GetStoreIdByIndex( ii );

	    n = pout->OutputStore( pdb, pdb->GetStoreNameById( store ), store );

	    if( n < 0 )
		return n;
	}

	n = pout->OutputStore( pdb, 0, 0 );

	if( n < 0 )
	    return n;

	// Items.

	for( unsigned ii = 0; ii < pdb->Count(); ii++ )
	{
	    SPNew<Item> spitem = pdb->UnPackRecord( ii );

	    if( !spitem )
		continue;

	    n = pout->OutputItem( pdb, spitem );

	    if( n < 0 )
		return n;
	}

	n = pout->Finished( pdb );

	if( n < 0 )
	    return n;

	return 0;
}

/*
 * Main.
 */

int __cdecl
main( int argc, char const** argv )
{
	// Parse arguments.

	if( argc )
	{
	    argc--;
	    argv++;
	}

	if( !argc )
	{
	    printf( "%s", c_pszUsageLong );
	    return 0;
	}

	Options opts;

	if( !opts.Parse( argc, argv, "/?cdfhinptxCXV", OPT_ANY, c_pszUsageShort ) )
	{
	    fprintf( stderr, "%s", opts.GetErrorString() );
	    return 1;
	}

	if( opts['?'] )
	{
	    if( argc && !LStrCmpI( argv[0], "examples" ) )
		printf( "%s", c_pszUsageExamples );
	    else
		printf( "%s", c_pszUsageLong );
	    return 0;
	}

	if( opts['V'] )
	{
	    printf( "%s", c_pszVersion );
	    return 0;
	}

	bool fCSV = !!opts['c'];
	bool fDumpRawPDB = !!opts['d'];
	bool fHSDB = !!opts['h'];
	bool fPDB = !!opts['p'];
	bool const fText = !!opts['t'];
	bool const fXML = !!opts['x'];

	if( fCSV + fDumpRawPDB + fHSDB + fPDB + fText + fXML > 1 )
	{
	    fprintf( stderr, "Can't use more than one of -c, -d, -h, -p, -t, and -x.\n" );
	    return 1;
	}

	bool const fForceOverwrite = !!opts['f'];
	bool const fUseStoreIds = !!opts['i'];
	bool const fNoLogo = !!opts['n'];
	bool const fUseCommaDecimal = !!opts['C'];
	bool const fSpreadsheetCompatibility = !!opts['X'];

	g_intl.SetCommaDecimalMode( fUseCommaDecimal );

	// Check number of arguments.

	int optFlags = OPT_TWO;

	if( fText || fXML )
	    optFlags = OPT_ONE|OPT_TWO;
	else if( fDumpRawPDB )
	    optFlags = OPT_ONE;

	if( !opts.Parse( argc, argv, "", optFlags, c_pszUsageShort ) )
	{
	    fprintf( stderr, "%s", opts.GetErrorString() );
	    return 1;
	}

	// Determine input and output file types.

	StrT sFrom;
	StrT sTo;

	if( argc > 0 )
	    sFrom.SetA( argv[0] );
	if( argc > 1 )
	    sTo.SetA( argv[1] );

	ConversionFormatType const formatFrom = ConversionFormatTypeFromExtension( sFrom.Text() );

	if( formatFrom == CFT_UNKNOWN )
	{
	    fprintf( stderr, "Unrecognized input file type extension.\n" );
	    return 1;
	}

	if( !fDumpRawPDB && !fText && !fXML )
	{
	    Assert( argc == 2 );

	    ConversionFormatType const formatTo = ConversionFormatTypeFromExtension( sTo.Text() );

	    switch( formatTo )
	    {
	    case CFT_MOBILE:
		if( fPDB || fCSV )
		{
		    fprintf( stderr, "Conflict between %s and '%s'.\n", fPDB ? "-p" : "-c", argv[1] );
		    return 1;
		}
		fHSDB = true;
		break;
	    case CFT_PALM:
		if( fHSDB || fCSV )
		{
		    fprintf( stderr, "Conflict between %s and '%s'.\n", fHSDB ? "-h" : "-c", argv[1] );
		    return 1;
		}
		fPDB = true;
		break;
	    case CFT_CSV:
		if( fHSDB || fPDB )
		{
		    fprintf( stderr, "Conflict between %s and '%s'.\n", fHSDB ? "-h" : "-p", argv[1] );
		    return 1;
		}
		fCSV = true;
		break;
	    default:
		Assert( fCSV + fHSDB + fPDB <= 1 );
		if( fCSV )
		    sTo.Append( TEXT( ".csv" ) );
#ifdef INCLUDE_HSDB_SUPPORT
		else if( fHSDB )
		    sTo.Append( TEXT( ".hsdb" ) );
#endif
		else if( fPDB )
		    sTo.Append( TEXT( ".pdb" ) );
		else
		{
		    fprintf( stderr, "Unable to determine which output format to use.\n" );
		    return 1;
		}
	    }
	}

	ConversionFormatType const formatTo = ConversionFormatTypeFromExtension( sTo.Text() );

	Assert( FImplies( !fCSV && !fDumpRawPDB && !fText && !fXML, formatTo != CFT_UNKNOWN ) );

	// Perform the indicated action(s).

	if( !fNoLogo )
	    printf( "%s", c_pszLogo );

	if( formatTo == formatFrom )
	{
	    fprintf( stderr, "Warning: input file type and output file type are the same.\n"
		     "The converter will continue in case that was intentional.\n\n" );
	}

	if( fDumpRawPDB )
	{
	    Assert( argc == 1 );

	    if( formatFrom != CFT_PALM )
	    {
		fprintf( stderr, "The -d flag requires a Palm (PDB) format database.\n" );
		return 1;
	    }

	    return DoDumpRawPDB( sFrom.Text() );
	}

	if( formatFrom == CFT_CSV )
	{
	    Assert( !fText );
	    Assert( !fXML );
	    Assert( argc == 2 );

	    return DoImportCSV( sFrom.Text(), sTo.Text(), fUseStoreIds, fForceOverwrite );
	}

	Assert( formatFrom == CFT_PALM || formatFrom == CFT_MOBILE );
	Assert( FImplies( fCSV, argc == 2 ) );

	if( formatTo == CFT_PALM || formatTo == CFT_MOBILE )
	    return DoDbConvert( sFrom.Text(), sTo.Text(), fForceOverwrite );

	Assert( fCSV + fHSDB + fPDB + fText + fXML == 1 );
	Assert( fCSV || fText || fXML );

	if( formatFrom == CFT_CSV && fSpreadsheetCompatibility )
	{
	    if( formatTo != CFT_CSV )
	    {
		fprintf( stderr, "The -X flag only applies when creating a CSV file; it is not supported when\n"
			 "reading a CSV file." );
		return 1;
	    }

	    fprintf( stderr, "Warning: the -X flag only applies to the output file; any single quotes in the\n"
		     "input file will become part of the text strings.\n\n" );
	}

	SPNew<OutputInterface> spout;

	if( fCSV )
	    spout = NewOutputCSV( fSpreadsheetCompatibility );
	else if( fText )
	    spout = NewOutputText();
	else if( fXML )
	    spout = NewOutputXML();

	Assert( spout );

	spout->SetFileName( sTo.Text(), fForceOverwrite );
	spout->SetFriendlyNames( !fUseStoreIds );

	SPNew<DbOpen> spdbFrom;
	SysError const eOpen = OpenDatabase( sFrom.Text(), &spdbFrom );

	if( eOpen )
	{
	    fprintf( stderr, "Unable to open database '%S'.\n", sFrom.Text() );
	    return 1;
	}

	return DoOutputDatabase( spdbFrom, spout );
}

