/*
 * Dbg.cpp - Debugging aids.
 */

// vim: set noet:			// Hard tabs.
// vim: set ts=8:			// Tab width 8.
// vim: set sw=4:			// Shift (indent) width 4.
// vim: set cino={0s:			// Braces flush with outer statement.
// vim: set cino=^1s:			// First indent 8 (add 1 shift width).

#include "precomp.h"

#include "dbg.h"
#include "str.h"
#include "hidpi.h"

#ifdef _WIN32_WCE
extern bool g_fSmartphone;
#endif

/*
 * __DbgTrace - show debug trace output.
 */

#ifdef DEBUG

static bool s_fTrace = false;
static bool s_fAsserts = true;

void
EnableTracing( bool fEnable )
{
	s_fTrace = fEnable;
}

void
EnableAsserts( bool fEnable )
{
	s_fAsserts = fEnable;
}

bool
AreAssertsEnabled()
{
	return s_fAsserts;
}

void
__DbgTrace( LPCSTR pszFmt, ... )
{
	if( s_fTrace )
	{
	    PreserveLastError ple;

	    StrA s;
	    StrA sFmt;
	    va_list args;
	    va_start( args, pszFmt );

	    sFmt.Set( pszFmt );
	    s.PrintfV( sFmt.Text(), args );

#if defined( OS_WINDOWSMOBILE )
	    s.AppendNativeLineEnding();
	    StrW sWide;
	    sWide.SetA( s );
	    OutputDebugStringW( sWide.Text() );
#elif defined( OS_WINDOWS )
	    s.AppendNativeLineEnding();
	    OutputDebugStringA( s.Text() );
#elif defined( OS_CRT )
	    fprintf( stderr, "%s\n", s.Text() );
#else
#error __DbgTrace not implemented for this platform.
#endif

	    va_end( args );
	}
}

#ifdef UNICODE
void
__DbgTrace( LPCWSTR pszFmt, ... )
{
	if( s_fTrace )
	{
	    PreserveLastError ple;

	    StrW s;
	    va_list args;
	    va_start( args, pszFmt );

	    s.PrintfV( pszFmt, args );

#if defined( OS_WINDOWS )
	    s.AppendNativeLineEnding();
	    OutputDebugStringW( s.Text() );
#elif defined( OS_CRT )
	    fwprintf( stderr, L"%s\n", s.Text() );
#else
#error __DbgTrace not implemented for this platform.
#endif

	    va_end( args );
	}
}
#endif

/*
 * __DbgMsg - show debug message.
 */

static void
__DbgMsgRaw( int tenths, LPCTSTR pszMsg )
{
	if( s_fTrace )
	{
	    PreserveLastError ple;

#if defined( OS_WINDOWSMOBILE )

	    RECT rcScreen;
	    RECT rcDraw;
	    RECT rcText;

	    HideCaret( 0 );

	    SetRect( &rcScreen, 0, 0, GetSystemMetrics( SM_CXSCREEN ), GetSystemMetrics( SM_CYSCREEN ) );
	    InflateRect( &rcScreen, 0, -SCALEY( 28 ) );
	    rcDraw = rcScreen;
	    InflateRect( &rcDraw, -SCALEX( 8 ), -SCALEY( 8 ) );

	    int const cxScreen = rcScreen.right - rcScreen.left;
	    int const cyScreen = rcScreen.bottom - rcScreen.top;

	    HDC const hdcScreen = GetDC( 0 );
	    HDC const hdcMemory = CreateCompatibleDC( hdcScreen );
	    HBITMAP const hbmpSave = CreateCompatibleBitmap( hdcScreen, cxScreen, cyScreen );
	    HBITMAP const hbmpOld = HBITMAP( SelectObject( hdcMemory, hbmpSave ) );

	    SaveDC( hdcScreen );
	    SelectObject( hdcScreen, GetStockObject( SYSTEM_FONT ) );
	    SetBkMode( hdcScreen, OPAQUE );
	    SetBkColor( hdcScreen, RGB( 0xff, 0xff, 0xff ) );
	    SetTextColor( hdcScreen, RGB( 0, 0, 0 ) );
	    SelectPen( hdcScreen, GetStockPen( BLACK_PEN ) );
	    SelectBrush( hdcScreen, GetStockBrush( WHITE_BRUSH ) );

	    BitBlt( hdcMemory, 0, 0, cxScreen, cyScreen,
		    hdcScreen, rcScreen.left, rcScreen.top, SRCCOPY );

	    rcText = rcDraw;
	    InflateRect( &rcText, -8, -6 );

	    DWORD const dwFlags = DT_NOPREFIX|DT_EXPANDTABS|DT_WORDBREAK;
	    DrawText( hdcScreen, pszMsg, lstrlen( pszMsg ), &rcText, DT_CALCRECT|dwFlags );

	    rcDraw.bottom = rcDraw.top + ( rcText.bottom - rcText.top ) + 12;

	    RoundRect( hdcScreen, rcDraw.left, rcDraw.top, rcDraw.right, rcDraw.bottom, 8, 8 );
	    DrawText( hdcScreen, pszMsg, lstrlen( pszMsg ), &rcText, dwFlags );

	    if( tenths > 0 )
	    {
		Sleep( tenths * 100 );
	    }
	    else
	    {
		MSG msg;
		GetMessage( &msg, 0, WM_LBUTTONDOWN, WM_LBUTTONDOWN );
	    }

	    BitBlt( hdcScreen, rcScreen.left, rcScreen.top, cxScreen, cyScreen,
		    hdcMemory, 0, 0, SRCCOPY );

	    RestoreDC( hdcScreen, -1 );
	    SelectObject( hdcMemory, hbmpOld );
	    DeleteObject( hbmpSave );
	    DeleteDC( hdcMemory );
	    ReleaseDC( 0, hdcScreen );

	    ShowCaret( 0 );

#else

	    SUPPRESS_UNUSED_VAR( tenths );

	    __DbgTrace( TEXT( "%s" ), pszMsg );

#endif
	}
}

static TCHAR
HexDigit( BYTE x )
{
	if( x < 10 )
	    return '0' + x;

	if( x < 16 )
	    return 'A' + x - 10;

	return '*';
}

void
__DbgDump( void const* pv, unsigned cb )
{
	if( s_fTrace )
	{
	    PreserveLastError ple;

	    register BYTE const* const pb = reinterpret_cast<BYTE const*>( pv );

	    enum { c_cbMax = 8 * 10 };

	    if( cb > c_cbMax )
		cb = c_cbMax;

	    LPTSTR const psz = LPTSTR( MemAlloc( ( ( cb * 3 ) + 1 ) * sizeof( *psz ) ) );

	    for( unsigned ii = 0; ii < cb; )
	    {
		register unsigned ich = ii * 3;
		psz[ich++] = HexDigit( pb[ii] >> 4 );
		psz[ich++] = HexDigit( pb[ii] & 0x0f );
		psz[ich++] = ( ++ii ) % 8 ? ' ' : '\n';
	    }

	    psz[cb * 3] = 0;

	    __DbgMsgRaw( 0, psz );

	    MemFree( psz );
	}
}

void
__DbgMsg( int tenths, LPCSTR pszFormat, ... )
{
	if( s_fTrace )
	{
	    PreserveLastError ple;

	    StrT s;
	    StrT sFmt;
	    va_list args;
	    va_start( args, pszFormat );

	    sFmt.SetA( pszFormat );
	    s.PrintfV( sFmt.Text(), args );
	    __DbgMsgRaw( tenths, s.Text() );

	    va_end( args );
	}
}

#ifdef UNICODE
void
__DbgMsg( int tenths, LPCWSTR pszFormat, ... )
{
	if( s_fTrace )
	{
	    PreserveLastError ple;

	    StrT s;
	    va_list args;
	    va_start( args, pszFormat );

	    s.PrintfV( pszFormat, args );
	    __DbgMsgRaw( tenths, s.Text() );

	    va_end( args );
	}
}
#endif

/*
 * __DbgAssertInternal - show assert box and return the user's response.
 */

AssertResponse
__DbgAssertInternalV( LPCTSTR pszFmt, LPCTSTR pszFile, int nLine, va_list args )
{
	PreserveLastError ple;

	TCHAR sz[1024];
	LPTSTR psz = sz;
	LPTSTR const pszEnd = sz + DimensionOf( sz );

	psz += _stprintf_s( psz, pszEnd - psz, TEXT( "%s, line %d:\n" ), pszFile, nLine );
	psz += _vstprintf_s( psz, pszEnd - psz, pszFmt, args );

	__DbgTrace( TEXT( "%s" ), sz );

	if( !s_fAsserts )
	    return AR_CONTINUE;

#if defined( OS_WINDOWS )
#ifdef _WIN32_WCE
	DWORD const dwMbFlags = g_fSmartphone ? MB_OKCANCEL : MB_ABORTRETRYIGNORE;
#else
	DWORD const dwMbFlags = MB_ABORTRETRYIGNORE;
#endif
	switch( MessageBox( GetForegroundWindow(), sz, TEXT( "Assert" ), dwMbFlags ) )
	{
	case IDABORT:
	    return AR_ABORT;
	case IDRETRY:
	case IDCANCEL:
	    return AR_BREAK;
	default:
	    return AR_CONTINUE;
	}
#elif defined( OS_CRT )
	return AR_BREAK;
#else
#error __DbgAssertInternalV not implemented for this platform.
#endif
}

AssertResponse
__DbgAssertInternal( LPCTSTR pszFmt, LPCTSTR pszFile, int nLine, ... )
{
	va_list args;
	va_start( args, nLine );

	AssertResponse const response = __DbgAssertInternalV( pszFmt, pszFile, nLine, args );

	va_end( args );

	return response;
}

/*
 * __DbgAssert - report an assertion and prompt the user to abort, break, or
 * continue.
 */

void
__DbgAssert( LPCTSTR pszFmt, LPCTSTR pszFile, int nLine, ... )
{
	PreserveLastError ple;

	va_list args;
	va_start( args, nLine );

	switch( __DbgAssertInternalV( pszFmt, pszFile, nLine, args ) )
	{
	case AR_ABORT:	__DbgAbort(); break;
	case AR_BREAK:	__DbgBreak(); break;
	}

	va_end( args );
}

#endif // DEBUG

