/*
 * DbOpen.cpp - Database abstraction layer.
 */

// vim: set noet:			// Hard tabs.
// vim: set ts=8:			// Tab width 8.
// vim: set sw=4:			// Shift (indent) width 4.
// vim: set cino={0s:			// Braces flush with outer statement.
// vim: set cino=^1s:			// First indent 8 (add 1 shift width).

#include "precomp.h"

// System headers.

// App headers.

#ifdef _WIN32_WCE
#include "res.h"
#endif

#include "dbopen.h"
#include "sys.h"
#include "storelist.h"
#include "item.h"
#include "viewdefs.h"

#include "palmlayer.h"			// for StrCaselessCompareCount used by StrLookup

/*
 * DbFormatTypeFromExtension.
 */

DbFormatType
DbFormatTypeFromExtension( LPCTSTR const pszFile )
{
	Assert( pszFile );

	TCHAR const* const pszExt = GetExtension( pszFile );

	if( pszExt )
	{
	    if( !_tcsicmp( pszExt, TEXT( ".pdb" ) ) )
		return DBFT_PALM;
#ifdef INCLUDE_HSDB_SUPPORT
	    else if( !_tcsicmp( pszExt, TEXT( ".hsdb" ) ) )
		return DBFT_MOBILE;
#endif
	}

	return DBFT_UNKNOWN;
}

/*
 * MultipleCategories.
 */

void
MultipleCategories::SetAll( bool fMarked )
{
	CASSERT( sizeof( m_rg ) == DimensionOf( m_rg ) );

	BYTE const value = fMarked ? true : false;

	m_Unfiled = value;
	memset( m_rg, value, sizeof( m_rg ) );
}

unsigned
MultipleCategories::CountMarked( DbOpen const* const pdb ) const
{
	Assert( pdb );

	unsigned cMarked = 0;

	// Count how many categories are marked.

	for( unsigned ii = pdb->GetCategoriesCount(); ii--; )
	    if( m_rg[pdb->GetCategoryIdByIndex( ii )] )
		cMarked++;

	if( m_Unfiled )
	    cMarked++;

	return cMarked;
}

/*
 * DbFont.
 */

FontId
FontIdFromPalmFont( BYTE idPalm )
{
	switch( idPalm )
	{
	case 1:		return FID_BOLD;
	case 7:		return FID_LARGEBOLD;
	case 0x80:	return FID_NARROW;
	case 0x82:	return FID_NARROW;	// Obsolete.
	case 0x81:	return FID_TINY;
	default:	return FID_NORMAL;
	}
}

BYTE
PalmFontFromFontId( FontId idFont )
{
	switch( idFont )
	{
	case FID_BOLD:		return 1;
	case FID_LARGEBOLD:	return 7;
	case FID_NARROW:	return 0x80;
	case FID_TINY:		return 0x81;
	default:		return 0;
	}
}

/*
 * MultipleStores.
 */

void
MultipleStores::SetAll( bool fMarked )
{
	CASSERT( sizeof( m_rg ) == DimensionOf( m_rg ) );

	BYTE const value = fMarked ? 0x03 : 0x00;

	m_Unfiled = value;
	memset( m_rg, value, sizeof( m_rg ) );
}

unsigned
MultipleStores::CountMarked( DbOpen* const pdb ) const
{
	Assert( pdb );

	unsigned cMarked = 0;

	// Count how many stores are marked.

	for( unsigned ii = pdb->GetStoresCount(); ii--; )
	    if( m_rg[pdb->GetStoreIdByIndex( ii )] )
		cMarked++;

	if( m_Unfiled )
	    cMarked++;

	return cMarked;
}

/*
 * DbGroupTransaction - Batches multiple transactions into a single
 * transaction.
 */

DbGroupTransaction::~DbGroupTransaction()
{
	AssertSz( !m_fActive, "***  DATA LOSS!  ***\r\n\r\nTransaction group was not ended properly!" );
}

/*
 * DbOpen - Database abstraction.
 */

PrivateRecordVisualMode DbOpenSimple::s_privDefaultRecordVisualMode = privShowRecords;

void
DbOpenSimple::SetPrivateRecordVisualMode( PrivateRecordVisualMode const priv )
{
	m_privRecordVisualMode = priv;
}

DbOpenSimple* SimpleOpenHs2Database( LPCTSTR pszFile );
DbQuickOpen* QuickOpenHs2Database( LPCTSTR pszFile );
#ifdef INCLUDE_HSDB_SUPPORT
DbOpenSimple* SimpleOpenHs3Database( LPCTSTR pszFile );
DbQuickOpen* QuickOpenHs3Database( LPCTSTR pszFile );
#endif

TCHAR DbOpen::s_szUnfiled[c_cchMaxCategoryName] = TEXT( "Unfiled" );
TCHAR DbOpen::s_szAll[c_cchMaxCategoryName] = TEXT( "All" );
TCHAR DbOpen::s_szAllStores[c_cchMaxStoreName] = TEXT( "All Stores" );
TCHAR DbOpen::s_szDefaultCurrency[8] = TEXT( "$" );

LPCTSTR
DbOpen::UnfiledString()
{
	return s_szUnfiled;
}

LPCTSTR
DbOpen::AllString()
{
	return s_szAll;
}

LPCTSTR
DbOpen::AllStoresString()
{
	return s_szAllStores;
}

LPCTSTR
DbOpen::DefaultCurrencyString()
{
	return s_szDefaultCurrency;
}

void
DbOpen::SetLocalizedStrings( LPCTSTR pszUnfiled, LPCTSTR pszAll, LPCTSTR pszAllStores, LPCTSTR pszDefaultCurrency )
{
	CchCopy( pszUnfiled, RgC( s_szUnfiled ) );
	CchCopy( pszAll, RgC( s_szAll ) );
	CchCopy( pszAllStores, RgC( s_szAllStores ) );
	CchCopy( pszDefaultCurrency, RgC( s_szDefaultCurrency ) );
}

DbOpen::DbOpen()
{
	m_pbits = new BitTable( this );
}

DbOpen::~DbOpen()
{
	delete m_pbits;
}

void
DbOpen::SetPrivateRecordVisualMode( PrivateRecordVisualMode const priv )
{
	if( m_pbits && priv != GetPrivateRecordVisualMode() )
	    m_pbits->Clear();

	base::SetPrivateRecordVisualMode( priv );
}

bool
DbOpen::SetLookup( StrT const* const psLookup, bool const fBroad )
{
	if( psLookup )
	{
	    bool const fNewText = ( LStrCmp( psLookup->Text(), m_sLookup.Text() ) != 0 );
	    bool const fNewBroad = ( fBroad != m_fFilteringDisabled );

	    if( !fNewText && !fNewBroad )
		return false;

	    bool const fCanRefineText = m_sLookup.Length() && !_tcsnicmp( psLookup->Text(), m_sLookup.Text(), m_sLookup.Length() );
	    bool const fCanRefineBroad = !fNewBroad || !fBroad;
	    bool const fRefine = fCanRefineText && fCanRefineBroad;

	    m_sLookup.Set( *psLookup );
#ifdef UNICODE
	    m_sLookupA.SetW( m_sLookup );
#else
	    m_sLookupA.SetA( m_sLookup );
#endif
	    m_fFilteringDisabled = fBroad;

	    if( fRefine )
		m_pbits->RefineScan();
	    else
		m_pbits->Clear();
	}
	else
	{
	    if( !m_sLookup.Length() && !m_fFilteringDisabled )
		return false;

	    m_sLookup.Clear();
	    m_sLookupA.Clear();
	    m_fFilteringDisabled = false;

	    m_pbits->Clear();
	}

	return true;
}

LPCTSTR
DbOpen::GetLookup() const
{
	if( !m_sLookup.Length() )
	    return 0;

	return m_sLookup.Text();
}

bool
DbOpen::IsLookupBroad() const
{
	return m_fFilteringDisabled;
}

bool
DbOpen::SetFilterReasonCompleteness( bool const fComplete )
{
	bool const fRet = m_fCompleteReasons;

	m_fCompleteReasons = fComplete;

	return fRet;
}

SysError
DbOpen::ConvertFrom( DbOpen const* const pdbFrom )
{
	Assert( !Count() );
	Assert( !GetCategoriesCount() );
	Assert( !GetStoresCount() );
	Assert( pdbFrom );

	unsigned ii;
	WORD rgXlateCategory[c_cMaxCategories];
	Store rgXlateStore[c_cMaxStores];

	for( ii = DimensionOf( rgXlateCategory ); ii--; )
	    rgXlateCategory[ii] = catUnfiledCategory;

	for( ii = DimensionOf( rgXlateStore ); ii--; )
	    rgXlateStore[ii] = storeUnfiledStore;

	// First check that the new format can contain all the data.

	if( pdbFrom->GetCategoriesCount() > GetCategoriesMaximum() )
	    return E_TOOMANYCATEGORIES;

	if( pdbFrom->GetStoresCount() > GetStoresMaximum() )
	    return E_TOOMANYSTORES;

	// Start a transaction group, to improve performance.

	SPNew<DbGroupTransaction> spXact;
	SysError const eXactTables = MakeTransactionGroup( spXact );

	if( eXactTables )
	    return eXactTables;

	// Copy categories, building a translation table.

	for( ii = 0; ii < pdbFrom->GetCategoriesCount(); ii++ )
	{
	    WORD const oldCategory = pdbFrom->GetCategoryIdByIndex( ii );
	    LPCTSTR const pszCategory = pdbFrom->GetCategoryNameByIndex( ii );
	    WORD const icon = pdbFrom->GetCategoryIcon( oldCategory );

	    if( !CategoryNew( pszCategory, icon ) )
		return SE_UNKNOWN;

	    WORD const newCategory = GetCategoryIdByName( pszCategory );

	    Assert( newCategory != catUnfiledCategory );

	    rgXlateCategory[oldCategory] = newCategory;
	}

	// Copy stores, building a translation table.

	for( ii = 0; ii < pdbFrom->GetStoresCount(); ii++ )
	{
	    Store const oldStore = pdbFrom->GetStoreIdByIndex( ii );
	    LPCTSTR const pszStore = pdbFrom->GetStoreNameByIndex( ii );

	    if( !StoreNew( pszStore ) )
		return SE_UNKNOWN;

	    Store const newStore = GetStoreIdByName( pszStore );

	    Assert( newStore != storeUnfiledStore );

	    rgXlateStore[oldStore] = newStore;
	}

	// Copy view settings.

	WORD category = pdbFrom->GetCategory();
	Store store = pdbFrom->GetStore();
	MultipleCategories multiCategoriesFrom;
	MultipleStores multiStoresFrom;

	if( !IsPhantomCategory( category ) )
	    category = rgXlateCategory[category];
	if( !IsPhantomStore( store ) )
	    store = rgXlateStore[store];

	SetView( pdbFrom->GetView() );
	SetCategory( category );
	SetStore( store );
	SetHideChecked( pdbFrom->GetHideChecked() );
	SetCouponsOnly( pdbFrom->GetCouponsOnly() );

	if( pdbFrom->GetMultipleCategories( &multiCategoriesFrom ) )
	{
	    MultipleCategories multiCategoriesTo;
	    multiCategoriesTo.SetAll( false );

	    for( ii = DimensionOf( multiCategoriesTo.m_rg ); ii--; )
	    {
		WORD const newCategory = rgXlateCategory[ii];

		if( IsPhantomCategory( newCategory ) )
		    continue;

		multiCategoriesTo.m_rg[newCategory] = multiCategoriesFrom.m_rg[ii];
	    }

	    SetMultipleCategories( &multiCategoriesTo );
	}

	if( pdbFrom->GetMultipleStores( &multiStoresFrom ) )
	{
	    MultipleStores multiStoresTo;
	    multiStoresTo.SetAll( false );

	    for( ii = DimensionOf( multiStoresTo.m_rg ); ii--; )
	    {
		Store const newStore = rgXlateStore[ii];

		if( IsPhantomStore( newStore ) )
		    continue;

		multiStoresTo.m_rg[newStore] = multiStoresFrom.m_rg[ii];
	    }

	    SetMultipleStores( &multiStoresTo );
	}

	for( ii = 0; ii < COLID_MAX; ii++ )
	    SetColumnWidth( ColumnIdentifier( ii ), pdbFrom->GetColumnWidth( ColumnIdentifier( ii ) ) );

	// Copy sort settings.

	SetSortOneOrder( pdbFrom->GetSortOneOrder() );

	for( ii = 0; ii < NUM_SORT_ORDERS; ii++ )
	{
	    SetSortOrderInfo( ii, pdbFrom->GetSortOrderInfo( ii ) );
	    SetCanLearnOrder( ii, pdbFrom->CanLearnOrder( ii ) );
	}

	// Copy other settings.

	SetRoundToWholeNumbers( pdbFrom->GetRoundToWholeNumbers() );
	SetCurrency( pdbFrom->GetCurrency() );
	SetTaxRate( pdbFrom->GetTaxRate() );
	SetTaxRate2( pdbFrom->GetTaxRate2() );
	SetTwoTaxes( pdbFrom->GetTwoTaxes() );
	SetPerStorePrices( pdbFrom->GetPerStorePrices() );
	SetAutoQuantity( pdbFrom->GetAutoQuantity() );
	SetAutoDelete( pdbFrom->GetAutoDelete() );
	SetRecordCreationDate( pdbFrom->GetRecordCreationDate() );
	SetRecordCompletionDate( pdbFrom->GetRecordCompletionDate() );
	SetTaxable( pdbFrom->GetTaxable() );
	SetDefaultPriority( pdbFrom->GetDefaultPriority() );
	SetRowBackgroundColor( pdbFrom->GetRowBackgroundColor() );
	SetFont( pdbFrom->GetFont() );
	SetMaxLinesPerItem( pdbFrom->GetMaxLinesPerItem() );
	SetIcon( pdbFrom->GetIcon() );

	SetExportFlags( pdbFrom->GetExportFlags() );

	// Copy records, using the tables from above to translate the
	// categories and stores.

	ScopedProgressHolder sph;
	sph.Create();
	sph.Init( pdbFrom->Count(), PROGRESS_PREDICTIVEWAIT );

	for( unsigned record = 0; record < pdbFrom->Count(); record++ )
	{
	    sph.Update( record );

	    // Use batched transaction groups to improve performance, while
	    // still limiting memory consumption.

	    if( !( record % c_cRecordsPerBatch ) || !spXact )
	    {
		SysError const eXactBatch = MakeTransactionGroup( spXact );

		if( eXactBatch )
		    return eXactBatch;
	    }

	    // Unpack the 'from' record.

	    SPNew<Item> spitem = pdbFrom->UnPackRecord( record );

	    Assert( spitem );

	    if( !spitem )
		return SE_UNKNOWN;

	    // Translate category and stores.

	    if( spitem->GetCategory() != catUnfiledCategory )
		spitem->SetCategory( rgXlateCategory[spitem->GetCategory()] );

	    for( ii = spitem->m_stores.Count(); ii--; )
	    {
		Store* const p = const_cast<Store*>( &spitem->m_stores.Array()[ii] );

		*p = rgXlateStore[*p];
	    }

	    for( ii = spitem->m_perstore.Count(); ii--; )
	    {
		PerStore* const p = const_cast<PerStore*>( &spitem->m_perstore.Array()[ii] );

		p->m_idStore = rgXlateStore[p->m_idStore];
	    }

	    // The store id's may have changed, so the lists must be resorted.

	    spitem->m_stores.Sort();
	    spitem->m_perstore.Sort();

	    // Pack the translated record into this database.

	    SysError const ePack = PackRecord( noRecordSelected, spitem.UnsafeAddress(), packNoSort|packNoUndo, 0 );

	    if( ePack )
		return ePack;
	}

	sph.End();

	if( spXact )
	{
	    SysError const eEndXactBatch = spXact->End();

	    if( eEndXactBatch )
		return eEndXactBatch;

	    spXact.Free();
	}

	SysError const eSort = SortDatabase();

	if( eSort )
	    return eSort;

	return SE_OK;
}

SysError
DbOpen::MakeTransactionGroup( SPNew<DbGroupTransaction>& spXact )
{
	if( spXact )
	{
	    SysError const eEnd = spXact->End();

	    if( eEnd )
		return eEnd;

	    spXact.Free();
	}

	SysError const eNew = NewDbGroupTransaction( &spXact );

	if( eNew )
	    return eNew;

	return SE_OK;
}

SysError
DbOpen::Checkout( bool const fProgress )
{
	SPNew<DbGroupTransaction> spXact;
	unsigned cRecords = Count();
	unsigned cProgress;
	SysError e = SE_OK;

	ClearUndo();

	ScopedProgressHolder sph;
	if( fProgress )
	    sph.Create();
	sph.Init( cRecords, PROGRESS_PREDICTIVEWAIT );

	cProgress = 0;
	for( unsigned record = 0; record < cRecords; cProgress++, record++ )
	{
	    if( !( cProgress % c_cRecordsPerBatch ) || !spXact )
	    {
		e = MakeTransactionGroup( spXact );

		if( e )
		    break;
	    }

	    sph.Update( cProgress );

	    WORD ces;
	    unsigned recordAfterPack;
	    e = CheckoutRecord( record, &ces, &recordAfterPack );

	    if( e )
		break;

	    if( ces & cesDeleted )
	    {
		cRecords--;
		record--;
	    }
	    else if( ( ces & cesMoved ) && recordAfterPack > record )
	    {
		record--;
		cProgress--;
	    }
	}

	// Deactivate the Learn Order mode if it is currently active.

	if( IsLearningOrder() )
	    ActivateLearningOrder( false );

	if( spXact )
	{
	    SysError const eEnd = spXact->End();

	    if( !e )
		e = eEnd;
	}

	return e;
}

SysError
DbOpen::CheckoutRecord( unsigned const record, WORD* const pces, unsigned* const precordAfterPack )
{
	Assert( pces );
	Assert( precordAfterPack );

	*pces = 0;
	*precordAfterPack = unsigned( -1 );

	// We can frequently avoid the more expensive act of unpacking the
	// record, by first check the item's quick info.

	ItemQuickInfo qi;

	if( !GetItemQuickInfo( record, &qi ) )
	    return SE_OK;

	// Old HS2 databases can potentially have records that are marked as
	// deleted but haven't actually been deleted.  This is a good
	// opportunity to clean them up.

	if( qi.IsRecordDeleted() )
	{
	    *pces = cesDeleted;
	    return DeleteRecord( record, false );
	}

	// Process items that are not needed but also not purchased (i.e. in
	// your shopping cart).

	if( qi.GetState() == ITEMSTATE_CHECKED )
	{
	    // Delete items that are Auto-Delete.

	    if( qi.IsAutoDelete() )
	    {
		*pces = cesDeleted;
		return DeleteRecord( record, false );
	    }

	    // Make the item purchased (not needed, not in your cart).

	    SPNew<Item> spItem = UnPackRecord( record );

	    if( !spItem )
		return SE_UNKNOWN;

	    spItem->SetState( ITEMSTATE_NONE, this );

	    if( GetRecordCompletionDate() )
		spItem->StampToday();

	    // Reset the alarm.

#if 0
	    if( flags & flagHasAlarm )
		Item_SetAlarm( pitem, 0 );
#endif

	    // Pack the record.

	    SysError const ePack = PackRecord( record, spItem.UnsafeAddress(), packForceSort|packNoUndo, pces );

	    if( ePack )
		return ePack;

	    *pces &= ~cesFiltered;

	    Assert( *pces == 0 || *pces == cesDeleted || *pces == cesMoved );	// Can't be both deleted and moved.

	    *precordAfterPack = GetRecordAfterPack();
	}
	else if( qi.GetState() == ITEMSTATE_DEFERRED )
	{
	    // Make the item needed.

	    SPNew<Item> spItem = UnPackRecord( record );

	    if( !spItem )
		return SE_UNKNOWN;

	    spItem->SetState( ITEMSTATE_NEED, this );

	    // Pack the record.

	    SysError const ePack = PackRecord( record, spItem.UnsafeAddress(), packForceSort|packNoUndo, pces );

	    if( ePack )
		return ePack;

	    *pces &= ~cesFiltered;

	    Assert( *pces == 0 || *pces == cesDeleted || *pces == cesMoved );	// Can't be both deleted and moved.

	    *precordAfterPack = GetRecordAfterPack();

	    AssertSz( *precordAfterPack <= record, "Deferred item got moved to a higher record index, will get double-processed by Checkout." );
	}

	return SE_OK;
}

bool
DbOpen::IsSortNeededBasedOnStoreChange( BYTE const newStore ) const
{
	// If the store isn't changing there's nothing to do.

	if( GetStore() == newStore )
	    return false;

	// If the store was a phantom store and still is a phantom store then
	// the sort order won't actually change and therefore sort is not
	// needed.

	bool const fWasOldPhantom = IsPhantomStore( GetStore() );
	bool const fIsNewPhantom = IsPhantomStore( newStore );

	if( fWasOldPhantom && fIsNewPhantom )
	    return false;

	// AutoOrder is always per-store, so when sorted by AutoOrder sort is
	// always needed when changing stores.

	if( CanLearnOrder( FAltSort() ) )
	    return true;

	// Otherwise if per-store prices/aisles are not activated then sort is
	// not needed when changing stores (particularly if the store isn't
	// actually changing!).

	if( !GetPerStorePrices() )
	    return false;

	// Finally if the sort order does not include the Price or Aisle
	// columns then the sort order won't actually change and therefore
	// sort is not needed.

	if( !IsSortedByColumn( COLID_PRICE ) &&
	    !IsSortedByColumn( COLID_AISLE ) )
	    return false;

	return true;
}

/*
 * Auto Order - learns the order in which you shop for items.
 *
 *	If an item is purchased "out of order" it gets kicked to the end of
 *	the order.  The main drawback is this means that all subsequent
 *	purchases now become "out of order" as well.  The approach can
 *	converge on a stable order, but any purchase out of order causes a new
 *	divergence.  The concern is that divergence will be common, and may
 *	cause the approach to never actually converge.
 */

enum WhichBounds
{
	WB_CHECKED_HI,
	WB_FIND,
	WB_ALL_HI,
	WB_ALL_ALL,
};

struct BoundingRecords
{
	unsigned rgRecordLo[ITEMSTATE_MAX];
	unsigned rgRecordHi[ITEMSTATE_MAX];
};

static bool
GetBoundingRecords( DbOpen const* const pdb, BoundingRecords* const p, WhichBounds const which, DWORD const ulFindOrder = 0 )
{
	Assert( pdb );
	Assert( p );

	// The sort order for flags is flagNeed, flagNeed|flagPurchased, 0,
	// flagPurchased.  So, the order values are ITEMSTATE_NEED=0,
	// ITEMSTATE_DEFERRED=1, ITEMSTATE_CHECKED=2, ITEMSTATE_NONE=3.

	static BYTE const c_rgCmpValue[] = { 3, 0, 2, 1 };
	CASSERT( DimensionOf( c_rgCmpValue ) == ITEMSTATE_MAX );

	// Find the lower/upper bounds for each item state.

	ZeroMemory( p, sizeof( p ) );

	for( unsigned ii = 0; ii < ITEMSTATE_MAX; ii++ )
	{
	    unsigned recordLo;
	    unsigned recordHi;
	    ItemQuickInfo qi;

	    // Find lower bound for each item state.

	    if( which == WB_ALL_ALL )
	    {
		recordLo = 0;
		recordHi = pdb->Count();

		while( recordLo < recordHi )
		{
		    unsigned recordMid = ( recordLo + recordHi ) / 2;

		    if( !pdb->GetItemQuickInfo( recordMid, &qi ) )
			return false;

		    if( c_rgCmpValue[qi.GetState()] >= c_rgCmpValue[ii] )
			recordHi = recordMid;
		    else
			recordLo = recordMid + 1;
		}

#ifdef DEBUG
		if( recordLo < pdb->Count() && pdb->GetItemQuickInfo( recordLo, &qi ) )
		    Assert( c_rgCmpValue[qi.GetState()] >= c_rgCmpValue[ii] );
		if( recordLo > 0 && pdb->GetItemQuickInfo( recordLo - 1, &qi ) )
		{
		    Assert( qi.GetState() != ItemState( ii ) );
		    Assert( c_rgCmpValue[qi.GetState()] < c_rgCmpValue[ii] );
		}
#endif

		p->rgRecordLo[ii] = recordLo;
	    }

	    // Find upper bound for each item state.

	    if( which != WB_CHECKED_HI || ii == ITEMSTATE_CHECKED )
	    {
		recordLo = 0;
		recordHi = pdb->Count();

		while( recordLo < recordHi )
		{
		    unsigned recordMid = ( recordLo + recordHi ) / 2;

		    if( !pdb->GetItemQuickInfo( recordMid, &qi ) )
			return false;

		    if( which == WB_FIND )
		    {
			if( c_rgCmpValue[qi.GetState()] < c_rgCmpValue[ii] )
			{
			    recordLo = recordMid + 1;
			}
			else if( c_rgCmpValue[qi.GetState()] > c_rgCmpValue[ii] )
			{
			    recordHi = recordMid;
			}
			else
			{
			    SPNew<Item> spitem = pdb->UnPackRecord( recordMid );
			    if( !spitem )
				return false;
			    if( spitem->GetAutoOrder( pdb->GetPerStorePrices(), pdb->GetStore() ) <= ulFindOrder )
				recordLo = recordMid + 1;
			    else
				recordHi = recordMid;
			}
		    }
		    else
		    {
			if( c_rgCmpValue[qi.GetState()] >= c_rgCmpValue[ii] + 1 )
			    recordHi = recordMid;
			else
			    recordLo = recordMid + 1;
		    }
		}

#ifdef DEBUG
		if( which != WB_FIND )
		{
		    if( recordLo < pdb->Count() && pdb->GetItemQuickInfo( recordLo, &qi ) )
		    {
			Assert( qi.GetState() != ItemState( ii ) );
			Assert( c_rgCmpValue[qi.GetState()] > c_rgCmpValue[ii] );
		    }
		}
		if( recordLo > 0 && pdb->GetItemQuickInfo( recordLo - 1, &qi ) )
		    Assert( c_rgCmpValue[qi.GetState()] <= c_rgCmpValue[ii] );
#endif

		p->rgRecordHi[ii] = recordLo;
	    }
	}

	return true;
}

bool
DbOpen::ActivateLearningOrderHelper( bool const fActivate )
{
	SUPPRESS_UNUSED_VAR( fActivate );

	Assert( CanLearnOrder( FAltSort() ) );

	if( !CanLearnOrder( FAltSort() ) )
	    return false;

	/*
	 * I think that with the leaping algorithm, both activating and
	 * deactivating involve no effort.
	 */

	return true;
}

void
DbOpen::AdjustAutoOrder( Item* pitem ) const
{
	DWORD const ulFloor = GetHighestAutoOrder( true );
	DWORD const ulCurrent = pitem->GetAutoOrder( GetPerStorePrices(), GetStore() );
	DWORD const ulNew = GetSuccessorOfAutoOrder( ulFloor, ulCurrent );

	if( ulCurrent != ulNew )
	    pitem->SetAutoOrder( ulNew, GetPerStorePrices(), GetStore() );
}

DWORD
DbOpen::GetHighestAutoOrder( bool fOnlyCheckedOffItems ) const
{
	/*
	 * If the list is sorted with m_fCheckedLast then it would be
	 * reasonably efficient to find (binary search) the checked off item
	 * with the highest auto order value.
	 *
	 * I specifically want to allow learning mode regardless of the value
	 * of m_fCheckedLast.  I could keep track of the highest value as
	 * items get checked off, but how can I efficiently update the highest
	 * value when the item with the current highest value gets unchecked
	 * for whatever reason?
	 *
	 * So for now learning mode will require m_fCheckedLast == true.
	 *
	 * Eventually I may maintain a look-aside table of checked off items,
	 * sorted by order value, but that will be hard to do on the Palm OS.
	 */

	BoundingRecords br;
	if( !GetBoundingRecords( this, &br, fOnlyCheckedOffItems ? WB_CHECKED_HI : WB_ALL_HI ) )
	    return 0;

	DWORD ulHighest = 0;
	for( unsigned ii = 0; ii < ITEMSTATE_MAX; ii++ )
	{
	    if( br.rgRecordHi[ii] > 0 && ( !fOnlyCheckedOffItems || ii == ITEMSTATE_CHECKED ) )
	    {
		SPNew<Item> spitem = UnPackRecord( br.rgRecordHi[ii] - 1 );
		if( spitem && ( !fOnlyCheckedOffItems || spitem->GetState() == ITEMSTATE_CHECKED ) )
		{
		    DWORD const ulAutoOrder = spitem->GetAutoOrder( GetPerStorePrices(), GetStore() );
		    if( ulHighest < ulAutoOrder )
			ulHighest = ulAutoOrder;
		}
	    }
	}

	return ulHighest;
}

DWORD
DbOpen::GetSuccessorOfAutoOrder( DWORD const ulFloor, DWORD const ulCurrent ) const
{
	Assert( CanLearnOrder( FAltSort() ) );

	if( ulCurrent >= ulFloor && ( ulCurrent || ulFloor ) )
	    return ulCurrent;

	/*
	 * Alternatively, this could just use (a transformation of) the
	 * current time as the next order value.  That seems rather attractive
	 * from a performance perspective, but at the cost of reducing how
	 * human-readable the order values are.
	 */
	//$ review: (chrisant) This could use a transformation of the current
	// time as the next order value, to improve performance.  The
	// transformation would need to be possible on both Windows Mobile and
	// Palm OS.
	DWORD const ulHighest = GetHighestAutoOrder( false );
	if( ulHighest >= ulCurrent )
	    return ulHighest + 1;

	return ulCurrent;
}

/*
 * DbOpenSimple.
 */

DbOpenSimple*
SimpleOpenDatabase( LPCTSTR pszFile )
{
	Assert( pszFile );

	LPCTSTR const pszExt = GetExtension( pszFile );

	if( pszExt && FileExists( pszFile ) )
	{
	    if( LStrCmpI( pszExt, TEXT( ".pdb" ) ) == 0 )
		return SimpleOpenHs2Database( pszFile );
#ifdef INCLUDE_HSDB_SUPPORT
	    if( LStrCmpI( pszExt, TEXT( ".hsdb" ) ) == 0 )
		return SimpleOpenHs3Database( pszFile );
#endif
	}

	return 0;
}

/*
 * DbQuickOpen.
 */

DbQuickOpen::DbQuickOpen()
{
}

DbQuickOpen::~DbQuickOpen()
{
}

bool
DbQuickOpen::IsEqual( const DbQuickOpen* pqo ) const
{
	if( LStrCmp( m_sName.Text(), pqo->m_sName.Text() ) )
	    return false;

	if( m_cItems != pqo->m_cItems )
	    return false;

	if( m_cNeeded != pqo->m_cNeeded )
	    return false;

	return true;
}

DbQuickOpen*
QuickOpenDatabase( LPCTSTR pszFile )
{
	Assert( pszFile );

	LPCTSTR const pszExt = GetExtension( pszFile );

	if( pszExt && FileExists( pszFile ) )
	{
	    if( LStrCmpI( pszExt, TEXT( ".pdb" ) ) == 0 )
		return QuickOpenHs2Database( pszFile );
#ifdef INCLUDE_HSDB_SUPPORT
	    if( LStrCmpI( pszExt, TEXT( ".hsdb" ) ) == 0 )
		return QuickOpenHs3Database( pszFile );
#endif
	}

	return 0;
}

/*
 * Open and Create databases.
 */

void MakeNewHs2DatabaseFileName( LPCTSTR pszDir, LPCTSTR const pszName, StrT* const psFile );
SysError NewHs2Database( LPCTSTR const pszFile, LPCTSTR const pszDisplayName, DbOpen** const ppdb );
SysError OpenHs2Database( LPCTSTR const pszFile, DbOpen** const ppdb );
void MakeNewHs3DatabaseFileName( LPCTSTR pszDir, LPCTSTR const pszName, StrT* const psFile );
SysError NewHs3Database( LPCTSTR const pszFile, LPCTSTR const pszDisplayName, DbOpen** const ppdb );
SysError OpenHs3Database( LPCTSTR const pszFile, DbOpen** const ppdb, bool const fReadonly );

SysError
MakeNewDatabaseFileName( LPCTSTR const pszDir, LPCTSTR const pszName, DbFormatType const format, StrT* const psFile )
{
	Assert( pszDir );
	Assert( pszName );
	Assert( psFile );

	psFile->Clear();

	if( format == DBFT_PALM )
	{
	    MakeNewHs2DatabaseFileName( pszDir, pszName, psFile );
	}
#if defined( INCLUDE_HSDB_SUPPORT ) && defined( CAN_CREATE_HSDB_FILES )
	else if( format == DBFT_MOBILE )
	{
	    MakeNewHs3DatabaseFileName( pszDir, pszName, psFile );
	}
#endif
	else
	{
	    Panic1( "UNEXPECTED format %u.", format );
	    return SE_UNKNOWN;
	}

	return SE_OK;
}

SysError
NewDatabase( LPCTSTR const pszFile, LPCTSTR const pszDisplayName, DbOpen** const ppdb )
{
	Assert( pszFile );
	Assert( pszDisplayName );
	Assert( ppdb );

	LPCTSTR const pszExt = GetExtension( pszFile );

	if( pszExt )
	{
	    if( LStrCmpI( pszExt, TEXT( ".pdb" ) ) == 0 )
		return NewHs2Database( pszFile, pszDisplayName, ppdb );
#if defined( INCLUDE_HSDB_SUPPORT )
	    if( LStrCmpI( pszExt, TEXT( ".hsdb" ) ) == 0 )
		return NewHs3Database( pszFile, pszDisplayName, ppdb );
#endif
	}

	Panic1( "UNEXPECTED extension on '%s'.", pszFile );
	return SE_UNKNOWN;
}

SysError
OpenDatabase( LPCTSTR const pszFile, DbOpen** const ppdb )
{
	Assert( pszFile );
	Assert( ppdb );

	if( !FileExists( pszFile ) )
	    return SE_FILENOTFOUND;

	LPCTSTR const pszExt = GetExtension( pszFile );

	if( pszExt )
	{
	    if( LStrCmpI( pszExt, TEXT( ".pdb" ) ) == 0 )
		return OpenHs2Database( pszFile, ppdb );
#ifdef INCLUDE_HSDB_SUPPORT
	    if( LStrCmpI( pszExt, TEXT( ".hsdb" ) ) == 0 )
		return OpenHs3Database( pszFile, ppdb, false );
#endif
	}

	return SE_UNKNOWN;
}

/*
 * StrLookup.
 */

//$ todo: (chrisant) use variation on Boyer-Moore fast string string

int
StrLookup( char const* const pszNeedle, size_t const cchNeedle, char const* const pszHaystack )
{
	Assert( FImplies( cchNeedle, pszNeedle ) );
	Assert( pszHaystack );

	if( !cchNeedle || !*pszHaystack )
	    return -1;

	// If the needle is alphanumeric then we'll look for matches starting
	// at a word boundary.

	bool const fAlphaNum = !!IsCharAlphaNumericA( *pszNeedle );

	for( char const* pszWalk = pszHaystack; *pszWalk; pszWalk++ )
	{
	    // Find the beginning of the next word.

	    if( fAlphaNum )
	    {
		if( !IsCharAlphaNumericA( *pszWalk ) )
		    continue;
	    }

	    // Compare the needle with the current position in the haystack.

	    if( StrCaselessCompareCount( pszWalk, pszNeedle, cchNeedle ) == 0 )
		return pszWalk - pszHaystack;

	    // Find the end of this word.

	    if( fAlphaNum )
	    {
		while( IsCharAlphaNumericA( *pszWalk ) && *( pszWalk + 1 ) )
		    pszWalk++;
	    }
	}

	return -1;
}

#ifdef UNICODE
int
StrLookup( WCHAR const* const pszNeedle, size_t const cchNeedle, WCHAR const* const pszHaystack )
{
	Assert( FImplies( cchNeedle, pszNeedle ) );
	Assert( pszHaystack );

	if( !cchNeedle || !*pszHaystack )
	    return -1;

	// If the needle is alphanumeric then we'll look for matches starting
	// at a word boundary.

	bool const fAlphaNum = !!IsCharAlphaNumericW( *pszNeedle );

	for( WCHAR const* pszWalk = pszHaystack; *pszWalk; pszWalk++ )
	{
	    // Find the beginning of the next word.

	    if( fAlphaNum )
	    {
		if( !IsCharAlphaNumericW( *pszWalk ) )
		    continue;
	    }

	    // Compare the needle with the current position in the haystack.

	    if( _wcsnicmp( pszWalk, pszNeedle, cchNeedle ) == 0 )
		return pszWalk - pszHaystack;

	    // Find the end of this word.

	    if( fAlphaNum )
	    {
		while( IsCharAlphaNumericW( *pszWalk ) && *( pszWalk + 1 ) )
		    pszWalk++;
	    }
	}

	return -1;
}
#endif

