/*
 * DbOpen.h - Database abstraction layer.
 */

#pragma once

// vim: set noet:			// Hard tabs.
// vim: set ts=8:			// Tab width 8.
// vim: set sw=4:			// Shift (indent) width 4.
// vim: set cino={0s:			// Braces flush with outer statement.
// vim: set cino=^1s:			// First indent 8 (add 1 shift width).

#if defined( _MSC_VER )
#pragma warning( disable: 4505 )
#endif

#include "str.h"
#include "sys.h"
#include "bittable.h"
#include "viewdefs.h"
#include "arrays.h"
#include "progress.h"

class Item;
class ItemQuickInfo;
class DbOpenHs2;

#ifdef noRecordSelected
CASSERT( noRecordSelected == unsigned( -1 ) );
#undef noRecordSelected
#undef noRecordSelectedID
#endif

#define noRecordSelected	unsigned( -1 )
#define noRecordSelectedID	unsigned( -1 )

#define E_TOOMANYCATEGORIES	SE_UNKNOWN
#define E_TOOMANYSTORES		SE_UNKNOWN

unsigned const c_cRecordsPerBatch = 4000;

/*
 * Category helpers.
 */

enum
{
	catAllCategories	= 0xff,
	catUnfiledCategory	= 0xfe,
	catEditCategories	= 0xff00,
	catMultiple		= 0xee00,

	c_cchMaxCategoryName	= 16,		// Includes null terminator.
	c_cMaxCategories	= 100,
};

inline bool
IsPhantomCategory( WORD const category )
{
	Assert( category != catMultiple );
	Assert( category != catEditCategories );

	return ( category == catAllCategories || category == catUnfiledCategory );
}

struct MultipleCategories
{
	void		SetAll( bool fMarked );
	unsigned	CountMarked( DbOpen const* pdb ) const;

	BYTE		m_Unfiled;
	BYTE		m_rg[c_cMaxCategories];
};

/*
 * DbSortOrder - Sort order details.
 */

typedef WORD SortType;

struct DbSortOrder
{
	bool		m_fCheckedLast;
	bool		m_rgfDescend[NUM_SORT_FIELDS];
	SortType	m_rgiSort[NUM_SORT_FIELDS];
};

CASSERT( sizeof( bool ) == 1 );
CASSERT( sizeof( DbSortOrder ) == 10 );

/*
 * DbFont.
 */

enum FontId
{
	FID_NORMAL,
	FID_BOLD,
	FID_LARGEBOLD,
	FID_NARROW,
	FID_TINY,
	FID_MAX
};

FontId FontIdFromPalmFont( BYTE idPalm );
BYTE PalmFontFromFontId( FontId idFont );

/*
 * Stores.
 */

enum
{
	c_cMaxStores		= 100,
	c_cchMaxStoreName	= 32,
};

struct MultipleStores
{
	void		SetAll( bool fMarked );
	unsigned	CountMarked( DbOpen* pdb ) const;

	BYTE		m_Unfiled;
	BYTE		m_rg[c_cMaxStores];
};

/*
 * DbOpen - Database abstraction.
 */

enum
{
	packNoSort		= 0x0001,	// Don't sort the record into place even if it appears to be out of place; this also implies that the AutoOrder must not be adjusted (as that could require sorting the item).
	packForceSort		= 0x0002,	// Force sorting the record into place (normally a record is only sorted into place if it appears to be out of place).
	packNoDelete		= 0x0004,	// Suppress the normal behavior where packing a record with no Description deletes the record.
	packNoCounts		= 0x0008,	// Don't update the bold store counts (USE WITH CAUTION!!).
	packNoUndo		= 0x0010,	// Don't add the record to the Undo stack.
	packInsertAtIndex	= 0x0020,	// Insert the record at the index (versus overwriting the existing record at the index).
	packUndoing		= 0x0040,	// This pack is part of an Undo operation.

	// The cesMoved and cesDeleted flags are mutually exclusive.  However,
	// cesFiltered can be combined with any other values, including 0.

	cesMoved		= 0x0001,	// PackRecord sorted the item to a new record number, which can be retrieved via GetRecordAfterPack.
	cesDeleted		= 0x0002,	// PackRecord deleted the item because it was empty (no description), and empty items are not allowed in the database.
	cesFiltered		= 0x0004,	// The item is now hidden by the view filter.
};

enum
{
	exportQuantity		= 0x0001,
	exportPrice		= 0x0002,
	exportNote		= 0x0004,
	exportTotal		= 0x0008,
	exportUnits		= 0x0010,
	exportAisle		= 0x0020,
	exportDash		= 0x0040,
};

enum FilterReason
{
	notFiltered		= 0x0000,
	filteredPrivate		= 0x0001,
	filteredLookup		= 0x0002,
	filteredState		= 0x0004,
	filteredCoupon		= 0x0008,
	filteredCategory	= 0x0010,
	filteredStore		= 0x0020,
	filteredDeleted		= 0x0040,
	filteredCorrupt		= 0x0080,
};

inline FilterReason operator|( FilterReason const a, FilterReason const b )
{
	return FilterReason( int( a ) | int( b ) );
}

inline FilterReason operator^( FilterReason const a, FilterReason const b )
{
	return FilterReason( int( a ) ^ int( b ) );
}

class DbOpenSimple : public Object
{
    public:
				DbOpenSimple() : m_privRecordVisualMode( s_privDefaultRecordVisualMode ) {}

	static PrivateRecordVisualMode GetDefaultPrivateRecordVisualMode() { return s_privDefaultRecordVisualMode; }
	static void		SetDefaultPrivateRecordVisualMode( PrivateRecordVisualMode priv ) { s_privDefaultRecordVisualMode = priv; }

	virtual LPCTSTR		Name() const = 0;
	virtual bool		HasPrefix() const = 0;
	virtual LPCTSTR		FileName() const = 0;
	virtual unsigned	Count() const = 0;

	// Read records.

	virtual bool		SetLookup( StrT const* psLookup, bool fBroad ) = 0;
	virtual LPCTSTR		GetLookup() const = 0;
	virtual bool		IsLookupBroad() const = 0;

	virtual bool		SetFilterReasonCompleteness( bool fComplete ) = 0;
	virtual FilterReason	IsRecordFiltered( unsigned record ) const = 0;

	virtual Item*		UnPackRecord( unsigned record ) const = 0;
	virtual bool		GetItemQuickInfo( unsigned record, ItemQuickInfo* pqi ) const = 0;

	virtual bool		GetRecordUniqueId( unsigned record, DWORD* pdwUniqueId ) const = 0;
	virtual bool		FindRecordByUniqueId( DWORD uniqueId, unsigned* precord ) const = 0;

	PrivateRecordVisualMode	GetPrivateRecordVisualMode() const { return m_privRecordVisualMode; }
	virtual void		SetPrivateRecordVisualMode( PrivateRecordVisualMode priv );

    private:
	PrivateRecordVisualMode	m_privRecordVisualMode;

	static PrivateRecordVisualMode s_privDefaultRecordVisualMode;
};

class DbGroupTransaction : public Object
{
    public:
				DbGroupTransaction() : m_fActive( true ) {}
				~DbGroupTransaction();

	virtual SysError	End() = 0;

    protected:
	bool			IsActive() const { return m_fActive; }
	void			MarkEnded() { Assert( m_fActive ); m_fActive = false; }

    private:
	bool			m_fActive;
};

class DbOpen : public DbOpenSimple
{
	typedef DbOpenSimple base;

	enum
	{
	    numUneditableCategories	= 1,
	};

    public:

	static LPCTSTR		UnfiledString();
	static LPCTSTR		AllString();
	static LPCTSTR		AllStoresString();
	static LPCTSTR		DefaultCurrencyString();
	static void		SetLocalizedStrings( LPCTSTR pszUnfiled, LPCTSTR pszAll, LPCTSTR pszAllStores, LPCTSTR pszDefaultCurrency );

    public:
				DbOpen();
				~DbOpen();

	virtual void		Close() = 0;

	virtual bool		CanFlush( bool *pfFlushSupported = 0 ) = 0;
	virtual SysError	Flush() = 0;

	virtual void		OmitPrefix() = 0;
	virtual SysError	SetName( LPCTSTR pszName ) = 0;

	Override( base, void, SetPrivateRecordVisualMode, ( PrivateRecordVisualMode priv ) );

	// Filtering.

	Override( base, bool, SetLookup, ( StrT const* psLookup, bool fBroad ) );
	OverrideConst( base, LPCTSTR, GetLookup, () );
	OverrideConst( base, bool, IsLookupBroad, () );
	Override( base, bool, SetFilterReasonCompleteness, ( bool fComplete ) );

	virtual bool		SetView( WORD view ) = 0;
	virtual void		SetCategory( WORD category ) = 0;
	virtual void		SetStore( BYTE idStore ) = 0;
	virtual void		SetHideChecked( bool fHideChecked ) = 0;
	virtual void		SetCouponsOnly( bool fCouponsOnly ) = 0;

	virtual WORD		GetView() const = 0;
	virtual WORD		GetCategory() const = 0;
	virtual BYTE		GetStore() const = 0;
	virtual bool		GetHideChecked() const = 0;
	virtual bool		GetCouponsOnly() const = 0;

	virtual void		SetMultipleCategories( MultipleCategories const* pmulti ) = 0;
	virtual bool		GetMultipleCategories( MultipleCategories* pmulti ) const = 0;

	virtual void		SetMultipleStores( MultipleStores const* pmulti ) = 0;
	virtual bool		GetMultipleStores( MultipleStores* pmulti ) const = 0;

	// Write records.

	virtual SysError	PackRecord( unsigned record, Item** ppitem, WORD wFlags, WORD* pces ) = 0;
	virtual SysError	DeleteRecord( unsigned record, bool fCanUndo ) = 0;

	virtual int		GetChangeCounter() const = 0;
	virtual unsigned	GetRecordsMaximum() const = 0;

	// Raw record bytes (for diagnostic purposes).

	virtual SysError	GetRawRecordBytes( unsigned record, BYTE** ppbOut, unsigned* pcb ) const = 0;

	// Undo.

	virtual bool		CanUndo() = 0;
	virtual SysError	Undo( unsigned* precord, bool* pfDidUndo = 0 ) = 0;
	virtual void		ClearUndo() = 0;

	// Sorting.

	virtual bool		IsSortedByColumn( ColumnIdentifier column ) const = 0;

	virtual bool		SetSortOneOrder( bool fSortOneOrder ) = 0;
	virtual bool		SetSortOrderInfo( int index, const DbSortOrder* pso ) = 0;

	virtual bool		GetSortOneOrder() const = 0;
	virtual const DbSortOrder* GetSortOrderInfo( int index ) const = 0;

	virtual SysError	SortRecord( unsigned record, unsigned* precordTo = 0 ) = 0;
	virtual SysError	SortDatabase() = 0;

	bool			IsSortNeededBasedOnStoreChange( BYTE const newStore ) const;

	virtual bool		SetCanLearnOrder( int index, bool f ) = 0;
	virtual bool		CanLearnOrder( int index ) const = 0;
	virtual bool		IsLearningOrder() const = 0;
	virtual bool		ActivateLearningOrder( bool fActivate ) = 0;
	bool			ActivateLearningOrderHelper( bool fActivate );
	void			AdjustAutoOrder( Item* pitem ) const;

	// Columns.

	virtual void		SetColumnWidth( ColumnIdentifier column, int cx ) = 0;
	virtual int		GetColumnWidth( ColumnIdentifier column ) const = 0;

	// Categories.

	virtual unsigned	GetCategoriesCount() const = 0;
	virtual unsigned	GetCategoriesMaximum() const = 0;
	virtual LPCTSTR		GetCategoryNameById( WORD category ) const = 0;
	virtual LPCTSTR		GetCategoryNameByIndex( int index ) const = 0;
	virtual WORD		GetCategoryIdByName( LPCTSTR pszCategory ) const = 0;
	virtual WORD		GetCategoryIdByIndex( int index ) const = 0;
	virtual int		GetCategoryIndexById( WORD category ) const = 0;
	virtual WORD		GetCategoryIcon( WORD category ) const = 0;

	virtual bool		CategoryExistsById( WORD category ) = 0;
	virtual bool		CategoryHasRecords( WORD category ) = 0;
	virtual bool		CategoryNew( LPCTSTR pszCategory, WORD icon ) = 0;
	virtual bool		CategoryMerge( WORD oldCategory, WORD newCategory ) = 0;
	virtual bool		CategoryRename( WORD category, LPCTSTR pszCategory ) = 0;
	virtual bool		CategoryDelete( WORD category ) = 0;
	virtual bool		SetCategoryIcon( WORD category, WORD icon ) = 0;

	// Stores.

	virtual unsigned	GetStoresCount() const = 0;
	virtual unsigned	GetStoresMaximum() const = 0;
	virtual LPCTSTR		GetStoreNameById( BYTE idStore ) const = 0;
	virtual LPCTSTR		GetStoreNameByIndex( int index ) const = 0;
	virtual BYTE		GetStoreIdByName( LPCTSTR pszStore ) const = 0;
	virtual BYTE		GetStoreIdByIndex( int index ) const = 0;
	virtual int		GetStoreIndexById( BYTE idStore ) const = 0;

	virtual unsigned	GetStoreNeededItemsCount( BYTE idStore ) const = 0;

	virtual bool		StoreExistsById( BYTE idStore ) = 0;
	virtual bool		StoreHasRecords( BYTE idStore ) = 0;
	virtual bool		StoreNew( LPCTSTR pszStore ) = 0;
	virtual bool		StoreNewWithId( LPCTSTR pszStore, BYTE idStore ) = 0;
	virtual bool		StoreMerge( BYTE oldStore, BYTE newStore ) = 0;
	virtual bool		StoreRename( BYTE idStore, LPCTSTR pszStore ) = 0;
	virtual bool		StoreDelete( BYTE idStore ) = 0;

	virtual void		ResetStoreItemCounts( bool fProgress ) = 0;

	// Settings.

	virtual void		SetRoundToWholeNumbers( bool fRoundWholeNumbers ) = 0;
	virtual void		SetCurrency( LPCTSTR pszCurrency ) = 0;
	virtual void		SetTaxRate( WORD wTaxRate ) = 0;
	virtual void		SetTaxRate2( WORD wTaxRate2 ) = 0;
	virtual void		SetTwoTaxes( bool fTwoTaxes ) = 0;
	virtual void		SetPerStorePrices( bool fPerStorePrices ) = 0;
	virtual void		SetAutoQuantity( bool fAutoQuantity ) = 0;
	virtual void		SetAutoDelete( bool fAutoDelete ) = 0;
	virtual void		SetRecordCreationDate( bool f ) = 0;
	virtual void		SetRecordCompletionDate( bool f ) = 0;
	virtual void		SetTaxable( bool f ) = 0;
	virtual void		SetDefaultPriority( BYTE prio ) = 0;
	virtual void		SetRowBackgroundColor( COLORREF crBackground ) = 0;
	virtual void		SetFont( FontId id ) = 0;
	virtual void		SetMaxLinesPerItem( int cLines ) = 0;
	virtual void		SetIcon( WORD icon ) = 0;

	virtual bool		GetRoundToWholeNumbers() const = 0;
	virtual LPCTSTR		GetCurrency() const = 0;
	virtual WORD		GetTaxRate() const = 0;
	virtual WORD		GetTaxRate2() const = 0;
	virtual bool		GetTwoTaxes() const = 0;
	virtual bool		GetPerStorePrices() const = 0;
	virtual bool		GetAutoQuantity() const = 0;
	virtual bool		GetAutoDelete() const = 0;
	virtual bool		GetRecordCreationDate() const = 0;
	virtual bool		GetRecordCompletionDate() const = 0;
	virtual bool		GetTaxable() const = 0;
	virtual BYTE		GetDefaultPriority() const = 0;
	virtual COLORREF	GetRowBackgroundColor() const = 0;
	virtual FontId		GetFont() const = 0;
	virtual int		GetMaxLinesPerItem() const = 0;
	virtual WORD		GetIcon() const = 0;

	virtual DWORD		GetNextAlarm() const = 0;

	// Export.

	virtual void		SetExportFlags( WORD flags ) = 0;
	virtual WORD		GetExportFlags() const = 0;

	// Maintenance.

	virtual bool		IsCauterized() const = 0;

	//$ todo: (chrisant) Compact the database file.

	// Debug.

#ifdef DEBUG
	virtual void		AssertNoRecordLocks( LPCTSTR pszContext ) = 0;
	virtual WORD		GetDbVersion() = 0;
	virtual bool		HasRollbackFlag() = 0;
#endif

    public:

	bool			FViewNeed( WORD view ) const { return ( view == VIEW_NEED || view == VIEW_ALLNOTPURCHASED ); }
	bool			FAltSort() const { return !GetSortOneOrder() && !FViewNeed( GetView() ); }

	unsigned		CountInView() { return m_pbits->CountInView(); }
	bool			InView( unsigned record ) { return m_pbits->InView( record ); }
	unsigned		IndexInView( unsigned record ) { return m_pbits->IndexInView( record ); }
	bool			FindNextRecord( unsigned* precord, bool fNext ) { return m_pbits->FindNextRecord( precord, fNext ); }
	unsigned		GetRecordAfterPack() { return m_recAfterPack; }

	SysError		ConvertFrom( DbOpen const* pdb );
	SysError		MakeTransactionGroup( SPNew<DbGroupTransaction>& spXact );

	SysError		Checkout( bool fProgress );
	SysError		CheckoutRecord( unsigned record, WORD* pces, unsigned* precordAfterPack );

    protected:
	virtual SysError	NewDbGroupTransaction( DbGroupTransaction** ppOut ) = 0;

	bool			RenumberAutoOrder( bool fForce = false );
	DWORD			GetHighestAutoOrder( bool fOnlyCheckedOffItems ) const;
	DWORD			GetSuccessorOfAutoOrder( DWORD ulFloor, DWORD ulCurrent ) const;
	DWORD			FindAutoOrder( DWORD ul );

    protected:
	BitTable*		m_pbits;
	unsigned		m_recAfterPack;

	StrT			m_sLookup;
	StrA			m_sLookupA;
	bool			m_fFilteringDisabled;
	bool			m_fCompleteReasons;

    private:
	static TCHAR		s_szUnfiled[c_cchMaxCategoryName];
	static TCHAR		s_szAll[c_cchMaxCategoryName];
	static TCHAR		s_szAllStores[c_cchMaxStoreName];
	static TCHAR		s_szDefaultCurrency[8];
};

class DbQuickOpen : public Object
{
	friend DbQuickOpen* QuickOpenHs2Database( LPCTSTR pszFile );
	friend DbQuickOpen* QuickOpenHs3Database( LPCTSTR pszFile );

    public:
				DbQuickOpen();
				~DbQuickOpen();

	bool			IsEqual( DbQuickOpen const* pqo ) const;

	const StrT*		Name() const { return &m_sName; }

	unsigned		Count() const { return m_cItems; }
	unsigned		CountNeeded() const { return m_cNeeded; }

	virtual SysError	Rename( TCHAR const* pszFile, StrT const* ps ) = 0;

#ifdef DEBUG
	const StrT*		File() const { return &m_sFile; }
	const StrT*		Path() const { return &m_sPath; }
	WORD			GetDbVersion() const { return m_version; }
	bool			HasRollbackFlag() const { return m_fRollbackFlag; }
#endif

    private:
	StrT			m_sName;
	unsigned		m_cItems;
	unsigned		m_cNeeded;

#ifdef DEBUG
	StrT			m_sFile;
	StrT			m_sPath;
	WORD			m_version;
	bool			m_fRollbackFlag;
#endif
};

enum DbFormatType
{
	DBFT_UNKNOWN,
	DBFT_PALM,
	DBFT_MOBILE,
};

DbFormatType DbFormatTypeFromExtension( LPCTSTR pszFile );

SysError MakeNewDatabaseFileName( LPCTSTR pszDir, LPCTSTR pszName, DbFormatType format, StrT* psFile );
SysError NewDatabase( LPCTSTR pszFile, LPCTSTR pszDisplayName, DbOpen** ppdb );
SysError OpenDatabase( LPCTSTR pszFile, DbOpen** ppdb );

DbOpenSimple* SimpleOpenDatabase( LPCTSTR pszFile );
DbQuickOpen* QuickOpenDatabase( LPCTSTR pszFile );

#ifdef DEBUG
SysError VerifyIntegrity( LPCTSTR pszFile );
#endif

#ifdef UNICODE
int StrLookup( WCHAR const* pszNeedle, size_t cchNeedle, WCHAR const* const pszHaystack );
inline int StrLookup( StrT const* psNeedle, TCHAR const* const pszHaystack ) { return StrLookup( psNeedle->Text(), psNeedle->Length(), pszHaystack ); }
#endif

int StrLookup( char const* pszNeedle, size_t cchNeedle, char const* const pszHaystack );
inline int StrLookup( StrA const* psNeedle, char const* const pszHaystack ) { return StrLookup( psNeedle->Text(), psNeedle->Length(), pszHaystack ); }
inline int StrLookup( StrT const* psNeedle, StrT const* const psHaystack ) { return StrLookup( psNeedle->Text(), psNeedle->Length(), psHaystack->Text() ); }

