/*
 * IoText.cpp - Text output interface for databases.
 */

// vim: set noet:			// Hard tabs.
// vim: set ts=8:			// Tab width 8.
// vim: set sw=4:			// Shift (indent) width 4.
// vim: set cino={0s:			// Braces flush with outer statement.
// vim: set cino=^1s:			// First indent 8 (add 1 shift width).

#include "precomp.h"

#ifdef OS_PALM
#error The converter is not designed for the Palm OS.
#endif

#include "std.h"
#include "str.h"
#include "sys.h"
#include "converter.h"
#include "storelist.h"
#include "number.h"
#include "item.h"
#include "dbopen.h"
#include "_types.h"

static void
DumpFlags( FILE* const pfile, WORD const flags )
{
	fprintf( pfile, "flags         =" );

	switch( flags & ( flagNeed|flagPurchased ) )
	{
	case flagNeed:			fprintf( pfile, " NEEDED" ); break;
	case flagPurchased:		fprintf( pfile, " CHECKEDOFF" ); break;
	case flagNeed|flagPurchased:	fprintf( pfile, " DEFERRED" ); break;
	case 0:				break;
	default:			Panic(); break;
	}

	if( flags & flagTaxable )
	    fprintf( pfile, " TAXABLE" );
	if( flags & flagTaxable2 )
	    fprintf( pfile, " TAXABLE2" );

	if( flags & flagCoupon )
	    fprintf( pfile, " COUPON" );

	if( flags & flagAutoDelete )
	    fprintf( pfile, " AUTODELETE" );

	if( flags & flagPrivate )
	    fprintf( pfile, " PRIVATE" );

	if( flags & flagHasAlarm )
	    fprintf( pfile, " ALARM" );

	fprintf( pfile, "\n" );
}

static void
EllipsifyStr( StrT& s, unsigned const cchMax )
{
	if( s.Length() <= cchMax )
	    return;

	if( s.Length() < 4 )
	    return;

	StrT tmp;

	tmp.Set( s );
	tmp.SetLength( cchMax - 3 );
	tmp.Append( TEXT( "..." ) );

	s.Set( tmp );
}

class OutText : public OutputInterface
{
    public:
			OutText();
			~OutText();

	DeclareOutputInterfaceMembers( IMPL );

    private:
	void		Close();

    private:
	FILE*		m_pfile;
	FILE*		m_pfileClose;
	int		m_cCategories;
	int		m_cStores;

	StrT		m_sName;
	bool		m_fForceOverwrite;
};

OutText::OutText()
{
	m_pfile = stdout;
	m_pfileClose = 0;
	m_cCategories = 0;
	m_cStores = 0;
	m_fForceOverwrite = false;
}

OutText::~OutText()
{
	Close();
}

void
OutText::Close()
{
	if( m_pfileClose )
	    fclose( m_pfileClose );
	m_pfileClose = 0;
	m_pfile = 0;
}

int
OutText::SetFileName( TCHAR const* const pszFile, bool const fForceOverwrite )
{
	m_sName.Set( pszFile );
	m_fForceOverwrite = fForceOverwrite;
	return 0;
}

int
OutText::SetFriendlyNames( bool const fUseFriendlyNames )
{
	SUPPRESS_UNUSED_VAR( fUseFriendlyNames );
	return 0;
}

int
OutText::Begin( DbOpen const* const pdb )
{
	SUPPRESS_UNUSED_VAR( pdb );

	if( m_sName.Length() )
	{
	    StrA name;
	    name.SetT( m_sName );

	    if( FileExists( m_sName.Text() ) )
	    {
		if( m_fForceOverwrite )
		{
		    __DeleteFile( m_sName.Text() );
		}
		else
		{
		    fprintf( stderr, "Output file '%s' already exists.\n", name.Text() );
		    return -1;
		}
	    }

	    if( fopen_s( &m_pfileClose, name.Text(), "w+" ) || !m_pfileClose )
	    {
		fprintf( stderr, "Unable to open output file.\n" );
		return -1;
	    }

	    m_pfile = m_pfileClose;
	}

	return 0;
}

int
OutText::OutputCategory( DbOpen const* const pdb, TCHAR const* const pszCategory, WORD const id )
{
	if( pszCategory )
	{
	    WORD const icon = pdb->GetCategoryIcon( id );

	    if( !m_cCategories )
		fprintf( m_pfile, "\nCATEGORIES\n" );

	    fprintf( m_pfile, "category%-6u= %-3u (%ls)", m_cCategories, id, pszCategory );
	    if( icon )
		fprintf( m_pfile, " icon=%u", icon );
	    fprintf( m_pfile, "\n" );

	    m_cCategories++;
	}

	return 0;
}

static void
OutputAppInfoHeader( bool& fDidHeader, FILE* const pfile )
{
	if( !fDidHeader )
	{
	    fDidHeader = true;
	    fprintf( pfile, "\nAPPINFO\n" );
	}
}

int
OutText::OutputPreferences( DbOpen const* const pdb )
{
	bool fDidHeader = false;

	UInt32 const alarm = pdb->GetNextAlarm();

	if( alarm )
	{
	    OutputAppInfoHeader( fDidHeader, m_pfile );

	    if( alarm >= c_UnixToPalm )
	    {
		time_t const time = alarm - c_UnixToPalm;
		struct tm tmAlarm;

		if( !_gmtime32_s( &tmAlarm, &time ) )
		{
		    fprintf( m_pfile, "next alarm    = %04u/%02u/%02u %02u:%02u:%02u\n",
			     tmAlarm.tm_year + 1900, tmAlarm.tm_mon + 1, tmAlarm.tm_mday,
			     tmAlarm.tm_hour, tmAlarm.tm_min, tmAlarm.tm_sec );
		}
		else
		{
		    fprintf( m_pfile, "next alarm    = <invalid value>\n" );
		}
	    }
	    else
	    {
		fprintf( m_pfile, "next alarm    = <before 1970>\n" );
	    }

	    fprintf( m_pfile, "       raw    = %u\n", alarm );
	}

	WORD const icon = pdb->GetIcon();

	if( icon )
	{
	    OutputAppInfoHeader( fDidHeader, m_pfile );
	    fprintf( m_pfile, "icon          = %u\n", icon );
	}

	return 0;
}

int
OutText::OutputStore( DbOpen const* const pdb, TCHAR const* const pszStore, BYTE const id )
{
	SUPPRESS_UNUSED_VAR( pdb );

	if( pszStore )
	{
	    if( !m_cStores )
		fprintf( m_pfile, "\nSTORES\n" );

	    fprintf( m_pfile, "store%-9u= %-3u (%ls)\n", m_cStores, id, pszStore );

	    m_cStores++;
	}

	return 0;
}

int
OutText::OutputItem( DbOpen const* const pdb, Item const* const pitem )
{
	if( pitem )
	{
	    fprintf( m_pfile, "\nITEMDATA\n" );

	    DumpFlags( m_pfile, pitem->m_flags );

	    fprintf( m_pfile, "prio          = %d\n", pitem->GetPriority() + 1 );

	    if( pitem->GetDescription()->Length() )
		fprintf( m_pfile, "desc          = '%ls'\n", pitem->GetDescription()->Text() );
	    if( pitem->GetCustomText()->Length() )
		fprintf( m_pfile, "custom        = '%ls'\n", pitem->GetCustomText()->Text() );

	    fprintf( m_pfile, "category      = '%ls'\n", pdb->GetCategoryNameById( pitem->GetCategory() ) );

	    if( pitem->GetQuantity() )
		fprintf( m_pfile, "quantity      = %ls\n", LocalizeNumber( pitem->GetQuantity(), lnTrimRightZeros, QTY_PRECISION, 0 ) );
	    if( pitem->GetUnits()->Length() )
		fprintf( m_pfile, "units         = %ls\n", pitem->GetUnits()->Text() );
	    if( pitem->GetPrice() )
		fprintf( m_pfile, "price         = %ls\n", LocalizeNumber( pitem->GetPrice(), lnCurrencySymbol|lnTrimRightZeros2, PRICE_PRECISION, 0 ) );
	    if( pitem->GetAisle()->Length() )
		fprintf( m_pfile, "aisle         = %ls\n", pitem->GetAisle()->Text() );

	    if( !pitem->FNoDate() )
	    {
		SYSTEMTIME const* const psystime = pitem->GetDate();

		fprintf( m_pfile, "date          = %04u/%02u/%02u\n", psystime->wYear, psystime->wMonth, psystime->wDay );
		fprintf( m_pfile, "rawdate       = 0x%04x\n", *reinterpret_cast<WORD const*>( &pitem->m_date ) );
	    }

	    if( pitem->m_stores.Count() )
	    {
		WORD c = pitem->m_stores.Count();
		Store const* p = pitem->m_stores.Array();
		fprintf( m_pfile, "stores        = %u\n", c );
		while( c-- )
		{
		    fprintf( m_pfile, "                %-3u (%ls)\n", *p, pdb->GetStoreNameById( *p ) );
		    p++;
		}
	    }
	    if( pitem->m_perstore.Count() )
	    {
		StrT s1;
		StrA s2;
		WORD c = pitem->m_perstore.Count();
		PerStore const* p = pitem->m_perstore.Array();
		fprintf( m_pfile, "perstore      = %u\n", c );
		while( c-- )
		{
		    s1.Set( pdb->GetStoreNameById( p->m_idStore ) );
		    EllipsifyStr( s1, 16 );
		    s2.Clear();
		    s2.Printf( "(%ls);", s1.Text() );

		    fprintf( m_pfile, "                %-3u %-19s  aisle = %ls;  price = %ls\n",
			     p->m_idStore, s2.Text(), p->m_szAisle,
			     LocalizeNumber( p->m_lPrice, lnCurrencySymbol|lnTrimRightZeros2, PRICE_PRECISION, 0 ) );
		    p++;
		}
	    }
	    if( pitem->CountAutoOrders() )
	    {
		StrT s1;
		StrA s2;
		fprintf( m_pfile, "autoorder     = %u\n", pitem->CountAutoOrders() );
		for( WORD ii = 0; ii < pitem->CountAutoOrders(); ii++ )
		{
		    AutoOrder const* const p = pitem->GetAutoOrderByIndex( ii );

		    if( !p->m_ulOrder )	// Skip entries that don't have any order information (for example the phantom store entry if it has never been filled in).
			continue;

		    s1.Set( pdb->GetStoreNameById( p->m_idStore ) );
		    EllipsifyStr( s1, 16 );
		    s2.Clear();
		    s2.Printf( "(%ls);", s1.Text() );

		    fprintf( m_pfile, "                %-3u %-19s  order = %u\n",
			     p->m_idStore, s2.Text(), p->m_ulOrder );
		}
	    }

	    if( pitem->GetNote()->Length() )
		fprintf( m_pfile, "note          = '%ls'\n", pitem->GetNote()->Text() );

	    if( pitem->GetAlarm() )
	    {
		if( pitem->GetAlarm() >= c_UnixToPalm )
		{
		    time_t const time = pitem->GetAlarm() - c_UnixToPalm;
		    struct tm tmAlarm;

		    if( !_gmtime32_s( &tmAlarm, &time ) )
		    {
			fprintf( m_pfile, "alarm         = %04u/%02u/%02u %02u:%02u:%02u\n",
				 tmAlarm.tm_year + 1900, tmAlarm.tm_mon + 1, tmAlarm.tm_mday,
				 tmAlarm.tm_hour, tmAlarm.tm_min, tmAlarm.tm_sec );
		    }
		    else
		    {
			fprintf( m_pfile, "alarm         = <invalid value>\n" );
		    }
		}
		else
		{
		    fprintf( m_pfile, "alarm         = <before 1970>\n" );
		}

		fprintf( m_pfile, "rawalarm      = %lu\n", ULONG( pitem->GetAlarm() ) );
	    }

	    if( pitem->GetAlarmMidi() )
		fprintf( m_pfile, "alarmmidi     = %lu\n", ULONG( pitem->GetAlarmMidi() ) );

	    if( pitem->GetRawIcon() )
		fprintf( m_pfile, "icon          = %u\n", ULONG( pitem->GetRawIcon() ) );

	    fprintf( m_pfile, "entry         = %u\n", pitem->GetEntryOrder() );
	}

	return 0;
}

int
OutText::Finished( DbOpen const* const pdb )
{
	SUPPRESS_UNUSED_VAR( pdb );

	Close();
	return 0;
}

OutputInterface*
NewOutputText()
{
	return new OutText;
}

