/*
 * IoXml.cpp - XML output interface for databases.
 */

// vim: set noet:			// Hard tabs.
// vim: set ts=8:			// Tab width 8.
// vim: set sw=4:			// Shift (indent) width 4.
// vim: set cino={0s:			// Braces flush with outer statement.
// vim: set cino=^1s:			// First indent 8 (add 1 shift width).

#include "precomp.h"

#ifdef OS_PALM
#error The converter is not designed for the Palm OS.
#endif

#include "std.h"
#include "str.h"
#include "sys.h"
#include "converter.h"
#include "storelist.h"
#include "number.h"
#include "item.h"
#include "dbopen.h"
#include "_types.h"

class OutXML : public OutputInterface
{
    public:
			OutXML();
			~OutXML();

	DeclareOutputInterfaceMembers( IMPL );

    private:
	void		Close();
	void		TransformOutput( TCHAR const* psz );

    private:
	FILE*		m_pfile;
	FILE*		m_pfileClose;
	int		m_cCategories;
	int		m_cStores;
	int		m_cItems;

	StrT		m_sName;
	bool		m_fForceOverwrite;
};

OutXML::OutXML()
{
	m_pfile = stdout;
	m_pfileClose = 0;
	m_cCategories = 0;
	m_cStores = 0;
	m_cItems = 0;
	m_fForceOverwrite = false;
}

OutXML::~OutXML()
{
	Close();
}

void
OutXML::Close()
{
	if( m_pfileClose )
	    fclose( m_pfileClose );
	m_pfileClose = 0;
	m_pfile = 0;
}

static bool
IsChrTransformable( TCHAR const ch )
{
	return ( ch == '<' ||
		 ch == '>' ||
		 ch == '\"' ||
		 ch == '&' ||
		 ch == '\'' );
}

static char const*
GetChrTransformation( TCHAR const ch )
{
	switch( ch )
	{
	case '<':	return "&lt;";
	case '>':	return "&gt;";
	case '\"':	return "&quot;";
	case '&':	return "&amp;";
	case '\'':	return "&apos;";
	}

	Panic();
	return 0;
}

void
OutXML::TransformOutput( TCHAR const* psz )
{
	StrA s;
	TCHAR const* pszWalk;

	while( *psz )
	{
	    pszWalk = psz;

	    while( *pszWalk && !IsChrTransformable( *pszWalk ) )
		pszWalk++;

	    s.SetT( psz, pszWalk - psz );
	    fwrite( s.Text(), s.Length(), 1, m_pfile );
	    psz = pszWalk;

	    if( !*pszWalk )
		break;

	    fprintf( m_pfile, GetChrTransformation( *pszWalk ) );

	    ++psz;
	}
}

int
OutXML::SetFileName( TCHAR const* const pszFile, bool const fForceOverwrite )
{
	m_sName.Set( pszFile );
	m_fForceOverwrite = fForceOverwrite;
	return 0;
}

int
OutXML::SetFriendlyNames( bool const fUseFriendlyNames )
{
	SUPPRESS_UNUSED_VAR( fUseFriendlyNames );
	return 0;
}

int
OutXML::Begin( DbOpen const* const pdb )
{
	SUPPRESS_UNUSED_VAR( pdb );

	if( m_sName.Length() )
	{
	    StrA name;
	    name.SetT( m_sName );

	    if( FileExists( m_sName.Text() ) )
	    {
		if( m_fForceOverwrite )
		{
		    __DeleteFile( m_sName.Text() );
		}
		else
		{
		    fprintf( stderr, "Output file '%s' already exists.\n", name.Text() );
		    return -1;
		}
	    }

	    if( fopen_s( &m_pfileClose, name.Text(), "w+" ) || !m_pfileClose )
	    {
		fprintf( stderr, "Unable to open output file.\n" );
		return -1;
	    }

	    m_pfile = m_pfileClose;
	}

	fprintf( m_pfile, "<Database>\n" );

	return 0;
}

int
OutXML::OutputCategory( DbOpen const* const pdb, TCHAR const* const pszCategory, WORD const id )
{
	if( !m_cCategories )
	    fprintf( m_pfile, "<Categories>\n" );

	if( pszCategory )
	{
	    WORD const icon = pdb->GetCategoryIcon( id );
	    StrA sIcon;

	    if( icon )
		sIcon.Printf( " ICON=\"%u\"", icon );

	    char const* const pszUnfiled = ( id == catUnfiledCategory ) ? " UNFILED=\"YES\"" : "";

	    fprintf( m_pfile, "  <Category%s%s>", sIcon.Text(), pszUnfiled );
	    TransformOutput( pszCategory );
	    fprintf( m_pfile, "</Category>\n" );

	    m_cCategories++;
	}
	else
	{
	    fprintf( m_pfile, "</Categories>\n" );
	}

	return 0;
}

int
OutXML::OutputPreferences( DbOpen const* const pdb )
{
	WORD const icon = pdb->GetIcon();

	if( icon )
	    fprintf( m_pfile, "<Icon>%u</Icon>\n", icon );

	return 0;
}

int
OutXML::OutputStore( DbOpen const* const pdb, TCHAR const* const pszStore, BYTE const id )
{
	SUPPRESS_UNUSED_VAR( pdb );
	SUPPRESS_UNUSED_VAR( id );

	if( !m_cStores )
	    fprintf( m_pfile, "<Stores>\n" );

	if( pszStore )
	{
	    fprintf( m_pfile, "  <Store>" );
	    TransformOutput( pszStore );
	    fprintf( m_pfile, "</Store>\n" );

	    m_cStores++;
	}
	else
	{
	    fprintf( m_pfile, "</Stores>\n" );
	}

	return 0;
}

int
OutXML::OutputItem( DbOpen const*  const pdb, Item const* const pitem )
{
	if( !m_cItems )
	    fprintf( m_pfile, "<Items>\n" );

	if( pitem )
	{
	    StrA s;

	    switch( pitem->m_flags & ( flagNeed|flagPurchased ) )
	    {
	    default:			break;
	    case flagNeed:		s.Append( " NEED=\"YES\"" ); break;
	    case flagPurchased:		s.Append( " HAVE=\"YES\"" ); break;
	    case flagNeed|flagPurchased: s.Append( " DEFERRED=\"YES\"" ); break;
	    }
	    if( pitem->m_flags & flagTaxable )
		s.Append( " TAX=\"YES\"" );
	    if( pitem->m_flags & flagTaxable2 )
		s.Append( " TAX2=\"YES\"" );
	    if( pitem->m_flags & flagCoupon )
		s.Append( " COUPON=\"YES\"" );
	    if( pitem->m_flags & flagAutoDelete )
		s.Append( " AUTODELETE=\"YES\"" );
	    if( pitem->m_flags & flagPrivate )
		s.Append( " PRIVATE=\"YES\"" );
	    if( pitem->m_flags & flagHasAlarm )
		s.Append( " ALARM=\"YES\"" );
	    fprintf( m_pfile, "  <Item%s>\n", s.Text() );

	    fprintf( m_pfile, "    <Priority>%u</Priority>\n", pitem->GetPriority() + 1 );

	    if( pitem->GetDescription()->Length() )
	    {
		fprintf( m_pfile, "    <Description>" );
		TransformOutput( pitem->GetDescription()->Text() );
		fprintf( m_pfile, "</Description>\n" );
	    }

	    if( pitem->GetCustomText()->Length() )
	    {
		fprintf( m_pfile, "    <CustomText>" );
		TransformOutput( pitem->GetCustomText()->Text() );
		fprintf( m_pfile, "</CustomText>\n" );
	    }

	    fprintf( m_pfile, "    <Category>" );
	    TransformOutput( pdb->GetCategoryNameById( pitem->GetCategory() ) );
	    fprintf( m_pfile, "</Category>\n" );

	    if( pitem->GetQuantity() )
	    {
		fprintf( m_pfile, "    <Quantity>" );
		TransformOutput( LocalizeNumber( pitem->GetQuantity(), lnTrimRightZeros, QTY_PRECISION, 0 ) );
		fprintf( m_pfile, "</Quantity>\n" );
	    }

	    if( pitem->GetUnits()->Length() )
	    {
		fprintf( m_pfile, "    <Unit>" );
		TransformOutput( pitem->GetUnits()->Text() );
		fprintf( m_pfile, "</Unit>\n" );
	    }

	    if( pitem->GetPrice() )
	    {
		fprintf( m_pfile, "    <Price>" );
		TransformOutput( LocalizeNumber( pitem->GetPrice(), lnTrimRightZeros2, PRICE_PRECISION, 0 ) );
		fprintf( m_pfile, "</Price>\n" );
	    }

	    if( pitem->GetAisle() )
	    {
		fprintf( m_pfile, "    <Aisle>" );
		TransformOutput( pitem->GetAisle()->Text() );
		fprintf( m_pfile, "</Aisle>\n" );
	    }

	    if( !pitem->FNoDate() )
	    {
		SYSTEMTIME const* const psystime = pitem->GetDate();

		fprintf( m_pfile, "    <Date>%04u/%02u/%02u</Date>\n", psystime->wYear, psystime->wMonth, psystime->wDay );
	    }

	    if( pitem->GetAlarm() )
	    {
		if( pitem->GetAlarm() >= c_UnixToPalm )
		{
		    time_t const time = pitem->GetAlarm() - c_UnixToPalm;
		    struct tm tmAlarm;

		    if( !_gmtime32_s( &tmAlarm, &time ) )
		    {
			fprintf( m_pfile, "    <Alarm>%04u/%02u/%02u %02u:%02u:%02u</Alarm>\n",
				 tmAlarm.tm_year + 1900, tmAlarm.tm_mon + 1, tmAlarm.tm_mday,
				 tmAlarm.tm_hour, tmAlarm.tm_min, tmAlarm.tm_sec );
		    }
		}
		else
		{
		    fprintf( m_pfile, "    <Alarm>Before 1970</Alarm>\n" );
		}
	    }

	    if( pitem->GetAlarmMidi() )
	    {
		fprintf( m_pfile, "    <AlarmMidi>%lu</AlarmMidi>\n", pitem->GetAlarmMidi() );
	    }

	    if( pitem->m_stores.Count() )
	    {
		WORD c = pitem->m_stores.Count();
		Store const* p = pitem->m_stores.Array();

		fprintf( m_pfile, "    <Stores>\n" );
		while( c-- )
		{
		    fprintf( m_pfile, "      <Store>" );
		    TransformOutput( pdb->GetStoreNameById( *p ) );
		    fprintf( m_pfile, "</Store>\n" );
		    p++;
		}
		fprintf( m_pfile, "    </Stores>\n" );
	    }
	    if( pitem->m_perstore.Count() )
	    {
		WORD c = pitem->m_perstore.Count();
		PerStore const* p = pitem->m_perstore.Array();

		fprintf( m_pfile, "    <PerStores>\n" );
		while( c-- )
		{
		    fprintf( m_pfile, "      <PerStore>\n" );

		    fprintf( m_pfile, "        <Store>" );
		    TransformOutput( pdb->GetStoreNameById( p->m_idStore ) );
		    fprintf( m_pfile, "</Store>\n" );

		    if( p->m_szAisle[0] )
		    {
			fprintf( m_pfile, "        <Aisle>" );
			TransformOutput( p->m_szAisle );
			fprintf( m_pfile, "</Aisle>\n" );
		    }

		    if( p->m_lPrice )
		    {
			fprintf( m_pfile, "        <Price>" );
			TransformOutput( LocalizeNumber( p->m_lPrice, lnTrimRightZeros2, PRICE_PRECISION, 0 ) );
			fprintf( m_pfile, "</Price>\n" );
		    }

		    fprintf( m_pfile, "      </PerStore>\n" );

		    p++;
		}

		fprintf( m_pfile, "    </PerStores>\n" );
	    }
	    if( pitem->CountAutoOrders() )
	    {
		fprintf( m_pfile, "    <AutoOrders>\n" );
		for( WORD ii = 0; ii < pitem->CountAutoOrders(); ii++ )
		{
		    AutoOrder const* const p = pitem->GetAutoOrderByIndex( ii );

		    if( !p->m_ulOrder )	// Skip entries that don't have any order information (for example the phantom store entry if it has never been filled in).
			continue;

		    fprintf( m_pfile, "      <AutoOrder>\n" );

		    fprintf( m_pfile, "        <Store>" );
		    TransformOutput( pdb->GetStoreNameById( p->m_idStore ) );
		    fprintf( m_pfile, "</Store>\n" );

		    fprintf( m_pfile, "        <Order>%u</Order>\n", p->m_ulOrder );

		    fprintf( m_pfile, "      </AutoOrder>\n" );
		}

		fprintf( m_pfile, "    </AutoOrders>\n" );
	    }

	    if( pitem->GetNote()->Length() )
	    {
		fprintf( m_pfile, "    <Note>" );
		TransformOutput( pitem->GetNote()->Text() );
		fprintf( m_pfile, "</Note>\n" );
	    }

	    if( pitem->GetRawIcon() )
	    {
		fprintf( m_pfile, "    <Icon>%u</Icon>\n", pitem->GetRawIcon() );
	    }

	    fprintf( m_pfile, "    <EntryOrder>%u</EntryOrder>\n", pitem->GetEntryOrder() );

	    fprintf( m_pfile, "  </Item>\n" );

	    m_cItems++;
	}
	else
	{
	    fprintf( m_pfile, "</Items>\n" );
	}

	return 0;
}

int
OutXML::Finished( DbOpen const* const pdb )
{
	SUPPRESS_UNUSED_VAR( pdb );

	if( m_cItems )
	    fprintf( m_pfile, "</Items>\n" );

	fprintf( m_pfile, "</Database>\n" );

	Close();
	return 0;
}

OutputInterface*
NewOutputXML()
{
	return new OutXML;
}

