/*
 * Item.cpp - Item object.
 */

// vim: set noet:			// Hard tabs.
// vim: set ts=8:			// Tab width 8.
// vim: set sw=4:			// Shift (indent) width 4.
// vim: set cino={0s:			// Braces flush with outer statement.
// vim: set cino=^1s:			// First indent 8 (add 1 shift width).

#include "precomp.h"

#include "str.h"
#include "dbopen.h"
#include "storelist.h"
#include "item.h"
#include "number.h"

/*
 * ItemQuickInfo class.
 */

void
ItemQuickInfo::Initialize( WORD const flags, bool const fDeleted )
{
	//$ review: (chrisant) Maybe this should just copy the flags and defer
	// the computation cost until actually needed.

	m_fDeleted = fDeleted;
	m_fAutoDelete = !!( flags & flagAutoDelete );
	m_fPrivate = !!( flags & flagPrivate );
	m_state = ItemStateFromFlags( flags );
}

/*
 * Item class.
 */

TCHAR const g_chPrioBase = '1';

unsigned const Item::offsetPriority = OffsetOf0( Item, m_prio );
unsigned const Item::offsetDescription = OffsetOf0( Item, m_sDescription );
unsigned const Item::offsetCustomText = OffsetOf0( Item, m_sCustomText );
unsigned const Item::offsetNote = OffsetOf0( Item, m_sNote );
unsigned const Item::offsetQuantity = OffsetOf0( Item, m_lQuantity );
unsigned const Item::offsetUnits = OffsetOf0( Item, m_sUnits );
unsigned const Item::offsetAisle = OffsetOf0( Item, m_sAisle );
unsigned const Item::offsetPrice = OffsetOf0( Item, m_lPrice );
unsigned const Item::offsetCategory = OffsetOf0( Item, m_category );
unsigned const Item::offsetDate = OffsetOf0( Item, m_date );
unsigned const Item::offsetEntryOrder = OffsetOf0( Item, m_ulEntryOrder );
unsigned const Item::offsetStores = OffsetOf0( Item, m_stores );
unsigned const Item::offsetPerStore = OffsetOf0( Item, m_perstore );
unsigned const Item::offsetIcon = OffsetOf0( Item, m_icon );
unsigned const Item::offsetAlarm = OffsetOf0( Item, m_alarm );
unsigned const Item::offsetAlarmMidi = OffsetOf0( Item, m_uidAlarmMidi );
unsigned const Item::offsetAutoOrder = OffsetOf0( Item, m_autoorder );

Item::Item()
{
	m_category = catUnfiledCategory;
	*reinterpret_cast<WORD*>( &m_date ) = dateNone;
	m_dirtySort = dirtyALL;

#ifdef DEBUG
	AssertClear();
#endif
}

Item::~Item()
{
}

void
Item::Clear()
{
	m_sDescription.Clear();
	m_sCustomText.Clear();
	m_sNote.Clear();
	m_sUnits.Clear();
	m_sAisle.Clear();

	m_flags = 0;
	m_lQuantity = 0;
	m_lPrice = 0;

	*reinterpret_cast<WORD*>( &m_date ) = dateNone;

	m_alarm = 0;
	m_alarmORIG = 0;
	m_uidAlarmMidi = 0;

	m_prio = 0;

	m_fStoresDIRTY = false;
	m_stores.Clear();
	m_storesORIG.Clear();
	m_perstore.Clear();

	m_icon = 0;
	m_category = catUnfiledCategory;
	m_ulEntryOrder = 0;

	m_dirtySort = dirtyALL;
}

#ifdef DEBUG
void
Item::AssertClear() const
{
	Assert( !m_sDescription.Length() );
	Assert( !m_sCustomText.Length() );
	Assert( !m_sNote.Length() );
	Assert( !m_sUnits.Length() );
	Assert( !m_sAisle.Length() );
	Assert( !m_flags );
	Assert( !m_lQuantity );
	Assert( !m_lPrice );
	Assert( !m_prio );
	Assert( !m_fStoresDIRTY );
	Assert( !m_icon );
	Assert( !m_ulEntryOrder );

	Assert( FNoDate() );
	Assert( dirtyALL == m_dirtySort );
	Assert( catUnfiledCategory == m_category );
}

void
Item::SetCanPack()
{
	m_flags |= flagCANPACK;
}
#endif

ItemState
Item::GetState() const
{
	return ItemStateFromFlags( m_flags );
}

StrT const*
Item::GetAisle( bool const fPerStore, Store const idStore ) const
{
	if( fPerStore )
	{
	    m_tmpRotateAisle = !m_tmpRotateAisle;
	    m_tmpAisle[m_tmpRotateAisle].Set( m_perstore.GetAisle( idStore ) );

	    return &m_tmpAisle[m_tmpRotateAisle];
	}

	return &m_sAisle;
}

WORD
Item::GetIcon( DbOpen* pdb ) const
{
	if( GetRawIcon() )
	    return GetRawIcon();

	return pdb->GetCategoryIcon( GetCategory() );
}

inline TCHAR
GetDecimalSeparator()
{
	return g_intl.GetDecimalSeparator();
}

long
Item::GetUnitCount( long const lWhenParseError ) const
{
	LPCTSTR psz = GetUnits()->Text();

	while( *psz && ( *psz > '9' || *psz < '0' ) && *psz != GetDecimalSeparator() )
	    psz++;

	if( !*psz )
	    return QTY_PRECISION;

	long const lUnitCount = ParseNumber( psz, QTY_PRECISION, 0, true );

	if( parseError == lUnitCount )
	    return lWhenParseError;

	return lUnitCount;
}

bool
Item::GetTaxable() const
{
	return !!( m_flags & flagTaxable );
}

bool
Item::GetTaxable2() const
{
	return !!( m_flags & flagTaxable2 );
}

LPSYSTEMTIME
Item::GetDate() const
{
	if( FNoDate() )
	    return 0;

	ZeroMemory( &m_systimeDummy, sizeof( m_systimeDummy ) );

	m_systimeDummy.wYear = m_date.year + 1904;
	m_systimeDummy.wMonth = m_date.month;
	m_systimeDummy.wDay = m_date.day;

	return &m_systimeDummy;
}

DWORD
Item::GetAutoOrder( bool fPerStore, Store idStore ) const
{
	if( !fPerStore || idStore == storeUnfiledStore )
	    idStore = storeAllStores;

	AutoOrder const* const p = m_autoorder.Find( idStore );

	return p ? p->m_ulOrder : 0;
}

bool
Item::GetAutoDelete() const
{
	return !!( m_flags & flagAutoDelete );
}

bool
Item::GetPrivate() const
{
	return !!( m_flags & flagPrivate );
}

bool
Item::GetCoupon() const
{
	return !!( m_flags & flagCoupon );
}

bool
Item::FInStore( Store const idStore ) const
{
	if( storeAllStores == idStore )
	    return true;

	if( storeUnfiledStore == idStore )
	    return !m_stores.Count();

	return m_stores.Find( idStore );
}

bool
Item::FInSelectedStores( DbOpen const* const pdb ) const
{
	Assert( pdb );

	MultipleStores multi;

	if( pdb->GetMultipleStores( &multi ) )
	{
	    WORD c = m_stores.Count();

	    if( c )
	    {
		for( Store const* p = m_stores.Array(); c--; p++ )
		    if( *p < DimensionOf( multi.m_rg ) && multi.m_rg[*p] )
			return true;
	    }
	    else
	    {
		if( multi.m_Unfiled )
		    return true;
	    }

	    return false;
	}

	Store const storeSelected = pdb->GetStore();

	return FInStore( storeSelected );
}

bool
Item::FInSelectedCategories( DbOpen const* const pdb ) const
{
	Assert( pdb );

	MultipleCategories multi;

	if( pdb->GetMultipleCategories( &multi ) )
	{
	    WORD const category = GetCategory();

	    if( category == catUnfiledCategory )
		return !!multi.m_Unfiled;

	    if( category < DimensionOf( multi.m_rg ) )
		return !!multi.m_rg[category];

	    return false;
	}

	WORD const categorySelected = pdb->GetCategory();

	return ( categorySelected == catAllCategories || categorySelected == GetCategory() );
}

bool
Item::ApplyDatabaseDefaults( DbOpen* pdb, bool fContinueTyping, bool* pfCantPickDefaultCategory )
{
	bool fChanged = false;

	if( pfCantPickDefaultCategory )
	    *pfCantPickDefaultCategory = false;

	// Get database defaults.

	fChanged = SetTaxable( pdb->GetTaxable() ) || fChanged;
	fChanged = SetTaxable2( pdb->GetTwoTaxes() && pdb->GetTaxable() ) || fChanged;
	fChanged = SetAutoDelete( pdb->GetAutoDelete() ) || fChanged;
	fChanged = SetPriority( pdb->GetDefaultPriority() ) || fChanged;

	// Get view defaults.

	if( pdb->FViewNeed( pdb->GetView() ) || ( VIEW_NOTNEEDED == pdb->GetView() && fContinueTyping ) )
	    fChanged = SetState( ITEMSTATE_NEED, 0 ) || fChanged;
	else
	    fChanged = SetState( ITEMSTATE_NONE, 0 ) || fChanged;

	fChanged = SetCoupon( pdb->GetCouponsOnly() ) || fChanged;

	// Category.

	WORD newCategory = pdb->GetCategory();

	if( catAllCategories == newCategory )
	{
	    if( pfCantPickDefaultCategory )
		*pfCantPickDefaultCategory = true;

	    newCategory = catUnfiledCategory;
	}

	fChanged = SetCategory( newCategory ) || fChanged;

	// Set the store based on the current view.

	Assert( !m_fStoresDIRTY );

	m_fStoresDIRTY = true;

	if( !IsPhantomStore( pdb->GetStore() ) )
	{
	    m_stores.Insert( pdb->GetStore() );
	    fChanged = true;
	}

	// Set the quantity based on the item state.

	if( GetState() == ITEMSTATE_NEED )
	    SetQuantity( QTY_PRECISION, 0 );

	return fChanged;
}

static WORD const c_rgFlagsFromState[] =
{
	flagPurchased,		// ITEMSTATE_NONE
	flagNeed,		// ITEMSTATE_NEED
	0,			// ITEMSTATE_CHECKED
	flagNeed|flagPurchased,	// ITEMSTATE_DEFERRED
};
CASSERT( DimensionOf( c_rgFlagsFromState ) == ITEMSTATE_MAX );

bool
Item::SetState( ItemState state, DbOpen const* const pdb, bool fAllowAdjustAutoOrder )
{
	Assert( ITEMSTATE_NONE == state ||
		ITEMSTATE_NEED == state ||
		ITEMSTATE_CHECKED == state ||
		ITEMSTATE_DEFERRED == state );

	WORD flags = m_flags & ~( flagNeed|flagPurchased );

	if( state >= 0 && state < ITEMSTATE_MAX )
	    flags |= c_rgFlagsFromState[state];

	if( m_flags == flags )
	    return false;

	bool const fWasChecked = ( GetState() == ITEMSTATE_CHECKED );

	m_flags = flags;
	m_dirtySort |= dirtyNeed;

	if( pdb )
	{
	    if( pdb->GetAutoQuantity() )
	    {
		if( ITEMSTATE_NONE == state && GetQuantity() )
		    SetQuantity( 0, 0 );
		else if( ITEMSTATE_NEED == state && !GetQuantity() )
		    SetQuantity( QTY_PRECISION, 0 );
	    }

	    if( fAllowAdjustAutoOrder &&
		!fWasChecked &&
		GetState() == ITEMSTATE_CHECKED &&
		pdb->IsLearningOrder() )
	    {
		pdb->AdjustAutoOrder( this );
	    }
	}

	return true;
}

bool
Item::SetAutoDelete( bool f )
{
	WORD flags = m_flags & ~flagAutoDelete;

	CASSERT( sizeof( m_flags ) == sizeof( flags ) );

	if( f )
	    flags |= flagAutoDelete;

	if( m_flags == flags )
	    return false;

	m_flags = flags;

	return true;
}

bool
Item::SetPrivate( bool f )
{
	WORD flags = m_flags & ~flagPrivate;

	CASSERT( sizeof( m_flags ) == sizeof( flags ) );

	if( f )
	    flags |= flagPrivate;

	if( m_flags == flags )
	    return false;

	m_flags = flags;

	return true;
}

bool
Item::SetCoupon( bool f )
{
	WORD flags = m_flags & ~flagCoupon;

	CASSERT( sizeof( m_flags ) == sizeof( flags ) );

	if( f )
	    flags |= flagCoupon;

	if( m_flags == flags )
	    return false;

	m_flags = flags;

	return true;
}

bool
Item::SetDescription( const StrT* ps )
{
	Assert( ps );

	if( ps->Equal( GetDescription() ) )
	    return false;

	m_sDescription.Set( *ps );
	m_dirtySort |= dirtyDescription;

	return true;
}

bool
Item::SetCustomText( const StrT* ps )
{
	Assert( ps );

	if( ps->Equal( GetCustomText() ) )
	    return false;

	m_sCustomText.Set( *ps );
	m_dirtySort |= dirtyCustomText;

	return true;
}

bool
Item::SetNote( const StrT* ps )
{
	if( ps )
	{
	    if( ps->Equal( GetNote() ) )
		return false;

	    m_sNote.Set( *ps );
	}
	else
	{
	    if( !GetNote()->Length() )
		return false;

	    m_sNote.Clear();
	}

	return true;
}

bool
Item::SetUnits( const StrT* ps )
{
	Assert( ps );

	if( ps->Equal( GetUnits() ) )
	    return false;

	m_sUnits.Set( *ps );
	m_dirtySort |= dirtyUnits;

	return true;
}

bool
Item::SetPriority( BYTE prio )
{
	if( prio >= MAX_PRIO )
	    return false;

	if( m_prio == prio )
	    return false;

	m_prio = prio;
	m_dirtySort |= dirtyPrio;

	return true;
}

bool
Item::SetAisle( const StrT* ps, bool fPerStore, Store idStore )
{
	Assert( ps );

	// Reject characters that HS2Convert will have trouble with.

	if( _tcschr( ps->Text(), '/' ) ||
	    _tcschr( ps->Text(), ',' ) ||
	    _tcschr( ps->Text(), '=' ) ||
	    _tcschr( ps->Text(), '\"' ) )
	{
	    return false;
	}

	if( fPerStore )
	{
	    PerStore* p = m_perstore.Find( idStore );

	    if( p )
	    {
		if( LStrCmp( ps->Text(), p->m_szAisle ) == 0 )
		    return false;
	    }
	    else
	    {
		if( !ps->Length() )
		    return false;
	    }

	    if( !p )
		p = m_perstore.Insert( idStore );

	    if( p )
	    {
		WORD const cch = WORD( Min( ps->Length(), DimensionOf( p->m_szAisle ) - 1 ) );

		memcpy( p->m_szAisle, ps->Text(), cch * sizeof( *p->m_szAisle ) );
		p->m_szAisle[cch] = 0;

		if( !p->m_szAisle[0] && !p->m_lPrice )
		    m_perstore.Remove( p );
	    }

	    m_dirtySort |= dirtyAisle|dirtyPerStore;
	}
	else
	{
	    if( ps->Equal( m_sAisle ) )
		return false;

	    m_sAisle.Set( *ps );
	    m_dirtySort |= dirtyAisle;
	}

	return true;
}

bool
Item::SetPrice( long const lPrice, bool const fPerStore, Store const idStore )
{
	if( lPrice <= -MAX_NUMBER * PRICE_PRECISION || lPrice >= MAX_NUMBER * PRICE_PRECISION )
	{
	    Panic1( "Price (%d) out of range.", lPrice );
	    return false;
	}

	if( fPerStore && IsPhantomStore( idStore ) )
	{
	    PanicSz( "Can't set price in pseudo-store." );
	    return false;
	}

	if( GetPrice( fPerStore, idStore ) == lPrice )
	    return false;

	if( fPerStore )
	{
	    PerStore* const p = m_perstore.Insert( idStore );

	    if( p )
	    {
		p->m_lPrice = lPrice;

		if( !p->m_szAisle[0] && !p->m_lPrice )
		    m_perstore.Remove( p );
	    }

	    m_dirtySort |= dirtyPrice|dirtyPerStore;
	}
	else
	{
	    m_lPrice = lPrice;
	    m_dirtySort |= dirtyPrice;
	}

	return true;
}

bool
Item::SetQuantity( long const lQuantity, DbOpen const* const pdb )
{
	if( lQuantity < 0 || lQuantity >= long( MAX_NUMBER ) * long( QTY_PRECISION ) )
	{
	    Panic1( "Quantity (%d) out of range.", lQuantity );
	    return false;
	}

	if( m_lQuantity == lQuantity )
	    return false;

	long const lBefore = m_lQuantity;

	m_lQuantity = lQuantity;
	m_dirtySort |= dirtyQuantity;

	if( pdb && pdb->GetAutoQuantity() )
	{
	    if( !lBefore && lQuantity )
		SetState( ITEMSTATE_NEED, 0 );
	    else if( lBefore && !lQuantity )
		SetState( ITEMSTATE_NONE, 0 );
	}

	return true;
}

bool
Item::SetTaxable( bool f )
{
	WORD flags = m_flags & ~flagTaxable;

	CASSERT( sizeof( m_flags ) == sizeof( flags ) );

	if( f )
	    flags |= flagTaxable;

	if( m_flags == flags )
	    return false;

	m_flags = flags;

	return true;
}

bool
Item::SetTaxable2( bool f )
{
	WORD flags = m_flags & ~flagTaxable2;

	CASSERT( sizeof( m_flags ) == sizeof( flags ) );

	if( f )
	    flags |= flagTaxable2;

	if( m_flags == flags )
	    return false;

	m_flags = flags;

	return true;
}

bool
Item::SetDate( LPSYSTEMTIME psystime )
{
	DateType date;

	//$ todo: (chrisant) must validate the value

	if( psystime && psystime->wYear )
	{
	    date.year = psystime->wYear - 1904;
	    date.month = psystime->wMonth;
	    date.day = psystime->wDay;
	}
	else
	{
	    *reinterpret_cast<WORD* >( &date ) = dateNone;
	}

	if( memcmp( &m_date, &date, sizeof( date ) ) == 0 )
	    return false;

	memcpy( &m_date, &date, sizeof( date ) );
	m_dirtySort |= dirtyDate;

	return true;
}

bool
Item::SetEntryOrder( DWORD const ulEntryOrder )
{
	if( ulEntryOrder == m_ulEntryOrder )
	    return false;

	m_ulEntryOrder = ulEntryOrder;
	m_dirtySort |= dirtyEntryOrder;

	return true;
}

bool
Item::SetAutoOrder( DWORD const ulAutoOrder, bool const fPerStore, Store idStore )
{
	Assert( idStore != storeMultipleStores );
	Assert( idStore != storeEditStores );

	if( !fPerStore || idStore == storeUnfiledStore )
	    idStore = storeAllStores;

	if( GetAutoOrder( fPerStore, idStore ) == ulAutoOrder )
	    return false;

	AutoOrder* const p = m_autoorder.Insert( idStore );

	if( p )
	{
	    p->m_ulOrder = ulAutoOrder;
	    m_dirtySort |= dirtyAutoOrder;

	    if( !ulAutoOrder )
	    {
		if( idStore != storeAllStores )
		    m_autoorder.Remove( p );
		if( m_autoorder.Count() == 1 && !m_autoorder[0].m_ulOrder )
		    m_autoorder.Clear();
	    }
	}

	return true;
}

bool
Item::SetRawIcon( WORD const icon )
{
	if( icon == m_icon )
	    return false;

	m_icon = icon;
	m_dirtySort |= dirtyIcon;

	return true;
}

bool
Item::SetAlarm( DWORD const alarmTime )
{
	if( alarmTime == m_alarm && !alarmTime == !( m_flags & flagHasAlarm ) )
	    return false;

	m_alarm = alarmTime;
	m_dirtySort |= dirtyAlarm;

	if( alarmTime )
	    m_flags |= flagHasAlarm;
	else
	    m_flags &= ~flagHasAlarm;

	return true;
}

bool
Item::SetAlarmMidi( DWORD const uidAlarmMidi )
{
	if( uidAlarmMidi == m_uidAlarmMidi )
	    return false;

	m_uidAlarmMidi = uidAlarmMidi;
	m_dirtySort |= dirtyAlarm;

	return true;
}

bool
Item::SetCategory( WORD const category )
{
	Assert( ( category >= 0 && category < c_cMaxCategories ) || category == catUnfiledCategory );

	if( ( category < 0 || category >= c_cMaxCategories ) && category != catUnfiledCategory )
	    return false;

	if( category == m_category )
	    return false;

	m_category = category;
	m_dirtySort |= dirtyCategory;

	m_flags |= flagCATEGORY;

	return true;
}

bool
Item::EnsureStoreAssociation( Store const idStore )
{
	if( IsPhantomStore( idStore ) )
	    return false;

	if( FInStore( idStore ) )
	    return false;

	if( !m_fStoresDIRTY )
	{
	    m_storesORIG.Copy( m_stores );
	    m_fStoresDIRTY = true;
	}

	m_stores.Insert( idStore );

	return true;
}

bool
Item::RemoveStoreAssociation( Store const idStore )
{
	if( IsPhantomStore( idStore ) )
	    return false;

	if( !m_fStoresDIRTY )
	{
	    m_storesORIG.Copy( m_stores );
	    m_fStoresDIRTY = true;
	}

	return m_stores.Remove( idStore );
}

bool
Item::ZapStoreData( Store const idStore )
{
	bool fDidAnyWork = RemoveStoreAssociation( idStore );

	PerStore const* const pPerStore = m_perstore.Find( idStore );
	if( pPerStore )
	{
	    m_perstore.Remove( pPerStore );
	    m_dirtySort |= dirtyPerStore;
	    fDidAnyWork = true;
	}

	AutoOrder const* const pAutoOrder = m_autoorder.Find( idStore );
	if( pAutoOrder )
	{
	    m_autoorder.Remove( pAutoOrder );
	    m_dirtySort |= dirtyAutoOrder;
	    fDidAnyWork = true;
	}

	return fDidAnyWork;
}

void
Item::StampToday()
{
	SYSTEMTIME systime;

#if defined( OS_WINDOWS )
	GetLocalTime( &systime );
#elif defined( OS_CRT )
	struct tm tmNow;
	time_t const timeNow = time( 0 );
	localtime_s( &tmNow, &timeNow );
	systime.wYear = LOWORD( tmNow.tm_year + 1900 );
	systime.wMonth = LOWORD( tmNow.tm_mon + 1 );
	systime.wDay = LOWORD( tmNow.tm_mday );
#else
#error Item::StampToday not implemented for this platform.
#endif

	SetDate( &systime );
}

void
Item::PrepareForOtherDatabase( DbOpen* pdbTo )
{
	// Say the item wasn't needed, so the target database doesn't get its
	// Unfiled bold count screwed up.

	m_flags &= ~flagWasNeeded;

	// Initialize the item per the prefs of destination database.

	if( pdbTo->GetRecordCreationDate() )
	    StampToday();

	// Clear the entry order and auto order because they are
	// non-transferrable.  The entry order will get initialized when the
	// record gets written to the database.  The auto order will get
	// updated when the item is checked off while Learn Order is active.

	m_ulEntryOrder = 0;
	ClearAutoOrders();

	// Make the category Unfiled for now.

	//$ todo: (chrisant) Try to translate the category?
	m_category = catUnfiledCategory;
}

bool
Item::ClearAutoOrders()
{
	if( m_autoorder.Count() )
	{
	    m_autoorder.Clear();
	    m_dirtySort |= dirtyAutoOrder;
	    return true;
	}
	return false;
}

bool
Item::ClearAutoOrder( Store const idStore )
{
	AutoOrder* const p = m_autoorder.Find( idStore );
	if( p )
	{
	    m_autoorder.Remove( p );
	    if( m_autoorder.Count() == 1 && !m_autoorder[0].m_ulOrder )
		m_autoorder.Clear();
	    m_dirtySort |= dirtyAutoOrder;
	    return true;
	}
	return false;
}

