/*
 * Item.h - Item object.
 */

#pragma once

// vim: set noet:			// Hard tabs.
// vim: set ts=8:			// Tab width 8.
// vim: set sw=4:			// Shift (indent) width 4.
// vim: set cino={0s:			// Braces flush with outer statement.
// vim: set cino=^1s:			// First indent 8 (add 1 shift width).

/*
 * Item class.
 */

#define dateNone		0xffff	// max possible, so it sorts last

#define cchMaxDescription	( 255 )
#define cchMaxCustomText	( 255 )
#define cchMaxNumberText	( 10 )	// "-99999.999", but NO null terminator
#define cchMaxUnit		( 16 )	// 15 + 1 for null terminator

#define MAX_PRIO		( 5 )	// max priority levels
#define MAX_NUMERIC_AISLES	( 200 )
#define MAX_AISLES		( MAX_NUMERIC_AISLES + 26 )

#include "str.h"

typedef DWORD DirtyType;
#define dirtyDescription	0x0001
#define dirtyNeed		0x0002
#define dirtyQuantity		0x0004
#define dirtyPrice		0x0008
#define dirtyPrio		0x0010
#define dirtyDate		0x0020
#define dirtyAlarm		0x0040
#define dirtyUnits		0x0080
#define dirtyCategory		0x0100
#define dirtyStores		0x0200
#define dirtyAisle		0x0400
#define dirtyEntryOrder		0x0800
#define dirtyPerStore		0x1000
#define dirtyCustomText		0x2000
#define dirtyIcon		0x4000
#define dirtyAutoOrder		0x8000
#define dirtyALL		0xffff


extern const TCHAR g_chPrioBase;

class DbOpen;

/*
 * Item flag values.
 *
 * IMPORTANT NOTE:  These must match the HS2 values for Palm compatibility!
 */

#define flagMaskPersisted	0x07FF	// only these bits get persisted
#define flagNeed		0x0001
// dead, don't reuse		0x0002	// used to be coupon long ago
#define flagTaxable		0x0004	// Tax1 (or GST for Canada)
#define flagPrivate		0x0008	// this item is private (used instead of dmRecAttrSecret)
#define flagHasAlarm		0x0010	// this item has an alarm set (persisted)
//#define flagUnits		0x0020	//$ review: (chrisant) not sure what the deal is with this one
#define flagPurchased		0x0040
#define flagHasStore		0x0080
#define flagTaxable2		0x0100	// Tax2 (or PST for Canada)
#define flagCoupon		0x0200
#define flagAutoDelete		0x0400	// delete this item at next Checkout, if item is completed

#define flagWasNeeded		0x0800	// so PackRecord can correctly update the bold counts
#define flagMetadataRecord	0x1000	// indicates this record is really the store list or category list for this database (Palm only)

#define flagCATEGORY		0x4000

#ifdef DEBUG
#define flagCANPACK		0x8000
#endif

/*
 * Item state values.
 */

enum ItemState
{
	ITEMSTATE_NONE,
	ITEMSTATE_NEED,
	ITEMSTATE_CHECKED,
	ITEMSTATE_DEFERRED,
	ITEMSTATE_MAX
};

inline ItemState
ItemStateFromFlags( WORD const flags )
{
	switch( flags & ( flagNeed|flagPurchased ) )
	{
	case flagNeed:			return ITEMSTATE_NEED;
	case flagPurchased:		return ITEMSTATE_NONE;
	case flagNeed|flagPurchased:	return ITEMSTATE_DEFERRED;
	}
	return ITEMSTATE_CHECKED;
}

/*
 * ItemQuickInfo class.
 */

class ItemQuickInfo
{
    public:
	void		Initialize( WORD flags, bool fDeleted = false );

	bool		IsRecordDeleted() const { return m_fDeleted; }
	bool		IsAutoDelete() const { return m_fAutoDelete; }
	bool		IsPrivate() const { return m_fPrivate; }
	ItemState	GetState() const { return m_state; }

    private:
	bool		m_fDeleted;
	bool		m_fAutoDelete;
	bool		m_fPrivate;
	ItemState	m_state;
};

/*
 * DateType for 16-bit date format.
 */

struct DateType
{
	WORD		Value() const { return *reinterpret_cast<WORD const*>( this ); }

	unsigned short	day	:5;
	unsigned short	month	:4; 
	unsigned short	year	:7;	// years since 1904 (MAC format)
};

/*
 * Item class.
 */

class Item
{
	friend class DbOpenHs2;
	friend class DbOpenSimpleHs2;
	//friend Err DbOpenHs2::PackRecord( UInt16 record, Item** ppitem, UInt16 wPackFlags, UInt16* pces );

    public:
	static unsigned const offsetPriority;
	static unsigned const offsetDescription;
	static unsigned const offsetCustomText;
	static unsigned const offsetNote;
	static unsigned const offsetQuantity;
	static unsigned const offsetUnits;
	static unsigned const offsetAisle;
	static unsigned const offsetPrice;
	static unsigned const offsetCategory;
	static unsigned const offsetDate;
	static unsigned const offsetEntryOrder;
	static unsigned const offsetStores;
	static unsigned const offsetPerStore;
	static unsigned const offsetIcon;
	static unsigned const offsetAlarm;
	static unsigned const offsetAlarmMidi;
	static unsigned const offsetAutoOrder;

    public:
			Item();
			~Item();

	void		Clear();

#ifdef DEBUG
	void		AssertClear() const;
	void		SetCanPack();
	bool		CanPack() const { return !!( m_flags & flagCANPACK ); }
	AutoOrderListValidity TestAutoOrderListValid() const { return m_autoorder.TestValid(); }
#endif

	// Get accessors.

	ItemState	GetState() const;
	bool		GetAutoDelete() const;
	bool		GetPrivate() const;
	bool		GetCoupon() const;

	StrT const*	GetDescription() const { return &m_sDescription; }
	StrT const*	GetCustomText() const { return &m_sCustomText; }
	StrT const*	GetNote() const { return &m_sNote; }
	StrT const*	GetUnits() const { return &m_sUnits; }
	StrT const*	GetAisle( bool fPerStore = false, Store idStore = storeUnfiledStore ) const;

	BYTE		GetPriority() const { return m_prio; }
	long		GetPrice( bool fPerStore = false, Store idStore = storeUnfiledStore ) const;
	long		GetQuantity() const { return m_lQuantity; }
	long		GetUnitCount( long lWhenParseError = 0 ) const;
	bool		GetTaxable() const;
	bool		GetTaxable2() const;
	LPSYSTEMTIME	GetDate() const;
	DWORD		GetEntryOrder() const { return m_ulEntryOrder; }
	DWORD		GetAutoOrder( bool fPerStore = false, Store idStore = storeUnfiledStore ) const;

	WORD		GetIcon( DbOpen* pdb ) const;
	WORD		GetRawIcon() const { return m_icon; }
	DWORD		GetAlarm() const { return m_alarm; }
	DWORD		GetAlarmMidi() const { return m_uidAlarmMidi; }

	WORD		GetCategory() const { return m_category; }

	Store		GetLowestPriceInfo( long* plPrice, Store idCheck ) const;

	bool		FNoDate() const { return dateNone == *reinterpret_cast<WORD const*>( &m_date ) || 0 == *reinterpret_cast<WORD const*>( &m_date ); }
	bool		FEmpty() const { Assert( this ); return !m_sDescription.Length() && !m_sCustomText.Length() && !m_sNote.Length(); }

	bool		FInStore( Store idStore ) const;
	bool		FInSelectedStores( DbOpen const* pdb ) const;
	bool		FInSelectedCategories( DbOpen const* pdb ) const;
	WORD		CountAutoOrders() const { return m_autoorder.Count(); }
	AutoOrder const* GetAutoOrderByIndex( WORD index ) const { return &m_autoorder[index]; }

	// Set accessors.

	bool		ApplyDatabaseDefaults( DbOpen* pdb, bool fContinueTyping = false, bool* pfCantPickDefaultCategory = 0 );

	bool		SetState( ItemState state, DbOpen const* pdb, bool fAllowAdjustAutoOrder = false );
	bool		SetAutoDelete( bool f );
	bool		SetPrivate( bool f );
	bool		SetCoupon( bool f );

	bool		SetDescription( const StrT* ps );
	bool		SetCustomText( const StrT* ps );
	bool		SetNote( const StrT* ps );
	bool		SetUnits( const StrT* ps );
	bool		SetAisle( const StrT* ps, bool fPerStore = false, Store idStore = storeUnfiledStore );

	bool		SetPriority( BYTE prio );
	bool		SetPrice( long lPrice, bool fPerStore = false, Store idStore = storeUnfiledStore );
	bool		SetQuantity( long lQuantity, DbOpen const* pdb );
	bool		SetTaxable( bool f );
	bool		SetTaxable2( bool f );
	bool		SetDate( LPSYSTEMTIME psystem );
	bool		SetEntryOrder( DWORD ulEntryOrder );
	bool		SetAutoOrder( DWORD ulAutoOrder, bool fPerStore = false, Store idStore = storeUnfiledStore );

	bool		SetRawIcon( WORD icon );
	bool		SetAlarm( DWORD alarmTime );
	bool		SetAlarmMidi( DWORD uidAlarmMidi );

	bool		SetCategory( WORD category );

	bool		EnsureStoreAssociation( Store idStore );
	bool		RemoveStoreAssociation( Store idStore );
	bool		ZapStoreData( Store idStore );

	void		StampToday();

	void		InitFromRecordHeader( WORD flags, WORD category );
	void		PrepareForOtherDatabase( DbOpen* pdb );
	bool		ClearAutoOrders();
	bool		ClearAutoOrder( Store idStore );

	void		InitAlarmOrig() { m_alarmORIG = m_alarm; }

	DirtyType	GetSortDirtyMask() const { return m_dirtySort; }

	//$ todo: (chrisant) maybe have accessors on Item to automatically
	// manage m_storesORIG versus m_stores correctly?

    public:
	WORD		m_flags;
	DateType	m_date;

	bool		m_fStoresDIRTY;
	StoreList	m_stores;
	StoreList	m_storesORIG;
	PerStoreList	m_perstore;

	DirtyType	m_dirtySort;		// used for stable sort

    private:
	StrT		m_sDescription;
	StrT		m_sCustomText;
	StrT		m_sNote;
	StrT		m_sUnits;
	StrT		m_sAisle;

	BYTE		m_prio;

	LONG		m_lQuantity;
	LONG		m_lPrice;

	DWORD		m_alarm;
	DWORD		m_alarmORIG;
	DWORD		m_uidAlarmMidi;

	WORD		m_icon;
	DWORD		m_ulEntryOrder;

	WORD		m_category;

	AutoOrderList	m_autoorder;

	mutable SYSTEMTIME	m_systimeDummy;
	mutable BYTE		m_tmpRotateAisle;
	mutable StrT		m_tmpAisle[2];
};

inline long
Item::GetPrice( bool fPerStore, Store idStore ) const
{
	if( !fPerStore )
	    return m_lPrice;

	return m_perstore.GetPrice( idStore );
}

inline Store
Item::GetLowestPriceInfo( long* plPrice, Store idCheck ) const
{
	return m_perstore.GetLowestPriceInfo( plPrice, idCheck );
}

inline void
Item::InitFromRecordHeader( WORD const flags, WORD const category )
{
	Assert( !( m_flags & flagWasNeeded ) );

	m_flags = flags;
	m_category = category;

	if( GetState() == ITEMSTATE_NEED )
	    m_flags |= flagWasNeeded;
}

