/*
 * Number.cpp - abstraction around a floating point number.
 */

// vim: set noet:			// Hard tabs.
// vim: set ts=8:			// Tab width 8.
// vim: set sw=4:			// Shift (indent) width 4.
// vim: set cino={0s:			// Braces flush with outer statement.
// vim: set cino=^1s:			// First indent 8 (add 1 shift width).

#include "precomp.h"

#include "number.h"
#include "str.h"

/*
 * NUMBER - abstraction around a floating point number.
 */

long
NUMBER::ToLong_Internal( WORD const precision ) const
{
	double const d = m_value * precision;

	if( d < 0 )
	    return long( d - 0.5 );

	return long( d + 0.5 );
}

static double const s_dMin = -MAX_NUMBER;
static double const s_dMax = MAX_NUMBER;

bool
NUMBER::IsOverflow() const
{
	return ( m_value < s_dMin || m_value > s_dMax );
}

/*
 * Number and string helpers.
 */

inline TCHAR
GetDecimalSeparator()
{
	return g_intl.GetDecimalSeparator();
}

inline TCHAR
GetThousandsSeparator()
{
	return g_intl.GetThousandsSeparator();
}

inline TCHAR const*
GetCurrency()
{
	return g_intl.GetCurrency();
}

void
SkipDigits( TCHAR const** ppsz )
{
	while( ( **ppsz >= '0' && **ppsz <= '9' ) || **ppsz == GetDecimalSeparator() )
	    ( *ppsz )++;
}

void
SkipSpaces( TCHAR const** ppsz )
{
	while( **ppsz == ' ' || **ppsz == '\t' )
	    ( *ppsz )++;
}

void
SkipLinefeeds( TCHAR const** ppsz )
{
	while( **ppsz == '\n' || **ppsz == '\r' )
	    ( *ppsz )++;
}

bool
SkipCurrency( TCHAR const** ppsz, LPCTSTR pszCurrency )
{
	LPCTSTR psz = *ppsz;
	bool fCurrency = false;

	// Skip currency symbol, but CASE INSENSITIVE.

	if( !pszCurrency )
	    pszCurrency = GetCurrency();

	while( *pszCurrency && _tcsnicmp( psz, pszCurrency, 1 ) == 0 )
	{
	    ++pszCurrency;
	    ++psz;
	}

	// Always allow $ as currency symbol.

	if( *psz == '$' )
	{
	    ++psz;
	    fCurrency = true;
	}

	if( pszCurrency && !*pszCurrency )
	    fCurrency = true;

	*ppsz = psz;

	return fCurrency;
}

static WORD
PlacesFromPrecision( WORD precision )
{
	WORD places = 0;

	// Get places from precision.

	while( precision > 1 )
	{
	    precision /= 10;
	    places++;
	}

	return places;
}

LPCTSTR
LocalizeNumber( long l, WORD flags, WORD precision, int* pcch, LPCTSTR pszCurrency )
{
	static TCHAR sz[cchMaxLocalizedNumber];

	int ii;
	LPTSTR psz = &sz[DimensionOf( sz ) - 1];
	bool fTrimZeros = !!( flags & ( lnTrimRightZeros|lnTrimRightZeros2 ) );
	bool fNegative = ( l < 0 );

	if( l == parseError )
	{
	    CchCopy( TEXT( "####" ), RgC( sz ) );
	    return sz;
	}

	*psz = 0;

	if( !pszCurrency || !*pszCurrency )
	    pszCurrency = GetCurrency();

	if( fNegative )
	    l = -l;

	if( l || ( flags & ( lnCurrencySymbol|lnShowZero ) ) )
	{
	    // Fractional part.

	    if( ( flags & lnShowDecimals ) || !( flags & lnAllowRounding ) )
	    {
		for( ii = PlacesFromPrecision( precision ); ii--; l /= 10 )
		{
		    if( ( flags & lnTrimRightZeros2 ) && ii < 2 )
			fTrimZeros = false;

		    if( !fTrimZeros || l % 10 )
		    {
			*( --psz ) = TCHAR( l % 10 ) + '0';
			fTrimZeros = false;
		    }
		}

		if( !fTrimZeros )
		    *( --psz ) = GetDecimalSeparator();
	    }
	    else
	    {
		//$ review: (chrisant) do stores always round up from .5?

		if( ( l % precision ) >= precision / 2 )
		    if( l < 99999 )	// don't allow rounding up into 6 digits
			l += precision;

		l /= precision;
	    }

	    // Whole number.

	    do
	    {
		*( --psz ) = TCHAR( l % 10 ) + '0';
		l /= 10;
	    }
	    while( l );

	    if( fNegative )
		*( --psz ) = '-';

	    // Currency symbol.

	    if( flags & lnCurrencySymbol )
	    {
		unsigned const cchCurrency = CchLen( pszCurrency );

		if( cchCurrency )
		{
		    LPCTSTR pszSym = pszCurrency + cchCurrency - 1;

		    Assert( pszSym >= pszCurrency );

		    // Automatically insert a space if it's not a symbol.

		    if( isalnum( *pszSym ) )
			*( --psz ) = ' ';

		    while( pszSym >= pszCurrency )
		    {
			*( --psz ) = *pszSym;
			pszSym--;
		    }
		}
	    }
	}

	if( pcch )
	    *pcch = ( sz + ( DimensionOf( sz ) - 1 ) - psz );

	return psz;
}

LPCTSTR
LocalizeNumber( NUMBER num, WORD flags, WORD precision, int* pcch, LPCTSTR pszCurrency )
{
#if 0
	// Try to compensate for rounding errors.
	NUMBER tweak;
	tweak.SetRawValue( 0.0000000001 );
	num.Add( num, tweak );
#endif
	long const l = num.ToLong_OverflowOk( precision );
	return LocalizeNumber( l, flags, precision, pcch, pszCurrency );
}

long
ParseNumber( LPCTSTR psz, WORD precision, TCHAR const** ppszOut, bool fSilent, LPCTSTR pszCurrency )
{
	SUPPRESS_UNUSED_VAR( fSilent );

	long l;
	long lPrice;
	bool fDecimal = false;
	WORD cDecimal;
	WORD places;
	bool fNegative = false;
	bool fDisallowNegative;

	l = lPrice = 0;
	cDecimal = 0;
	fDisallowNegative = short( precision ) < 0;

	if( fDisallowNegative )
	    precision = -precision;

	places = PlacesFromPrecision( precision );

	if( !psz )
	    return 0;

	// Skip currency symbol.

	SkipSpaces( &psz );
	SkipCurrency( &psz, pszCurrency );
	SkipSpaces( &psz );

	// Check for negative.

	if( *psz == '-' )
	{
	    psz++;
	    fNegative = true;
	}

	// Parse number.

	for( ; *psz; ++psz )
	{
	    if( ( *psz ) >= '0' && ( *psz ) <= '9' )
	    {
		if( fDecimal && cDecimal >= places )
		{
		    if( psz[1] >= '5' && psz[1] <= '9' ) // banker's rule
			l++;
		    break;
		}

		l *= 10;
		l += ( ( *psz ) - '0' );
		cDecimal++;

		if( !fDecimal && cDecimal > MAX_PLACES )
		{
		    //$ todo: (chrisant) error message
#if 0
		    if( !fSilent )
			Error( errNumericOverflow );
#endif
		    return parseError;
		}
	    }
	    else if( *psz == GetDecimalSeparator() && !fDecimal )
	    {
		lPrice = l;
		l = 0;
		fDecimal = true;
		cDecimal = 0;
	    }
	    else if( GetThousandsSeparator() != ' ' && !( *psz == GetThousandsSeparator() /*|| *psz == '$'*/ ) )
		break;
	}

	if( !fDecimal )
	{
	    lPrice = l;
	    l = 0;
	}

	lPrice *= precision;

	while( cDecimal++ < places )
	    l *= 10;

	lPrice += l;

	if( fNegative )
	    lPrice = -lPrice;

	if( fDisallowNegative && lPrice < 0 )
	{
	    //$ todo: (chrisant) error message
#if 0
	    if( !fSilent )
		Error( errNumericNegative );
#endif
	    return parseError;
	}

	if( ppszOut )
	{
	    *ppszOut = psz;
	    SkipDigits( ppszOut );
	}

	return lPrice;
}

