/*
 * SecureGlue.h - Ease compilation for compilers that don't support the secure
 * CRT function signatures.  This a shim layer, to allow folks who don't have
 * access to the secure CRT functions to still compile these sources.
 *
 * DANGER:  Some of these are secure, but some are just syntactic sugar to
 * allow compilation without actually being secure.  For example the time
 * related ones are still not threadsafe and the itoa ones can still overrun
 * the buffer.
 */

#pragma once

// vim: set noet:			// Hard tabs.
// vim: set ts=8:			// Tab width 8.
// vim: set sw=4:			// Shift (indent) width 4.
// vim: set cino={0s:			// Braces flush with outer statement.
// vim: set cino=^1s:			// First indent 8 (add 1 shift width).

#if !defined( __STDC_WANT_SECURE_LIB__ )

#include <wchar.h>

typedef int errno_t;

inline errno_t
fopen_s( FILE** ppfile, char const* name, char const* mode )
{
	*ppfile = fopen( name, mode );

	if( *ppfile )
	    return 0;

#if defined( OS_WINDOWSMOBILE )
	errno_t const errno = 1;
#endif

	return errno;
}

inline errno_t
_tfopen_s( FILE** ppfile, TCHAR const* name, TCHAR const* mode )
{
	*ppfile = _tfopen( name, mode );

	if( *ppfile )
	    return 0;

#if defined( OS_WINDOWSMOBILE )
	errno_t const errno = 1;
#endif

	return errno;
}

inline void
_itoa_s( int x, char* pszOut, size_t cchMax, int radix )
{
	_itoa( x, pszOut, radix );
}

inline void
_itot_s( int x, TCHAR* pszOut, size_t cchMax, int radix )
{
	_itot( x, pszOut, radix );
}

#if !defined( OS_WINDOWS )

inline errno_t
localtime_s( struct tm* ptm, time_t const* ptime )
{
	struct tm const* const ptmTemp = localtime( ptime );

	if( ptmTemp )
	{
	    *ptm = *ptmTemp;
	    return 0;
	}

#if defined( OS_WINDOWSMOBILE )
	errno_t const errno = 1;
#endif

	return errno;
}

#endif

#if !defined( OS_WINDOWSMOBILE )

inline errno_t
_gmtime32_s( struct tm* ptm, time_t const* ptime )
{
	struct tm const* const ptmTemp = gmtime( ptime );

	if( ptmTemp )
	{
	    *ptm = *ptmTemp;
	    return 0;
	}

#if defined( OS_WINDOWSMOBILE )
	errno_t const errno = 1;
#endif

	return errno;
}

#endif

int vsprintf_s( char* pszOut, size_t cchMax, char const* pszFormat, va_list args );
int vswprintf_s( WCHAR* pszOut, size_t cchMax, WCHAR const* pszFormat, va_list args );
int _vstprintf_s( TCHAR* pszOut, size_t cchMax, TCHAR const* pszFormat, va_list args );

int sprintf_s( char* pszOut, size_t cchMax, char const* pszFormat, ... );
int swprintf_s( WCHAR* pszOut, size_t cchMax, WCHAR const* pszFormat, ... );
int _stprintf_s( TCHAR* pszOut, size_t cchMax, TCHAR const* pszFormat, ... );

#endif

