/*
 * SmartPointers.h - Smart pointers.
 */

#pragma once

// vim: set noet:			// Hard tabs.
// vim: set ts=8:			// Tab width 8.
// vim: set sw=4:			// Shift (indent) width 4.
// vim: set cino={0s:			// Braces flush with outer statement.
// vim: set cino=^1s:			// First indent 8 (add 1 shift width).

#if defined( _MSC_VER )
#if _MSC_VER >= 1400
#pragma warning( disable: 4284 )
#endif
#endif

#ifdef USE_RTTI
#define TypeIdName( x )		typeid( x ).name()
#define TypeIdName2( x, y )	typeid( x, y ).name()
#define TypeIdName3( x, y, z )	typeid( x, y, z ).name()
#else
#define TypeIdName( x )		TEXT( #x )
#define TypeIdName2( x, y )	TEXT( #x ) TEXT( "," ) TEXT( #y )
#define TypeIdName3( x, y, z )	TEXT( #x ) TEXT( "," ) TEXT( #y ) TEXT( "," ) TEXT( #z )
#endif

#ifdef _MSC_VER
//$ WARNING: (chrisant) __if_exists() is not reliable; must find a different
// mechanism.
//#define USE_TYPE_TRAITS
#endif

#ifdef USE_TYPE_TRAITS

#define DEFINE_HAS_MEMBER( Member ) \
namespace type_traits \
{ \
	template<typename T> struct has_##Member \
	{ \
	    enum { value = \
	    __if_exists( T::Member ) { true } \
	    __if_not_exists( T::Member ) { false } \
	    }; \
	}; \
}

DEFINE_HAS_MEMBER( AddRef );
DEFINE_HAS_MEMBER( Release );

#endif // USE_TYPE_TRAITS

typedef void ( *FreeFn )( void* pv );

/*
 * SPNew<Type> - Smart pointer for new/delete objects.
 */

template<class Type>
class SPNew
{
    public:
	SPNew( Type* p = 0 )		{ m_p = p; CASSERT_TRAITS(); }
	~SPNew()			{ delete m_p; }
	operator Type*() const		{ return m_p; }
	Type* Pointer() const		{ return m_p; }
	Type& operator*() const		{ Assert( m_p ); return *m_p; }
	Type& operator[]( int i ) const	{ Assert( m_p ); return m_p[i]; }
	Type& operator[]( unsigned i ) const { Assert( m_p ); return m_p[i]; }
#ifdef _WIN64
	Type& operator[]( __int64 i ) const { Assert( m_p ); return m_p[i]; }
	Type& operator[]( unsigned __int64 i ) const { Assert( m_p ); return m_p[i]; }
#endif
	Type** operator &()		{ Assert1( !m_p, "Non-empty %s as out param.", TypeIdName( SPNew<Type> ) ); return &m_p; }
	Type* operator=( Type* p )	{ Assert1( !m_p, "Non-empty %s in assignment.", TypeIdName( SPNew<Type> ) ); return m_p = p; }
	Type* Transfer()		{ Type* const p = m_p; m_p = 0; return p; }
	void Free()			{ delete m_p; m_p = 0; }
	Type* operator->() const	{ return m_p; }
	bool operator!()		{ return !m_p; }

	Type** UnsafeAddress()		{ return &m_p; }

    protected:
	Type* m_p;

    private:
	SPNew<Type>& operator=( SPNew<Type> const& ); // Forbidden.
	SPNew( SPNew<Type> const& ); // Forbidden.

	void CASSERT_TRAITS()
	{
#ifdef USE_TYPE_TRAITS
	    CASSERT_INLINE( !type_traits::has_AddRef<Type>::value &&
			    !type_traits::has_Release<Type>::value );
#endif
	}
};

/*
 * SPNewArray<Type> - Smart pointer for new[]/delete[] objects.
 */

template<class Type>
class SPNewArray
{
    public:
	SPNewArray( Type* p = 0 )	{ m_p = p; CASSERT_TRAITS(); }
	~SPNewArray()			{ delete [] m_p; }
	operator Type*() const		{ return m_p; }
	Type* Pointer() const		{ return m_p; }
	Type& operator*() const		{ Assert( m_p ); return *m_p; }
	Type& operator[]( int i ) const	{ Assert( m_p ); return m_p[i]; }
	Type& operator[]( unsigned i ) const { Assert( m_p ); return m_p[i]; }
#ifdef _WIN64
	Type& operator[]( __int64 i ) const { Assert( m_p ); return m_p[i]; }
	Type& operator[]( unsigned __int64 i ) const { Assert( m_p ); return m_p[i]; }
#endif
	Type** operator &()		{ Assert1( !m_p, "Non-empty %s as out param.", TypeIdName( SPNewArray<Type> ) ); return &m_p; }
	Type* operator=( Type* p )	{ Assert1( !m_p, "Non-empty %s in assignment.", TypeIdName( SPNewArray<Type> ) ); return m_p = p; }
	Type* Transfer()		{ Type* const p = m_p; m_p = 0; return p; }
	void Free()			{ delete [] m_p; m_p = 0; }
	Type* operator->() const	{ return m_p; }
	bool operator!()		{ return !m_p; }

	Type** UnsafeAddress()		{ return &m_p; }

    protected:
	Type* m_p;

    private:
	SPNewArray<Type>& operator=( SPNewArray<Type> const& ); // Forbidden.
	SPNewArray( SPNewArray<Type> const& ); // Forbidden.

	void CASSERT_TRAITS()
	{
#ifdef USE_TYPE_TRAITS
	    CASSERT_INLINE( !type_traits::has_AddRef<Type>::value &&
			    !type_traits::has_Release<Type>::value );
#endif
	}
};

/*
 * SPMem<Type> - Smart pointer for MemAlloc/MemFree memory blocks.
 */

template<class Type>
class SPMem
{
    public:
	SPMem( Type* p = 0 )		{ m_p = p; CASSERT_TRAITS(); }
	~SPMem()			{ MemFree( m_p ); }
	operator Type*() const		{ return m_p; }
	Type* Pointer() const		{ return m_p; }
	Type& operator*() const		{ Assert( m_p ); return *m_p; }
	Type& operator[]( int i ) const	{ Assert( m_p ); return m_p[i]; }
	Type& operator[]( unsigned i ) const { Assert( m_p ); return m_p[i]; }
#ifdef _WIN64
	Type& operator[]( __int64 i ) const { Assert( m_p ); return m_p[i]; }
	Type& operator[]( unsigned __int64 i ) const { Assert( m_p ); return m_p[i]; }
#endif
	Type** operator &()		{ Assert1( !m_p, "Non-empty %s as out param.", TypeIdName( SPMem<Type> ) ); return &m_p; }
	Type* operator=( Type* p )	{ Assert1( !m_p, "Non-empty %s in assignment.", TypeIdName( SPMem<Type> ) ); return m_p = p; }
	Type* Transfer()		{ Type* const p = m_p; m_p = 0; return p; }
	void Free()			{ MemFree( m_p ); m_p = 0; }
	Type* operator->() const	{ return m_p; }
	bool operator!()		{ return !m_p; }

	Type** UnsafeAddress()		{ return &m_p; }

    protected:
	Type* m_p;

    private:
	SPMem<Type>& operator=( SPMem<Type> const& ); // Forbidden.
	SPMem( SPMem<Type> const& ); // Forbidden.

	void CASSERT_TRAITS()
	{
#ifdef USE_TYPE_TRAITS
	    CASSERT_INLINE( !type_traits::has_AddRef<Type>::value &&
			    !type_traits::has_Release<Type>::value );
#endif
	}
};

typedef SPMem<TCHAR>			SPTSZ;

/*
 * SPI<Interface, Cast> - Smart pointer for AddRef/Release interfaces.
 * SPQI<Interface, IID> - Adds QueryInterface helpers FQuery and HrQuery.
 *
 * Note that the following is possible:
 *
 *	SPI<SomeClass, IUnknown> spFoo = new SomeClass;
 */

#ifdef OS_WINDOWS

template<class IFace>
class PrivateRelease : public IFace
{
    private:
	// Force Release to be private to prevent "spfoo->Release()".
	STDMETHODIMP_( ULONG ) Release();
};

template<class IFace, class ICast = IFace>
class SPI
{
    public:
	SPI()				{ m_p = 0; CASSERT_TRAITS(); }
	//SPI( IFace *p )		{ m_p = p; if ( m_p ) m_p->AddRef(); }
	~SPI()				{ if( m_p ) static_cast<ICast*>( m_p )->Release(); }
	operator IFace*() const		{ return m_p; }
	PrivateRelease<IFace>* operator->() const
					{ return reinterpret_cast<PrivateRelease<IFace>*>( m_p ); }
	IFace** operator &()		{ Assert1( !m_p, "Non-empty %s as out param.", TypeIdName( SPI<IFace> ) ); return &m_p; }
	IFace* operator=( IFace* p )	{ Assert1( !m_p, "Non-empty %s in assignment.", TypeIdName( SPI<IFace> ) ); return m_p = p; }
	IFace* Transfer()		{ IFace* p = m_p; m_p = 0; return p; }
	IFace* Copy()			{ if ( m_p ) static_cast<ICast*>( m_p )->AddRef(); return m_p; }
	void Release()			{ Set( 0 ); }
	void Set( IFace* p )		{ ICast* pRelease = static_cast<ICast*>( m_p ); if( p ) static_cast<ICast*>( p )->AddRef(); m_p = p; if( pRelease ) pRelease->Release(); }
	bool operator!()		{ return ( m_p == NULL ); }

    protected:
	IFace* m_p;

    private:
	SPI<IFace>& operator=( SPI<IFace> const& sp ); // Forbidden.
	SPI( SPI<IFace> const& ); // Forbidden.

	void CASSERT_TRAITS()
	{
#ifdef USE_TYPE_TRAITS
	    CASSERT_INLINE( type_traits::has_AddRef<ICast>::value &&
			    type_traits::has_Release<ICast>::value );
#endif
	}
};

struct IUnknown;

template<class IFace, IID const& iid = __uuidof( IFace )>
class SPQI : public SPI<IFace>
{
    public:
	//SPQI( IFace* p )		{ m_p = p; if ( m_p ) m_p->AddRef(); }
	//SPQI( IUnknown* p )		{ m_p = 0; HrQuery( p ); }

	bool FQuery( IUnknown* punk )	{ return SUCCEEDED( HrQuery( punk ) ); }
	HRESULT HrQuery( IUnknown* punk )
					{ Assert1( !m_p, "Non-empty %s in HrQuery().", TypeIdName2( SPI<IFace, iid> ) ); return punk->QueryInterface( iid, ( void** )&m_p ); }

    private:
	SPQI<IFace, iid>& operator=( SPQI<IFace, iid> const& sp ); // Forbidden.
	SPQI( SPQI<IFace, iid> const& ); // Forbidden.
};

#define DeclareSPQI( TAG, IFace )\
	EXTERN_C const GUID CDECL IID_##IFace;\
	typedef SPQI<IFace, &IID_##IFace> SP##TAG;

#endif

