/*
 * StdMem.cpp - Memory allocator, with debug tracking.
 */

// vim: set noet:			// Hard tabs.
// vim: set ts=8:			// Tab width 8.
// vim: set sw=4:			// Shift (indent) width 4.
// vim: set cino={0s:			// Braces flush with outer statement.
// vim: set cino=^1s:			// First indent 8 (add 1 shift width).

#include "precomp.h"

#include "str.h"

#ifdef USE_MEMORY_TRACKING

/*
 * Memory tracking subsystem.
 */

#undef malloc
#undef realloc
#undef free

#ifdef ASSERT_WITH_STACK
#error MAX_STACK_STRING_LENGTH
#else
enum
{
	MAX_STACK_STRING_LENGTH = 1,
};
#endif

size_t const c_cbPreRun		= 4;
size_t const c_cbOverRun	= 4;
BYTE const s_bAllocFill		= 0xcd;
BYTE const s_bReallocFill	= 0xce;
BYTE const s_bDeadFill		= 0xdd;
BYTE const s_bMemTrackDeadFill	= 0xde;
BYTE const s_bMemTrackPreRun	= 0xab;	// Left chevron.
BYTE const s_bMemTrackOverRun	= 0xbb;	// Right chevron.

struct CRITICAL_SECTION_autoinit : public CRITICAL_SECTION
{
    public:
	CRITICAL_SECTION_autoinit() { InitializeCriticalSection( this ); }
};

static CRITICAL_SECTION_autoinit s_cs;
static DWORD s_cEntered = 0;
static DWORD s_idThread = 0;
static MemTracking* s_pHead = 0;
static DWORD s_iAlloc = 0;
static DWORD s_iAllocBreak = 0;

class AutoSynchronize
{
    public:
	AutoSynchronize()
	{
	    EnterCriticalSection( &s_cs );
	    s_cEntered++;
	    s_idThread = GetCurrentThreadId();
	}
	~AutoSynchronize()
	{
	    s_cEntered--;
	    if( !s_cEntered )
		s_idThread = 0;
	    LeaveCriticalSection( &s_cs );
	}
};

inline void AssertSynchronized() { Assert( s_idThread == GetCurrentThreadId() ); }

static DWORD
DbgNewAllocNumber()
{
	AssertSynchronized();

	s_iAlloc++;

	if( s_iAllocBreak && s_iAllocBreak == s_iAlloc )
	    __DbgBreak();

	return s_iAlloc;
}

char const*
DbgInspectMemory( void const* pv, ULONG cb )
{
	ULONG const cbMaxInspect = 32;
	static char const c_szHex[] = "0123456789ABCDEF";
	static char szHex[5 + ( 3 * cbMaxInspect ) + 3 + 1];
	static char szChr[1 + ( 3 * cbMaxInspect ) + 1 + 3 + 1];

	unsigned char const* pb = static_cast<unsigned char const*>( pv );
	char* pszHex = szHex;
	char* pszChr = szChr;
	char* pszEnd = 0;
	bool fTruncated = false;
	unsigned cFiltered = 0;
	ULONG cbMax;

	if( !cb )
	    return "";

	if( cb > cbMaxInspect )
	{
	    fTruncated = true;
	    cb = cbMaxInspect;
	}
	cbMax = cb;

	CchCopy( "HEX: ", pszHex, DimensionOf( szHex ) - ( pszHex - szHex ) );
	pszHex += 5;
	*( pszChr++ ) = '\"';

	while( cb-- )
	{
	    // Hex.
	    *( pszHex++ ) = c_szHex[*pb >> 4];
	    *( pszHex++ ) = c_szHex[*pb & 0x0f];
	    *( pszHex++ ) = ' ';

	    // Chr.
	    if( *pb >= 32 && *pb <= 126 )
	    {
		*( pszChr++ ) = *pb;
	    }
	    else
	    {
		*( pszChr++ ) = '.';
		cFiltered++;
	    }

#if 0
	    // Ignore trailing nulls in strings.
	    if( *pb )
		pszEnd = pszChr;
#else
	    // Ignore null termination of strings.
	    if( !cb && !*pb && ( cbMax == 1 || *( pb - 1 ) ) )
		pszEnd = pszChr - 1;
#endif

	    // Next byte.
	    pb++;
	}

	if( pszEnd )
	    pszChr = pszEnd;

	*( pszChr++ ) = '\"';
	*pszChr = 0;
	*pszHex = 0;

	if( fTruncated )
	{
	    Assert( CchLen( pszHex ) < DimensionOf( szHex ) - 3 );
	    Assert( CchLen( pszChr ) < DimensionOf( szChr ) - 3 );

	    CchCopy( "...", pszHex, DimensionOf( szHex ) - ( pszHex - szHex ) );
	    CchCopy( "...", pszChr, DimensionOf( szChr ) - ( pszChr - szChr ) );
	}

	return ( cFiltered > ( cbMax / 2 ) ) ? szHex : szChr;
}

inline BYTE*
BytePtr( void const* const pv )
{
	return static_cast<BYTE*>( const_cast<void*>( pv ) );
}

size_t
MemTracking::PadSize( size_t cbSize )
{
	Assert( cbSize );
	return sizeof( MemTracking ) + c_cbPreRun + cbSize + c_cbOverRun;
}

MemTracking*
MemTracking::FromPv( void const* const pv, char const* const pszFailureMessage )
{
	SUPPRESS_UNUSED_VAR( pszFailureMessage );

	AutoSynchronize as;

	MemTracking* const p = pv ? reinterpret_cast<MemTracking*>( BytePtr( pv ) - ( sizeof( MemTracking ) + c_cbPreRun ) ) : 0;
	bool fTraced;
	BYTE const* pVerify;
	ULONG cb;

	// Verify prerun.

	fTraced = false;
	pVerify = BytePtr( pv ) - c_cbPreRun;
	for( cb = c_cbPreRun; cb--; )
	{
	    if( pVerify[cb] != s_bMemTrackPreRun )
	    {
		// Panic first, to avoid crashing in the trace before we've
		// even alerted anyone about the prerun.
		Panic3( "Memory Prerun at 0x%8.8x == 0x%2.2x\n\tPrerun area: %s",
			pVerify + cb, pVerify[cb], DbgInspectMemory( pVerify, c_cbPreRun ) );

		if( !fTraced )
		{
		    fTraced = true;

		    char szStack[MAX_STACK_STRING_LENGTH];
		    szStack[0] = 0;
#ifdef ASSERT_WITH_STACK
		    if( p->m_rgStack[0] )
			p->GetStackString( RgC( szStack ) );
#endif

		    Trace(( "Prerun block:  #%u,  %u,  0x%p,  %u,  %s,  %s%s%s",
			    p->m_iNumber, p->m_cbSize, pv, p->m_cReAlloc, p->SzLabel(),
			    DbgInspectMemory( pv, p->m_cbSize ), szStack[0] ? ",  context:" : "", szStack ));
		}

		// Only scream once.
		break;
	    }
	}

	// Verify overrun.

	fTraced = false;
	pVerify = BytePtr( pv ) + p->m_cbSize;
	for( cb = 0; cb < c_cbOverRun; cb++ )
	{
	    if( pVerify[cb] != s_bMemTrackOverRun )
	    {
		Panic3( "Memory Overrun at 0x%p == 0x%2.2x\n\tOverrun area: %s",
			pVerify + cb, ( long )pVerify[cb], DbgInspectMemory( pVerify, c_cbOverRun ) );

		if( !fTraced )
		{
		    fTraced = true;

		    char szStack[MAX_STACK_STRING_LENGTH];
		    szStack[0] = 0;
#ifdef ASSERT_WITH_STACK
		    if( p->m_rgStack[0] )
			p->GetStackString( RgC( szStack ) );
#endif

		    Trace(( "Overrun block:  #%u,  %u,  0x%p,  %u,  %s,  %s%s%s",
			    p->m_iNumber, p->m_cbSize, pv, p->m_cReAlloc, p->SzLabel(),
			    DbgInspectMemory( pv, p->m_cbSize ), szStack[0] ? ",  context:" : "", szStack ));
		}

		// Only scream once.
		break;
	    }
	}

	return p;
}

void*
MemTracking::ToPv() const
{
	return BytePtr( this ) + sizeof( MemTracking ) + c_cbPreRun;
}

void
MemTracking::Fill( size_t const cbSize, bool const fDead )
{
	BYTE* const pb = BytePtr( ToPv() );

	if( !fDead )
	{
	    memset( pb - c_cbPreRun, s_bMemTrackPreRun, c_cbPreRun );
	    memset( pb + cbSize, s_bMemTrackOverRun, c_cbOverRun );
	}

	memset( pb, fDead ? s_bDeadFill : s_bAllocFill, cbSize );
}

void
MemTracking::Link()
{
	AssertSynchronized();

	m_pNext = s_pHead;
	if( s_pHead )
	    s_pHead->m_pPrev = this;
	s_pHead = this;
}

void
MemTracking::Unlink()
{
	AssertSynchronized();

	if( m_pNext )
	    m_pNext->m_pPrev = m_pPrev;

	if( m_pPrev )
	    m_pPrev->m_pNext = m_pNext;
	else
	    s_pHead = m_pNext;
}

void
MemTracking::MakeZombie()
{
	if( m_fFreeLabel )
	    rawMemFree( const_cast<char*>( m_pszLabel ) );

	memset( this, s_bMemTrackDeadFill, sizeof( MemTracking ) );
}

void
MemTracking::CopyFromAndFree( MemTracking* const p, size_t const cbSize )
{
	AssertSynchronized();

	// Copy from pv.

	memcpy( ToPv(), p->ToPv(), min( p->m_cbSize, cbSize ) );

	m_iNumber = p->m_iNumber;
	m_dwFlags = p->m_dwFlags;
	m_pszLabel = p->m_pszLabel;
	m_fFreeLabel = p->m_fFreeLabel;
	m_cReAlloc = p->m_cReAlloc;

	p->m_fFreeLabel = false;

	// Free.

	DbgFree( p->ToPv() );
}

char const*
MemTracking::SzLabel() const
{
	if( m_pszLabel )
	    return m_pszLabel;

	if( m_dwFlags & memObject )
	{
#ifdef USE_RTTI
	    return TypeIdName( *( Object* )ToPv() );
#else
	    return "(Object)";
#endif
	}

	if( m_dwFlags & memNew )
	    return "(new)";

	return "(MemAlloc)";
}

void
MemTracking::SetLabel( char const* const pszLabel, bool const fAutoFree )
{
	m_pszLabel = pszLabel;
	m_fFreeLabel = pszLabel && fAutoFree;
}

#ifdef ASSERT_WITH_STACK
void
MemTracking::GetStackString( char* pszBuf, int cchMax )
{
	if( !cchMax )
	    return;

	cchMax--;	// Reserve space for null terminator.

	char sz[cchMaxAssertStackLevelStringLen];
	for( size_t ii = 0; ii < DimensionOf( m_rgStack ) && m_rgStack[ii]; ii++ )
	{
	    char* psz = sz;

	    *psz = ' ';
	    psz++;

	    if( ii )
	    {
		strcpy( psz, "/ " );
		psz += 2;
	    }

	    psz += GetStringFromAddr( m_rgStack[ii], psz, fTrue );
	    *psz = 0;

	    CchCopy( sz, pszBuf, cchMax );

	    pszBuf += size_t( psz - sz );
	    cchMax -= size_t( psz - sz );

	    if( cchMax <= 1 )
		break;
	}

	*pszBuf = 0;	// Fill in null terminator (space was reserved at top).
}
#endif

void*
DbgAlloc( size_t cbSize, DWORD dwFlags )
{
	AutoSynchronize as;

	if( !( dwFlags & memNoSizeCheck ) )
	    Assert1( cbSize <= 16 * 1024, "DbgAlloc:  Is this a bogus size?  (%u bytes)", cbSize );

	MemTracking* const p = static_cast<MemTracking*>( rawMemAlloc( MemTracking::PadSize( cbSize ) ) );
	if( !p )
	    return p;

	// Fill values.

	memset( p, 0, sizeof( MemTracking ) );
	p->Fill( cbSize, false );

	// Memory block details.

	p->m_iNumber = DbgNewAllocNumber();
	p->m_cbSize = cbSize;
	p->m_dwFlags = dwFlags;

	//p->m_nLine = nLine;
	//p->m_pszFilename = pszFilename;	// Assume pointer to constant string.

#ifdef ASSERT_WITH_STACK
	UINT const cLevels = GetStackLevels( 1, DimensionOf( p->m_rgStack ), p->m_rgStack );
	if( cLevels < DimensionOf( p->m_rgStack ) )
	    p->m_rgStack[cLevels] = 0;	// Null terminate.
#endif

	// Link the block into the list.

	p->Link();

	// Return pointer to caller-data portion of the block.

	return p->ToPv();
}

void*
DbgReAlloc( void *pv, size_t cbSize, DWORD dwFlags )
{
	if( !cbSize )
	{
	    DbgFree( pv );
	    return 0;
	}

	if( !pv )
	    return DbgAlloc( cbSize, dwFlags );

	AutoSynchronize as;

	MemTracking* p;
	MemTracking* p2 = MemTracking::FromPv( pv );

	Assert( !( p2->m_dwFlags & memNew ) );
	Assert2( ( p2->m_dwFlags & ~memNoSizeCheck ) == ( dwFlags & ~memNoSizeCheck ), "DbgReAlloc:  Not allowed to change flags.  (old=0x%x, new=0x%x)", p2->m_dwFlags, dwFlags );

	p = static_cast<MemTracking*>( rawMemAlloc( MemTracking::PadSize( cbSize ) ) );
	if( !p )
	    return 0;

	Assert1( cbSize != p2->m_cbSize, "DbgReAlloc:  Why realloc when the size isn't changing?  (%u bytes)", cbSize );
	if( !( p2->m_dwFlags & memNoSizeCheck ) )
	    Assert1( cbSize <= 64 * 1024, "DbgReAlloc:  Is this a bogus size?  (%u bytes)", cbSize );

	// Fill values.

	memset( p, 0, sizeof( MemTracking ) );
	p->Fill( cbSize, false );

	// Copy from pv.

	p->CopyFromAndFree( p2, cbSize );

	// Update memory block details.

	p->m_cReAlloc++;
	p->m_cbSize = cbSize;

	//p->m_nLine = nLine;
	//p->m_pszFilename = pszFilename;	// Assume pointer to constant string.

#ifdef ASSERT_WITH_STACK
	UINT const cLevels = GetStackLevels( 1, DimensionOf( p->m_rgStack ), p->m_rgStack );
	if( cLevels < DimensionOf( p->m_rgStack ) )
	    p->m_rgStack[cLevels] = 0;	// Null terminate.
#endif

	// Link the block into the list.

	p->Link();

	// Return pointer to caller-data portion of the block.

	return p->ToPv();
}

void
DbgFree( void* pv )
{
	if( !pv )
	    return;

	AutoSynchronize as;

	MemTracking* p = MemTracking::FromPv( pv );

#if 0
	for( MemTracking* pFind = s_pHead; pFind && p != pFind; pFind = pFind->m_pNext ) {}
	if( !pFind )
	{
	    PanicSz( "DbgFree:  Freeing dead/bogus memory." );
	    return;
	}
#endif

	// Dead fill.

	p->Fill( p->m_cbSize, true );

	// Unlink the memory block from the list.

	p->Unlink();

	// More dead fill.

	memset( p, s_bMemTrackDeadFill, sizeof( MemTracking ) );

	// Free the memory.

	rawMemFree( p );
}

static int
ListLeaks()
{
	AutoSynchronize as;

	MemTracking* p;
	UINT nLeaks = 0;
	size_t cb = 0;
	char szStack[MAX_STACK_STRING_LENGTH];

	Trace(( "---- Checking for leaks ----" ));
	for( p = s_pHead; p; p = p->m_pNext, ++nLeaks )
	{
	    cb += p->m_cbSize;

	    szStack[0] = 0;
#ifdef ASSERT_WITH_STACK
	    if( p->m_rgStack[0] )
		p->GetStackString( RgC( szStack ) );
#endif

	    Trace(( "Leak:  #%d,  %u bytes (%u reallocs),  0x%p,  %s,  %s%s%s",
		    p->m_iNumber, p->m_cbSize, p->m_cReAlloc, p->ToPv(), p->SzLabel(),
		    DbgInspectMemory( p->ToPv(), p->m_cbSize ), szStack[0] ? ",  context:" : "", szStack ));
	}

	Trace(( "---- %d leaks, %d bytes total ----", nLeaks, cb ));
	return nLeaks;
}

void
DbgCheck()
{
	AutoSynchronize as;

	UINT const nLeaks = ListLeaks();

	if( nLeaks )
	{
	    Assert1( false, "%d leaks detected.", nLeaks );
	    ListLeaks();
	}
}

void
DbgSetLabel( void* pv, char const* pszLabel, bool fCopy )
{
	MemTracking* const p = MemTracking::FromPv( pv );

	if( fCopy )
	{
	    size_t const cchMax = CchLen( pszLabel ) + sizeof( *pszLabel );
	    char* const pszCopy = static_cast<char*>( rawMemAlloc( cchMax ) );
	    CchCopy( pszLabel, pszCopy, cchMax );
	    pszLabel = pszCopy;
	}

	p->SetLabel( pszLabel, fCopy );
}

#endif // USE_MEMORY_TRACKING

/*
 * New/delete operators.
 */

void* _cdecl
operator new( size_t cbSize )
{
	void* const pv = MemAlloc( cbSize _MEM_NEW );
	if( pv )
	    memset( pv, 0, cbSize );
	return pv;
}

void _cdecl
operator delete( void* pv )
{
	if( pv )
	    MemFree( pv );
}

