/*
 * StdMem.h - Memory allocator, with debug tracking.
 */

// vim: set noet:			// Hard tabs.
// vim: set ts=8:			// Tab width 8.
// vim: set sw=4:			// Shift (indent) width 4.
// vim: set cino={0s:			// Braces flush with outer statement.
// vim: set cino=^1s:			// First indent 8 (add 1 shift width).

#pragma once

/*
 * General allocation stuff.
 */

inline void* rawMemAlloc( size_t cb ) { return malloc( cb ); }
inline void* rawMemReAlloc( void* pv, size_t cb ) { return realloc( pv, cb ); }
inline void rawMemFree( void* pv ) { return free( pv ); }

#define malloc __use_MemAlloc_instead__
#define realloc __use_MemReAlloc_instead__
#define free __use_MemFree_instead__

void* _cdecl operator new( size_t cbSize );
void _cdecl operator delete( void* pv );

#ifndef __PLACEMENT_NEW_INLINE
#	define __PLACEMENT_NEW_INLINE
inline void* _cdecl operator new( size_t, void *pv ) { return pv; }
#	if defined( _MSC_VER )
#		if _MSC_VER >= 1200
inline void _cdecl operator delete( void *, void * ) { return; }
#		endif
#	endif
#endif

#ifdef USE_MEMORY_TRACKING

/*
 * Memory tracking subsystem.
 */

enum
{
#ifdef ASSERT_WITH_STACK
	MAX_STACK_DEPTH		= 12,
#endif

	memNew			= 0x00000001,
	memObject		= 0x00000002,
	memNoSizeCheck		= 0x00000004,
};

struct MemTracking
{
	static size_t	PadSize( size_t cbSize );
	static MemTracking* FromPv( const void* pv, char const* pszFailureMessage = 0 );
	void*		ToPv() const;
	void		Fill( size_t cbSize, bool fDead );
	void		Link();
	void		Unlink();
	void		MakeZombie();
	void		CopyFromAndFree( MemTracking* p, size_t cbSize );

	char const*	SzLabel() const;
	void		SetLabel( char const* pszLabel, bool fAutoFree );

#ifdef ASSERT_WITH_STACK
	void		GetStackString( char* pszBuf, int cchMax ) const;
#endif

	MemTracking*	m_pPrev;
	MemTracking*	m_pNext;

	DWORD		m_iNumber;
	DWORD		m_dwFlags;
	size_t		m_cbSize;
	DWORD		m_cReAlloc;

	//char const*	m_pszFilename;
	//ULONG		m_nLine;

#ifdef ASSERT_WITH_STACK
	DWORD_PTR	m_rgStack[MAX_STACK_DEPTH];	// Max or null terminated.
#endif

    private:
	char const*	m_pszLabel;			// Friendly name of allocation.
	unsigned	m_fFreeLabel : 1;		// Whether to free m_pszLabel.
};

#define _MEM_0			, 0
#define _MEM_NEW		, memNew
#define _MEM_OBJECT		, memObject
#define _MEM_NOSIZECHECK	, memNoSizeCheck

#define MemAlloc		DbgAlloc
#define MemReAlloc		DbgReAlloc
#define MemFree			DbgFree

void* DbgAlloc( size_t cbSize, DWORD dwFlags = 0 );
void* DbgReAlloc( void* pv, size_t cbSize, DWORD dwFlags = 0 );
void DbgFree( void* pv );
void DbgSetLabel( void* pv, char const* pszLabel, bool fCopy );
void DbgCheck();

#else // !USE_MEMORY_TRACKING

/*
 * Definitions when memory tracking is not enabled.
 */

#define _MEM_0
#define _MEM_NEW
#define _MEM_OBJECT
#define _MEM_NOSIZECHECK

#define MemAlloc		rawMemAlloc
#define MemReAlloc		rawMemReAlloc
#define MemFree			rawMemFree

#endif // !USE_MEMORY_TRACKING

/*
 * General purpose allocation helpers.
 */

struct StructFastNewOperator
{
	void* _cdecl operator new( size_t cbSize ) { return MemAlloc( cbSize _MEM_NEW ); }
	void operator delete( void* pv ) { MemFree( pv ); }
};

#define ZeroClassData() do { ZeroMemory( LPBYTE( this ) + sizeof( void* ), sizeof( *this ) - sizeof( void* ) ); } while( 0 )
#define ZeroClassDataUpTo( m ) do { ZeroMemory( LPBYTE( this ) + sizeof( void* ), OffsetOf( *this, m ) - sizeof( void* ) ); } while( 0 )

