/*
 * StoreList.cpp - list of stores.
 */

// vim: set noet:			// Hard tabs.
// vim: set ts=8:			// Tab width 8.
// vim: set sw=4:			// Shift (indent) width 4.
// vim: set cino={0s:			// Braces flush with outer statement.
// vim: set cino=^1s:			// First indent 8 (add 1 shift width).

#include "precomp.h"

#include "storelist.h"
#include "dbopen.h"

/*
 * StoreList - list of stores.
 */

static int
CmpRawStoreId( void const* pA, void const* pB )
{
	return ( static_cast<int>( *reinterpret_cast<Store const*>( pA ) ) -
		 static_cast<int>( *reinterpret_cast<Store const*>( pB ) ) );
}

Store const*
StoreList::Alloc( WORD c )
{
	Assert( !FBorrowed() );
	Assert( BYTE( c ) == c );

	if( c > m_cAlloc )
	{
	    Store* prg;

	    // Allocate some extra room.

	    if( m_cAlloc && c < m_cAlloc + 8 )
		c = m_cAlloc + 8;

	    // Allocate the new block.

	    prg = new Store[c];

	    if( !prg )
		return 0;

	    // Copy contents from the old block.

	    if( m_c )
	    {
		memmove( prg, m_prg, m_c * sizeof( *prg ) );
		delete [] m_prg;
	    }

	    // Set the new pointer.

	    m_prg = prg;
	    m_cAlloc = BYTE( c );
	}

	return m_prg;
}

bool
StoreList::Find( Store const id ) const
{
	Assert( !IsPhantomStore( id ) );

	for( WORD ii = 0; ii < m_c; ii++ )
	    if( id == m_prg[ii] )
		return true;

	return false;
}

bool
StoreList::Insert( Store const id )
{
	Assert( !FBorrowed() );
	Assert( !Find( id ) );

	if( !Alloc( m_c + 1 ) )
	    return false;

	// Just append to the end, as nothing specifically cares about the
	// order.

	m_prg[m_c++] = id;
	return true;
}

bool
StoreList::Remove( Store const id )
{
	Assert( !FBorrowed() );
	Assert( !IsPhantomStore( id ) );

	for( WORD ii = 0; ii < m_c; ii++ )
	    if( id == m_prg[ii] )
	    {
		memmove( &m_prg[ii], &m_prg[ii + 1], sizeof( m_prg[0] ) * ( m_c - ( ii + 1 ) ) );
		return true;
	    }

	return false;
}

void
StoreList::Borrow( WORD const c, Store* const prg )
{
	Assert( FPristine() );
	Assert( BYTE( c ) == c );

	m_c = BYTE( c );
	m_cAlloc = 0;
	m_prg = c ? prg : 0;
}

bool
StoreList::Copy( WORD const c, Store const* const prg )
{
	Assert( !Count() );
	Assert( !FBorrowed() );

	if( c )
	{
	    if( !Alloc( c ) )
		return false;

	    memmove( m_prg, prg, c * sizeof( *prg ) );
	    m_c = BYTE( c );
	}

	return true;
}

void
StoreList::Clear()
{
	if( FBorrowed() )
	    m_prg = 0;
	m_c = 0;
}

void
StoreList::Free()
{
	if( m_cAlloc )
	{
	    delete [] m_prg;
	    m_cAlloc = 0;
	}

	m_prg = 0;
	m_c = 0;
}

void
StoreList::Sort()
{
	Assert( !FBorrowed() );

	// RemoveDuplicates implicitly sorts the list.  And since we might as
	// well remove duplicates now anyway, we'll just let it run the show.

	RemoveDuplicates();
}

void
StoreList::SortWorker()
{
	Assert( !FBorrowed() );

	// Sort the list.

	qsort( m_prg, m_c, sizeof( *m_prg ), CmpRawStoreId );
}

XlateResult
StoreList::Translate( StoreTranslationInfo* pxi ) const
{
	WORD ii;
	XlateResult xr = xrAll;
	Store* pstore;
	Store* pstoreHare;

	Assert( !pxi->storesXlate.Count() );
	Assert( !pxi->storesAdd.Count() );

	if( !Count() )
	    goto LOut;

	if( Count() > BYTE( -1 ) )
	{
	    xr = xrFailure;
	    goto LOut;
	}

	// Allocate memory for translated stores.

	if( !pxi->storesXlate.Alloc( Count() ) ||
	    !pxi->storesAdd.Alloc( Count() ) )
	{
	    //$ todo: (chrisant) show an error message?
	    Panic();
	    xr = xrFailure;
	    goto LOut;
	}

	// Initialize the translated stores.

	pxi->storesXlate.Copy( *this );

	// Translate store id's.

	pstore = pstoreHare = const_cast<Store*>( pxi->storesXlate.Array() );

	for( ii = Count(); ii--; pstoreHare++ )
	{
	    TCHAR const* const pszStore = pxi->pdbFrom->GetStoreNameById( *pstoreHare );
	    Store const idTo = pxi->pdbTo->GetStoreIdByName( pszStore );

	    if( storeUnfiledStore == idTo )
	    {
		pxi->storesAdd.Insert( *pstoreHare );
		xr = xrSome;
	    }
	    else
	    {
		*pstore = idTo;
		pstore++;
	    }
	}

	pxi->storesXlate.m_c = BYTE( pstore - pxi->storesXlate.Array() );

#ifdef DEBUG
	switch( xr )
	{
	case xrSome:
	    Assert( pxi->storesAdd.Count() );
	    Assert( pxi->storesAdd.Count() + pxi->storesXlate.Count() == Count() );
	    break;
	case xrAll:
	    Assert( pxi->storesXlate.Count() == Count() );
	    Assert( !pxi->storesAdd.Count() );
	    break;
	}
#endif

LOut:
	return xr;
}

bool
StoreList::RemoveDuplicates()
{
	WORD const cOrig = m_c;
	Store* pStores;
	Store* pStoresHare;

	Assert( !FBorrowed() );

	// First sort the list, to make it easy to find duplicates.

	SortWorker();

	// Walk the list, with a slow pointer and a fast pointer.

	pStoresHare = pStores = m_prg;

	for( WORD ii = m_c; ii--; )
	{
	    if( ii && pStoresHare[0] == pStoresHare[1] )
	    {
		// Duplicate; discard it.

		//DbgMsg( ( 8, "Record %u\n-- Stores --\nDuplicate, %u", record, *pStoresHare ) );
		m_c--;
	    }
	    else
	    {
		// Unique; keep it.

		if( pStoresHare > pStores )
		    *pStores = *pStoresHare;

		pStores++;
	    }

	    pStoresHare++;
	}

	// Return non-zero if there were any duplicates.

	Assert( cOrig >= m_c );
	return cOrig != m_c;
}

#if 0
bool
StoreList::RepairList()
{
	WORD const cOrig = m_c;
	Store *pStores;
	Store *pStoresHare;

	// Remove duplicates.

	RemoveDuplicates();

	// Look for bogus entries, walking the list with slow/fast pointers.

	pStoresHare = pStores = m_prg;

	for( WORD ii = m_c; ii--; )
	{
	    if( *pStoresHare >= MAX_STORES ||
		IndexFromId_Slow( *pStoresHare ) == Store( -1 ) )
	    {
		// Bogus entry; discard it.

		//DbgMsg( ( 8, "Record %u\n-- Stores --\nBogus, %u", record, *pStoresHare ) );
		m_c--;
	    }
	    else
	    {
		// Valid; keep it.

		*pStores = *pStoresHare;
		pStores++;
	    }

	    pStoresHare++;
	}

	// Return non-zero if there were any duplicate or bogus entries.

	Assert( cOrig >= m_c );
	return cOrig != m_c;
}
#endif

void
StoreList::Swap( StoreList& other )
{
	BYTE const c = m_c;
	BYTE const cAlloc = m_cAlloc;
	Store* const prg = m_prg;

	m_c = other.m_c;
	m_cAlloc = other.m_cAlloc;
	m_prg = other.m_prg;

	other.m_c = c;
	other.m_cAlloc = cAlloc;
	other.m_prg = prg;
}

/*
 * PerStoreList - list of per-store data.
 */

static void
CopyLong( LONG* const _plDst, LONG const* const _plSrc )
{
	register BYTE* const pDst = reinterpret_cast<BYTE*>( _plDst );
	register BYTE const* const pSrc = reinterpret_cast<BYTE const*>( _plSrc );

	pDst[0] = pSrc[0];
	pDst[1] = pSrc[1];
	pDst[2] = pSrc[2];
	pDst[3] = pSrc[3];
}

static int
CmpStoreId( void const* pvFind, WORD indexCompare, void* pvCookie )
{
	return ( reinterpret_cast<int>( pvFind ) -
		 static_cast<int>( const_cast<PerStore const*>( static_cast<PerStore*>( pvCookie ) )[indexCompare].m_idStore ) );
}

static int
CmpRawPerStore( void const* pA, void const* pB )
{
	return ( static_cast<int>( static_cast<PerStore const*>( pA )->m_idStore ) -
		 static_cast<int>( static_cast<PerStore const*>( pB )->m_idStore ) );
}

PerStore const*
PerStoreList::Alloc( WORD c )
{
	Assert( !FBorrowed() );
	Assert( BYTE( c ) == c );

	if( c > m_cAlloc )
	{
	    PerStore* prg;

	    // Allocate some extra room.

	    if( c < m_cAlloc + 2 )
		c = m_cAlloc + 2;

	    // Allocate the new block.

	    prg = new PerStore[c];

	    if( !prg )
		return 0;

	    // Copy contents from the old block.

	    if( m_c )
	    {
		memmove( prg, m_prg, m_c * sizeof( *prg ) );
		delete [] m_prg;
	    }

	    // Set the new pointer.

	    m_prg = prg;
	    m_cAlloc = BYTE( c );
	}

	return m_prg;
}

PerStore*
PerStoreList::Find( Store const id, WORD* pindex ) const
{
	WORD ii;

	Assert( !IsPhantomStore( id ) );

	if( !pindex )
	    pindex = &ii;

	if( !FBinarySearch( reinterpret_cast<void const*>( id ),
			    const_cast<PerStore*>( Array() ),
			    m_c, pindex, CmpStoreId ) )
	    return 0;

	return &m_prg[*pindex];
}

PerStore*
PerStoreList::Insert( Store const id )
{
	Assert( !FBorrowed() );

	WORD ii;

	PerStore* const p = Find( id, &ii );

	if( p )
	    return p;

	if( !Alloc( m_c + 1 ) )
	    return 0;

	memmove( m_prg + ii + 1, m_prg + ii, ( m_c - ii ) * sizeof( *m_prg ) );

	ZeroMemory( m_prg + ii, sizeof( *m_prg ) );
	m_prg[ii].m_idStore = id;
	m_c++;

	return m_prg + ii;
}

void
PerStoreList::Remove( PerStore const* const p )
{
	Assert( !FBorrowed() );

	if( p )
	{
	    Assert( p >= m_prg );
	    Assert( p < m_prg + m_c );
	    Assert( !( ( reinterpret_cast<DWORD>( p ) - reinterpret_cast<DWORD>( m_prg ) ) % sizeof( *p ) ) );

	    memmove( const_cast<PerStore*>( p ), p + 1, ( m_c - ( p - m_prg ) - 1 ) * sizeof( *p ) );
	    m_c--;
	}
}

#if 0
void
PerStoreList::Borrow( WORD const c, PerStore* const prg )
{
	Assert( FPristine() );
	Assert( BYTE( c ) == c );

	m_c = BYTE( c );
	m_cAlloc = 0;
	m_prg = c ? prg : 0;
}
#endif

bool
PerStoreList::Copy( WORD const c, PerStore const* const prg )
{
	Assert( !Count() );
	Assert( !FBorrowed() );

	if( c )
	{
	    if( !Alloc( c ) )
		return false;

	    memmove( m_prg, prg, c * sizeof( *prg ) );
	    m_c = BYTE( c );
	}

	return true;
}

void
PerStoreList::Swap( PerStoreList& other )
{
	Assert( !FBorrowed() );

	BYTE const c = m_c;
	BYTE const cAlloc = m_cAlloc;
	PerStore* const prg = m_prg;

	m_c = other.m_c;
	m_cAlloc = other.m_cAlloc;
	m_prg = other.m_prg;

	other.m_c = c;
	other.m_cAlloc = cAlloc;
	other.m_prg = prg;
}

void
PerStoreList::Clear()
{
	if( FBorrowed() )
	    m_prg = 0;
	m_c = 0;
}

void
PerStoreList::Free()
{
	if( m_cAlloc )
	{
	    delete [] m_prg;
	    m_cAlloc = 0;
	}

	m_prg = 0;
	m_c = 0;
}

LONG
PerStoreList::GetPrice( Store const id ) const
{
	PerStore* p;
	WORD ii;
	LONG l = 0;
	LONG lPrice;

	if( Count() )
	{
	    if( IsPhantomStore( id ) )
	    {
		LONG const lBound = LONG_MAX;

		p = m_prg;
		l = lBound;

		for( ii = Count(); ii--; p++ )
		{
		    CopyLong( &lPrice, &p->m_lPrice );
		    if( lPrice && lPrice < l )
			l = lPrice;
		}

		if( l == lBound )
		    l = 0;
	    }
	    else
	    {
		p = Find( id );

		if( p )
		    CopyLong( &l, &p->m_lPrice );
	    }
	}

	return l;
}

Store
PerStoreList::GetLowestPriceInfo( LONG* const plPrice, Store const idCheck ) const
{
	Store id = storeUnfiledStore;
	LONG l = LONG_MAX;
	LONG lPrice;
	PerStore const* p = m_prg;

	for( WORD ii = Count(); ii--; p++ )
	{
	    CopyLong( &lPrice, &p->m_lPrice );

	    if( lPrice && lPrice <= l )
	    {
		// When idCheck is given, don't update id when the price is
		// equal and we already found idCheck.

		if( lPrice == l && idCheck != storeUnfiledStore && id == idCheck )
		    continue;

		l = lPrice;
		id = p->m_idStore;
	    }
	}

	if( LONG_MAX == l )
	    l = 0;
	if( plPrice )
	    *plPrice = l;

	return id;
}

TCHAR const*
PerStoreList::GetAisle( Store const idStore ) const
{
	if( Count() && !IsPhantomStore( idStore ) )
	{
	    PerStore const* const p = Find( idStore );

	    if( p )
		return p->m_szAisle;
	}

	return TEXT( "" );
}

void
PerStoreList::Sort()
{
	Assert( !FBorrowed() );

	qsort( m_prg, m_c, sizeof( *m_prg ), CmpRawPerStore );
}

XlateResult
PerStoreList::Translate( StoreTranslationInfo* const pxi ) const
{
	WORD ii;
	XlateResult xr = xrAll;
	PerStore* pperstore;
	PerStore* pperstoreHare;

	Assert( !pxi->perstoreXlate.Count() );
	Assert( !pxi->storesAddPer.Count() );

	if( !Count() )
	    goto LOut;

	if( Count() > BYTE( -1 ) )
	{
	    xr = xrFailure;
	    goto LOut;
	}

	// Allocate memory for translated stores.

	if( !pxi->perstoreXlate.Alloc( Count() ) ||
	    !pxi->storesAddPer.Alloc( Count() ) )
	{
	    //$ todo: (chrisant) show an error message?
	    Panic();
	    xr = xrFailure;
	    goto LOut;
	}

	// Initialize the translated stores.

	pxi->perstoreXlate.Copy( *this );

	// Translate per-store id's.

	pperstore = pperstoreHare = const_cast<PerStore*>( pxi->perstoreXlate.Array() );

	for( ii = Count(); ii--; pperstoreHare++ )
	{
	    TCHAR const* const pszStore = pxi->pdbFrom->GetStoreNameById( pperstoreHare->m_idStore );
	    Store const idTo = pxi->pdbTo->GetStoreIdByName( pszStore );

	    if( storeUnfiledStore == idTo )
	    {
		pxi->storesAddPer.Insert( pperstoreHare->m_idStore );
		xr = xrSome;
	    }
	    else
	    {
		*pperstore = *pperstoreHare;
		pperstore->m_idStore = idTo;
		pperstore++;
	    }
	}

	pxi->perstoreXlate.m_c = BYTE( pperstore - pxi->perstoreXlate.Array() );

#ifdef DEBUG
	switch( xr )
	{
	case xrSome:
	    Assert( pxi->storesAddPer.Count() );
	    Assert( pxi->storesAddPer.Count() + pxi->perstoreXlate.Count() == Count() );
	    break;
	case xrAll:
	    Assert( pxi->perstoreXlate.Count() == Count() );
	    Assert( !pxi->storesAddPer.Count() );
	    break;
	}
#endif

LOut:
	return xr;
}

#if 0
bool
PerStoreList::RepairList()
{
	Assert( !FBorrowed() );

	bool fDamaged = false;
	WORD ii;
	WORD c = 0;
	PerStore* pPerStore;
	PerStore* pPerStoreHare;

	// Ensure the per-store info is sorted (very important).

	pPerStore = m_prg;

	for( ii = m_c; ii-- > 1; pPerStore++ )
	    if( pPerStore[0].m_idStore >= pPerStore[1].m_idStore )
	    {
		//DbgMsg( ( 8, "Record %u\n-- PerStore --\nSorted wrong", record ) );
		fDamaged = true;
		Sort();
		break;
	    }

	// Look for bad entries.

	pPerStoreHare = pPerStore = m_prg;

	for( ii = m_c; ii--; )
	{
	    // Duplicate?

	    if( ii && pPerStoreHare[0].m_idStore == pPerStoreHare[1].m_idStore )
	    {
		//DbgMsg( ( 8, "Record %u\n-- PerStore --\nDuplicate, %u", record, pPerStoreHare->m_idStore ) );
		goto LDamaged;
	    }

	    // Bogus entry?

	    if( pPerStoreHare->m_idStore >= MAX_STORES ||
		IndexFromId_Slow( pPerStoreHare->m_idStore ) == Store( -1 ) )
	    {
		//DbgMsg( ( 8, "Record %u\n-- PerStore --\nBogus, %u", record, pPerStoreHare->m_idStore ) );
LDamaged:
		fDamaged = true;
		m_c--;
		goto LNextPerStore;
	    }

	    // Keep it.

	    if( pPerStoreHare > pPerStore )
		MemMove( pPerStore, pPerStoreHare, sizeof( *pPerStore ) );

	    pPerStore++;

LNextPerStore:
	    pPerStoreHare++;
	}

	return fDamaged;
}
#endif

/*
 * AutoOrderList - list of per-store AutoOrder data.
 */

static int
CmpStoreId_AutoOrder( void const* pvFind, WORD indexCompare, void* pvCookie )
{
	return ( reinterpret_cast<int>( pvFind ) -
		 static_cast<int>( const_cast<AutoOrder const*>( static_cast<AutoOrder*>( pvCookie ) )[indexCompare].m_idStore ) );
}

static int
CmpRawAutoOrder( void const* pA, void const* pB )
{
	return ( static_cast<int>( static_cast<AutoOrder const*>( pA )->m_idStore ) -
		 static_cast<int>( static_cast<AutoOrder const*>( pB )->m_idStore ) );
}

AutoOrder const*
AutoOrderList::Alloc( WORD c )
{
	Assert( !FBorrowed() );
	Assert( BYTE( c ) == c );

	if( c > m_cAlloc )
	{
	    AutoOrder* prg;

	    // Allocate some extra room.

	    if( c < m_cAlloc + 2 )
		c = m_cAlloc + 2;

	    // Allocate the new block.

	    prg = new AutoOrder[c];

	    if( !prg )
		return 0;

	    // Copy contents from the old block.

	    if( m_c )
	    {
		memmove( prg, m_prg, m_c * sizeof( *prg ) );
		delete [] m_prg;
	    }

	    // Set the new pointer.

	    m_prg = prg;
	    m_cAlloc = BYTE( c );
	}

	return m_prg;
}

bool
AutoOrderList::AllocForXfer( WORD c )
{
	Assert( FPristine() );
	Assert( c > 0 );
	Assert( BYTE( c ) == c );

	// Clever optimization, allocate room for an extra entry to avoid
	// realloc simply from later adding one store.

	if( !Alloc( c + 1 ) )
	    return false;

	Assert( m_cAlloc > c );

	m_c = BYTE( c );

	return true;
}

AutoOrder*
AutoOrderList::Find( Store const id, WORD* pindex ) const
{
	Assert( id != storeUnfiledStore );

	WORD ii;
	if( !pindex )
	    pindex = &ii;

	if( id == storeAllStores )
	{
	    *pindex = 0;
	    if( Count() < 1 )
		return 0;
	    return const_cast<AutoOrder*>( Array() + 0 );
	}
	else
	{
	    *pindex = 1;
	    if( Count() < 2 )
		return 0;
	}

	Assert( !IsPhantomStore( id ) );

	bool const fFound = FBinarySearch( reinterpret_cast<void const*>( id ),
					   const_cast<AutoOrder*>( Array() + 1 ),
					   m_c - 1, pindex, CmpStoreId_AutoOrder );
	( *pindex ) += 1;
	return fFound ? &m_prg[*pindex] : 0;
}

AutoOrder*
AutoOrderList::Insert( Store const id )
{
	Assert( !FBorrowed() );
	Assert( id != storeUnfiledStore );

	WORD ii;

	AutoOrder* const p = Find( id, &ii );

	if( p )
	    return p;

	BYTE const cOld = m_c;
	BYTE cNew = 0;
	bool fShiftToMakeRoom = false;
	if( id == storeAllStores )
	{
	    Assert( ii == 0 );
	    Assert( Count() < 1 );
	    cNew = BYTE( ii + 1 );
	}
	else
	{
	    Assert( ii >= 1 );
	    cNew = Max( cOld, BYTE( 1 ) ) + 1;
	    fShiftToMakeRoom = ( cOld > 0 );
	}

	Assert( cNew );
	Assert( ii < cNew );

	if( !Alloc( cNew ) )
	    return 0;

	if( cOld < 1 && cNew >= 1 )
	    ZeroMemory( m_prg + 0, sizeof( *m_prg ) );
	if( fShiftToMakeRoom )
	    memmove( m_prg + ii + 1, m_prg + ii, ( m_c - ii ) * sizeof( *m_prg ) );
	ZeroMemory( m_prg + ii, sizeof( *m_prg ) );

	m_prg[ii].m_idStore = id;
	m_c = cNew;

	return m_prg + ii;
}

void
AutoOrderList::Remove( AutoOrder const* const p )
{
	Assert( !FBorrowed() );

	if( p )
	{
	    Assert( p >= m_prg + 1 );	// Plus 1 because the first entry is for phantom stores, and once it exists it cannot be deleted.
	    Assert( p < m_prg + m_c );
	    Assert( !( ( reinterpret_cast<DWORD>( p ) - reinterpret_cast<DWORD>( m_prg ) ) % sizeof( *p ) ) );

	    memmove( const_cast<AutoOrder*>( p ), p + 1, ( m_c - ( p - m_prg ) - 1 ) * sizeof( *p ) );
	    m_c--;
	}
}

void
AutoOrderList::Swap( AutoOrderList& other )
{
	Assert( !FBorrowed() );

	BYTE const c = m_c;
	BYTE const cAlloc = m_cAlloc;
	AutoOrder* const prg = m_prg;

	m_c = other.m_c;
	m_cAlloc = other.m_cAlloc;
	m_prg = other.m_prg;

	other.m_c = c;
	other.m_cAlloc = cAlloc;
	other.m_prg = prg;
}

void
AutoOrderList::Clear()
{
	if( FBorrowed() )
	    m_prg = 0;
	m_c = 0;
}

void
AutoOrderList::Free()
{
	if( m_cAlloc )
	{
	    delete [] m_prg;
	    m_cAlloc = 0;
	}

	m_prg = 0;
	m_c = 0;
}

DWORD
AutoOrderList::GetAutoOrder( Store const id ) const
{
	AutoOrder const* p = Find( id );
	return p ? p->m_ulOrder : 0;
}

void
AutoOrderList::Sort()
{
	Assert( !FBorrowed() );

	if( m_c > 1 )
	    qsort( m_prg + 1, m_c - 1, sizeof( *m_prg ), CmpRawAutoOrder );
}

#ifdef DEBUG
AutoOrderListValidity
AutoOrderList::TestValid() const
{
	if( m_c > 2 )
	{
	    for( BYTE ii = 2; ii < m_c - 1; ii++ )
	    {
		if( m_prg[ii].m_idStore < m_prg[ii - 1].m_idStore )
		    return aolUnsorted;
	    }
	}

	if( m_c > 0 )
	{
	    if( m_prg[0].m_idStore != 0 )
		return aolInvalidStore_AllStores;

	    bool rgF[c_cMaxStores] = {};
	    for( BYTE ii = 1; ii < m_c; ii++ )
	    {
		if( m_prg[ii].m_idStore >= c_cMaxStores )
		    return aolInvalidStore_OutOfBounds;
		if( rgF[m_prg[ii].m_idStore] )
		    return aolDuplicates;
		rgF[m_prg[ii].m_idStore] = true;
	    }
	}

	return aolValid;
}
#endif

