/*
 * storeshs2.cpp - Stores functions.
 */

// vim: set noet:			// Hard tabs.
// vim: set ts=8:			// Tab width 8.
// vim: set sw=4:			// Shift (indent) width 4.
// vim: set cino={0s:			// Braces flush with outer statement.
// vim: set cino=^1s:			// First indent 8 (add 1 shift width).

#include "precomp.h"

#include "BitTable.h"
#include "storelist.h"
#include "item.h"
#include "palmlayer.h"

#include "globalhs2.h"
#include "columns.h"
#include "data1x.h"
#include "storeshs2.h"
#include "db.h"
#include "pdbhs2.h"

CASSERT( MAX_STORES == c_cMaxStores );

extern void* GetMemberPtr( void* pvRecord, BYTE tag );

inline void
__itoa_s( int x, char* pszOut, size_t cchMax, int radix )
{
#if defined( OS_WINDOWSMOBILE )
	SUPPRESS_UNUSED_VAR( cchMax );
	_itoa( x, pszOut, radix );
#else
	_itoa_s( x, pszOut, cchMax, radix );
#endif
}

/*
 * Stores Forms.
 */

static Boolean s_fDirty;

#if 0
static int
CmpDupStores( const void *pvFind, UInt16 index, UInt32 dwCookie )
{
	return StrCompare( ( PCSZ )pvFind, StoreFromIndex( index )->m_szStore );
}
#endif

void 
DbOpenHs2::PutMasterStoreList()
{
	Assert( m_pStores );

	// If the stores aren't actually dirty, there's nothing to do.

	if( !m_fStoresDirty )
	    return;

	// If we don't have a list and we don't need a list, there's nothing
	// to do.

	unsigned record;
	MemHandle h;
	MasterStoreList* pStores;
	bool const fHaveList = FindMetadataRecord( m_pdb, tagMasterStoreList, &record, &h, &pStores );
	bool const fNeedList = !!m_pStores->m_cStores;

	if( fHaveList || fNeedList )
	{
	    // If there is no list and there needs to be one, create the
	    // record.

	    if( !fHaveList && fNeedList )
	    {
		h = DmNewRecord( m_pdb, &record, sizeof( *pStores ) );

		if( h )
		{
		    m_cMetadataRecords++;

		    Assert( m_cMetadataRecords <= 2 );

		    pStores = ( MasterStoreList* )MemHandleLock( h );

		    if( m_pbits )
			m_pbits->NewRecord( record, false );
		}
		else
		    Panic1( "PutMasterStoreList\n  DmNewRecord failed (%d)", m_pdb->DmGetLastErr() );
	    }

	    // If we have a pointer to the list record, update it.

	    if( pStores )
	    {
		Assert( MemHandleSize( h ) == sizeof( *pStores ) );

		memcpy( pStores, m_pStores, sizeof( *pStores ) );

		MemHandleUnlock( h );
		DmReleaseRecord( m_pdb, record, true );
	    }

	    // If there is a list record but there doesn't need to be, remove
	    // the record.

	    if( fHaveList && !fNeedList )
	    {
		m_cMetadataRecords--;

		DmRemoveRecord( m_pdb, record );

		if( m_pbits )
		    m_pbits->RemoveRecord( record );
	    }

	    SetDirty();
	}

	m_fStoresDirty = false;
}

void
DbOpenHs2::GetMasterStoreList()
{
	Assert( !m_pStores );

	unsigned record;
	MemHandle h;
	MasterStoreList *pStores = 0;

	m_pStores = new MasterStoreList;

	if( FindMetadataRecord( m_pdb, tagMasterStoreList, &record, &h, &pStores ) )
	{
	    Assert( MemHandleSize( h ) == sizeof( *m_pStores ) );

	    memcpy( m_pStores, pStores, sizeof( *m_pStores ) );

	    MemHandleUnlock( h );
	    DmReleaseRecord( m_pdb, record, false );
	}
}

void
DbOpenHs2::ZapStore( UInt8 const id )
{
	MemHandle		h;
	UInt8*			pRec;
	UInt8			cStores;
	Boolean			fDirty;
	UInt16			attr;

	Assert( this );
	Assert( m_pdb );
	Assert( allStores != id );
	Assert( noStore != id );

	// Loop thru all records in database, removing all references to the
	// specified store id.

	for( unsigned dmRecord = m_pdb->DmNumRecords(); dmRecord-- > m_cMetadataRecords; )
	{
	    Boolean fUnpackAndZap = false;

	    m_pdb->DmRecordInfo( dmRecord, &attr, 0, 0 );
	    if( attr & ( dmRecAttrDelete|dmRecAttrBusy ) )
	    {
#ifdef DEBUG
		h = m_pdb->DmQueryRecord( dmRecord );
		pRec = ( UInt8* )MemHandleLock( h );
		Assert1( tagMasterStoreList == *pRec, "dmRecord %u is busy, but is not the master store list!", dmRecord );
		MemHandleUnlock( h );
#endif
		continue;
	    }

	    fDirty = false;

	    h = m_pdb->DmGetRecord( dmRecord );
	    Assert( h );
	    pRec = ( UInt8* )MemHandleLock( h );
	    if( tagMasterStoreList != *pRec && tagCategoryList != *pRec )
	    {
		Assert( tagFlags == *pRec );

		// Figure out whether to unpack the record.  If the store list
		// does not mention the specified store, and the record has
		// neither per-store data nor auto order data (which is also
		// per-store) then the record does not need to be unpacked.

		if( !fUnpackAndZap )
		{
		    UInt8 const* pStores = static_cast<UInt8 const*>( GetMemberPtr( pRec, tagStores ) );
		    if( pStores )
		    {
			cStores = *pStores;
			pStores++;

			while( cStores-- )
			{
			    if( id == *pStores )
			    {
				fUnpackAndZap = true;
				break;
			    }
			    pStores++;
			}
		    }
		}

		if( !fUnpackAndZap )
		{
		    fUnpackAndZap = ( GetMemberPtr( pRec, tagPerStore ) ||
				      GetMemberPtr( pRec, tagPerStoreEx ) ||
				      GetMemberPtr( pRec, tagAutoOrder ) );
		}
	    }

	    MemHandleUnlock( h );
	    m_pdb->DmReleaseRecord( dmRecord, fDirty );

	    if( fUnpackAndZap )
	    {
		const unsigned record = dmRecord - m_cMetadataRecords;
		SPNew<Item> spitem = UnPackRecord( record );
		if( spitem->ZapStoreData( id ) )
		    PackRecord( record, spitem.UnsafeAddress(), packNoSort|packNoDelete|packNoUndo, 0 );
	    }
	}

	//DbgMsg(( 20, "ZapStore took %d clock ticks", TimGetTicks() - ticks ));
}

MasterStoreList::MasterStoreList()
	: m_tag( tagMasterStoreList )
	, m_cMax( MAX_STORES )
	, m_reserved( 0 )
{
#ifdef DEBUG
	BYTE const* pb = reinterpret_cast<BYTE const*>( this );
	for( unsigned ii = 2; ii < sizeof( *this ); ii++ )
	    Assert( !pb[ii] );
#endif
}

UInt8
MasterStoreList::IdFromIndex( UInt8 i ) const
{
	Assert( this );
	Assert( i < m_cStores );
	return m_rgIndexToId[i];
}

UInt8
MasterStoreList::IndexFromId_Slow( UInt8 id ) const
{
	UInt8 c;
	UInt8 const* pid;

	Assert( this );

	c = m_cStores;
	pid = m_rgIndexToId;
	while( c-- )
	{
	    if( *pid == id )
		return ( UInt8 )( pid - m_rgIndexToId );
	    pid++;
	}

	Panic1( "BAD: store id %d not found!", id );
	return UInt8( -1 );
}

StoreDef*
MasterStoreList::StoreFromIndex( UInt8 i )
{
	Assert( this );
	Assert( i < m_cStores );
	return &m_rgStore[m_rgIndexToId[i]];
}

void
DbOpenHs2::IncStore( Int8 n )
{
	Assert( this );
	Assert( m_pStores );
	Assert( SameSize( n, m_pStores->m_cStores ) );

	m_pStores->m_cStores = m_pStores->m_cStores + n;
}

int
DbOpenHs2::CompareStores( MemPtr pA, MemPtr pB, Int32 other )
{
	Assert( other );

	DbOpenHs2* const pThis = reinterpret_cast<DbOpenHs2*>( other );

	Assert( pThis );
	Assert( pThis->m_pStores );

	return StrCaselessCompareSort( pThis->m_pStores->m_rgStore[*( UInt8* )pA].m_szStore,
				       pThis->m_pStores->m_rgStore[*( UInt8* )pB].m_szStore );
}

void
DbOpenHs2::SortStores( bool fDirty )
{
	UInt8 rgid[MAX_STORES];

	Assert( this );
	Assert( m_pStores );
	Assert( DimensionOf( rgid ) == m_pStores->m_cMax );
	AssertSameSize( m_pStores->m_rgIndexToId[0], rgid[0] );

	MemMove( rgid, m_pStores->m_rgIndexToId, sizeof( m_pStores->m_rgIndexToId ) );

	if( m_pStores->m_cStores )
	{
	    SysInsertionSort( rgid, m_pStores->m_cStores,
			      sizeof( rgid[0] ), CompareStores, ( Int32 )this );
	}

	DmWrite( m_pStores, OffsetOf( *m_pStores, m_rgIndexToId ),
		 rgid, sizeof( m_pStores->m_rgIndexToId ) );

	if( fDirty )
	    SetStoresDirty();
}

void
DbOpenHs2::FillInStore( UInt8 id, PCSZ psz, Boolean fNew )
{
	StoreDef *pstore;

	Assert( m_pStores );
	Assert( id >= 0 && id < m_pStores->m_cMax );
	AssertSameSize( m_pStores->m_rgIndexToId[0], id );

	// Fill in new store info.
	if( fNew )
	{
	    DmWrite( m_pStores, ( UInt32 )&m_pStores->m_rgIndexToId[m_pStores->m_cStores] - ( UInt32 )m_pStores, &id, sizeof( id ) );
	    IncStore( 1 );
	}
	pstore = &m_pStores->m_rgStore[id];
	if( fNew )
	    DmSet( m_pStores, ( UInt8* )pstore - ( UInt8* )m_pStores, sizeof( *pstore ), 0 );
	DmStrCopy( m_pStores, ( UInt8* )pstore->m_szStore - ( UInt8* )m_pStores, psz );

	// Resort stores.
	SortStores();
}

UInt8
DbOpenHs2::HashStoreId( PCSZ psz )
{
	UInt8 idNew = 0;

	// Hash to find new store index.
	if( *psz >= 'a' && *psz <= 'z' )
	    idNew = 18 + 3 * ( *psz - 'a' );
	else if( *psz >= 'A' && *psz <= 'Z' )
	    idNew = 18 + 3 * ( *psz - 'A' );
	while( *( m_pStores->m_rgStore[idNew].m_szStore ) )
	{
	    idNew++;
	    idNew %= MAX_STORES;
	}
	return idNew;
}

UInt8
DbOpenHs2::StoreNewRaw( PCSZ psz, UInt8 id )
{
	UInt8 idNew;

	Assert( m_pStores );

	// Get new store id.
	idNew = ( id == noStore ) ? HashStoreId( psz ) : id;

	// Fill in store info.
	FillInStore( idNew, psz, true );
	return idNew;
}

/*
 * Bold Counts.
 */

void
DbOpenHs2::UpdateCounts( StoreList *pstores, Int16 delta )
{
	UInt16 c;

	Assert( this );
	Assert( m_pdb );		// Otherwise m_prefs is not filled in!
	Assert( FImplies( pstores->Count(), m_pStores ) );
	Assert( delta == 1 || delta == -1 );

	m_prefs.add_cBoldAll( delta );
	c = pstores->Count();
	if( c )
	{
	    Assert( pstores->Count() < m_pStores->m_cMax );
	    while( c-- )
	    {
		m_pStores->m_rgStore[pstores->operator[]( c )].add_m_cBold( delta );
	    }
	}
	else
	    m_prefs.add_cBoldNo( delta );

	SetStoresDirty();
	WriteAppInfo();
}

void
DbOpenHs2::ResetStoreItemCounts( bool fProgress )
{
	UInt16		record;
	MemHandle	h;
	UInt8*		pRec;
	UInt8		cStores;
	UInt16		attr;
	UInt16		flags;
	UInt16		cBoldAll = 0;
	UInt16		cBoldNo = 0;
	SPNew<MasterStoreList> spmslBefore;
	SPNew<MasterStoreList> spmsl;
	UInt16			cProgress;

	ASSERTGLOBALS;
	Assert( this );
	Assert( m_pdb );

	// Get a copy of the master store list, and work on that.

	spmsl = new MasterStoreList;
	spmslBefore = new MasterStoreList;

	memcpy( spmslBefore, m_pStores, sizeof( *spmslBefore ) );
	memcpy( spmsl, spmslBefore, sizeof( *spmslBefore ) );

	for( cStores = spmsl->m_cMax; cStores--; )
	    spmsl->m_rgStore[cStores].set_m_cBold( 0 );

	// Loop thru all records in database.

	record = UInt16( DmNumRecords( m_pdb ) );

	ScopedProgressHolder sph;
	if( fProgress )
	    sph.Create();
	sph.Init( record, PROGRESS_FORCEVISIBLE );
	cProgress = 0;

	while( record-- )
	{
	    sph.Update( cProgress++ );

	    DmRecordInfo( m_pdb, record, &attr, 0, 0 );
	    if( attr & dmRecAttrDelete )
		continue;

	    h = DmQueryRecord( m_pdb, record );
	    Assert( h );
	    pRec = ( UInt8* )MemHandleLock( h );
	    if( tagMasterStoreList == *pRec || tagCategoryList == *pRec )
		goto Next;

	    Assert( !( attr & dmRecAttrBusy ) );
	    Assert( tagFlags == *pRec );
	    pRec++;
	    CASSERT_( SameSize( flags, ( ( ItemPtr )0 )->m_flags ) );
	    flags = ( pRec[0] << 8 ) | pRec[1];
	    if( ( flags & ( flagNeed|flagPurchased ) ) != flagNeed )
		goto Next;

	    cBoldAll++;
	    if( !( flags & flagHasStore ) )
	    {
		cBoldNo++;
		goto Next;
	    }

	    Assert( *( pRec + 2 ) == tagStores );
	    pRec += 4;
	    for( cStores = *( pRec - 1 ); cStores--; pRec++ )
		spmsl->m_rgStore[*pRec].add_m_cBold( 1 );

Next:
	    MemHandleUnlock( h );
	}

	DbgMsg(( 10, TEXT( "all = %d\nnone = %d" ), cBoldAll, cBoldNo ));

	sph.End();

	// Only try to write to the database if the counts were wrong.

	if( memcmp( spmsl, spmslBefore, sizeof( *spmsl ) ) != 0 ||
	    cBoldAll != m_prefs.get_cBoldAll() ||
	    cBoldNo != m_prefs.get_cBoldNo() )
	{
	    memcpy( m_pStores, spmsl, sizeof( *spmsl ) );
	    SetStoresDirty();

	    m_prefs.set_cBoldAll( cBoldAll );
	    m_prefs.set_cBoldNo( cBoldNo );
	    WriteAppInfo();
	}
}

/*
 * Add Stores.
 */

bool
FAddStores( DbOpen* pdbFrom, DbOpen* pdbTo,
	    StoreList const* pstoresAdd, StoreList const* pstoresAddPer )
{
	UInt16 ii;
	UInt16 c;
	bool fSuccess = false;
	StoreList storesMerged;
	const Store *prg;

	// Allocate memory for merged list.
	if( !storesMerged.Alloc( pstoresAdd->Count() + pstoresAddPer->Count() ) )
	    //$ review: ( chrisant ) show an error message?
	    goto LError;

	// Merge lists of stores.
	storesMerged.Copy( *pstoresAdd );
	for( ii = 0; ii < pstoresAddPer->Count(); ii++ )
	    storesMerged.Insert( pstoresAddPer->operator[]( ii ) );
	storesMerged.RemoveDuplicates();

	// See if there's room for the indicated stores.
	if( pdbTo->GetStoresCount() + storesMerged.Count() > MAX_STORES )
	{
	    //$ API: (chrisant) how to make this info available?
#if 0
	    Error( errCantAddStores );
#endif
	    goto LError;
	}

	// Add the prg list of stores from popenFrom to popenTo.
	DbgMsg(( 10, TEXT( "adding %d stores" ), storesMerged.Count() ));
	prg = storesMerged.Array();
	for( c = storesMerged.Count(); c--; prg++ )
	{
	    LPCTSTR pszNew = pdbFrom->GetStoreNameById( *prg );
	    pdbTo->StoreNew( pszNew );
	}

	fSuccess = true;

LError:
	return fSuccess;
}

/*
 * PerStoreList - list of per-store data.
 */

static void
CopyLongHs2( Int32* const _plDst, Int32 const* const _plSrc )
{
	register BYTE* const pDst = reinterpret_cast<BYTE*>( _plDst );
	register BYTE const* const pSrc = reinterpret_cast<BYTE const*>( _plSrc );

	pDst[0] = pSrc[0];
	pDst[1] = pSrc[1];
	pDst[2] = pSrc[2];
	pDst[3] = pSrc[3];
}

static int
CmpStoreIdHs2( void const* pvFind, WORD indexCompare, void* pvCookie )
{
	return ( reinterpret_cast<int>( pvFind ) -
		 static_cast<int>( const_cast<PerStore2 const*>( static_cast<PerStore2*>( pvCookie ) )[indexCompare].m_idStore ) );
}

PerStoreHs2List::PerStoreHs2List( PerStoreList const& list )
	: m_c( 0 )
	, m_cAlloc( 0 )
	, m_prg( 0 )
{
	if( !list.Count() )
	    return;

	Alloc( list.Count() );

	Assert( m_prg );

	if( !m_prg )
	    return;

	PerStore2* p = m_prg;

	for( UInt16 ii = 0; ii < list.Count(); ii++, p++ )
	{
	    p->m_idStore = list[ii].m_idStore;
	    p->m_lPrice = list[ii].m_lPrice;

	    size_t const cch = TCharToNarrowString( list[ii].m_szAisle, RgC( p->m_szAisle ) );

	    Assert( cch < DimensionOf( p->m_szAisle ) );
	    SUPPRESS_UNUSED_VAR( cch );

	    m_c++;
	}

	Assert( m_c == m_cAlloc );
}

PerStore2 const*
PerStoreHs2List::Alloc( UInt16 c )
{
	Assert( UInt8( c ) == c );

	if( c > m_cAlloc )
	{
	    PerStore2* prg;

	    // Allocate some extra room.

	    if( m_cAlloc && c < m_cAlloc + 2 )
		c = m_cAlloc + 2;

	    // Allocate the new block.

	    prg = new PerStore2[c];

	    if( !prg )
		return 0;

	    // Copy contents from the old block.

	    if( m_c )
	    {
		memmove( prg, m_prg, m_c * sizeof( *prg ) );
		delete [] m_prg;
	    }

	    // Set the new pointer.

	    m_prg = prg;
	    m_cAlloc = UInt8( c );
	}

	return m_prg;
}

bool
PerStoreHs2List::Copy( UInt16 c, PerStore0 const* const prg )
{
	Assert( !Count() );

	if( c )
	{
	    if( !Alloc( c ) )
		return false;

	    m_c = UInt8( c );

	    while( c-- )
	    {
		m_prg[c].m_idStore = prg[c].m_idStore;

		m_prg[c].m_szAisle[0] = 0;
		if( prg[c].m_aisle )
		    if( prg[c].m_aisle <= MAX_NUMERIC_AISLES )
		    {
			__itoa_s( prg[c].m_aisle, RgC( m_prg[c].m_szAisle ), 10 );
		    }
		    else if( prg[c].m_aisle <= MAX_AISLES )
		    {
			m_prg[c].m_szAisle[0] = 'A' + ( prg[c].m_aisle - ( MAX_NUMERIC_AISLES + 1 ) );
			m_prg[c].m_szAisle[1] = 0;
		    }

		CASSERT( sizeof( prg[0].m_lPrice ) == sizeof( m_prg[0].m_lPrice ) );
		memcpy( &m_prg[c].m_lPrice, &prg[c].m_lPrice, sizeof( prg[c].m_lPrice ) );
	    }
	}

	return true;
}

bool
PerStoreHs2List::Copy( UInt16 const c, PerStore2 const* const prg )
{
	Assert( !Count() );

	if( c )
	{
	    if( !Alloc( c ) )
		return false;

	    memmove( m_prg, prg, c * sizeof( *prg ) );
	    m_c = UInt8( c );
	}

	return true;
}

void
PerStoreHs2List::Clear()
{
	m_c = 0;
}

void
PerStoreHs2List::Free()
{
	if( m_cAlloc )
	{
	    delete [] m_prg;
	    m_cAlloc = 0;
	}

	m_prg = 0;
	m_c = 0;
}

Int32
PerStoreHs2List::GetPrice( Store const id ) const
{
	PerStore2* p;
	UInt16 ii;
	Int32 l = 0;
	Int32 lPrice;

	if( Count() )
	{
	    if( id == allStores || id == noStore )
	    {
		Int32 const lBound = LONG_MAX;

		p = m_prg;
		l = lBound;

		for( ii = Count(); ii--; p++ )
		{
		    CopyLongHs2( &lPrice, &p->m_lPrice );
		    if( lPrice && lPrice < l )
			l = lPrice;
		}

		if( l == lBound )
		    l = 0;
	    }
	    else
	    {
		p = Find( id );

		if( p )
		    CopyLongHs2( &l, &p->m_lPrice );
	    }
	}

	return l;
}

Store
PerStoreHs2List::GetLowestPriceInfo( Int32* const plPrice, Store const idCheck ) const
{
	Store id = storeUnfiledStore;
	Int32 l = LONG_MAX;
	Int32 lPrice;
	PerStore2 const* p = m_prg;

	for( UInt16 ii = Count(); ii--; p++ )
	{
	    CopyLongHs2( &lPrice, &p->m_lPrice );

	    if( lPrice && lPrice <= l )
	    {
		// When idCheck is given, don't update id when the price is
		// equal and we already found idCheck.

		if( lPrice == l && idCheck != noStore && id == idCheck )
		    continue;

		l = lPrice;
		id = p->m_idStore;
	    }
	}

	if( LONG_MAX == l )
	    l = 0;
	if( plPrice )
	    *plPrice = l;

	return id;
}

Char const*
PerStoreHs2List::GetAisle( Store const idStore ) const
{
	if( Count() && idStore != allStores && idStore != noStore )
	{
	    PerStore2 const* const p = Find( idStore );

	    if( p )
		return p->m_szAisle;
	}

	return "";
}

PerStore2*
PerStoreHs2List::Find( Store const id, UInt16* pindex ) const
{
	UInt16 ii;

	Assert( id != noStore );
	Assert( id != allStores );

	if( !pindex )
	    pindex = &ii;

	if( !FBinarySearch( reinterpret_cast<void const*>( id ),
			    const_cast<PerStore2*>( Array() ),
			    m_c, pindex, CmpStoreIdHs2 ) )
	    return 0;

	return &m_prg[*pindex];
}

void
PerStoreHs2List::MacAdjust()
{
	for( UInt16 ii = m_c; ii--; )
	    CopyMacDWord( &m_prg[ii].m_lPrice, &m_prg[ii].m_lPrice );
}

void
PerStoreHs2List::UpConvert( PerStoreList* const pList )
{
	Assert( pList->FPristine() );

	if( !Count() )
	    return;

	PerStore* const prgNew = new PerStore[Count()];

	Assert( prgNew );

	if( !prgNew )
	    return;

	PerStore* p = prgNew;

	for( UInt16 ii = 0; ii < Count(); ii++, p++ )
	{
	    p->m_idStore = m_prg[ii].m_idStore;
	    p->m_lPrice = m_prg[ii].m_lPrice;

	    size_t const cch = NarrowToTCharString( m_prg[ii].m_szAisle, RgC( p->m_szAisle ) );

	    Assert( cch < DimensionOf( p->m_szAisle ) );
	    SUPPRESS_UNUSED_VAR( cch );
	}

	pList->Copy( Count(), prgNew );

	delete [] prgNew;
}

