/*
 * Str.h - Simplistic string class.
 */

#pragma once

// vim: set noet:			// Hard tabs.
// vim: set ts=8:			// Tab width 8.
// vim: set sw=4:			// Shift (indent) width 4.
// vim: set cino={0s:			// Braces flush with outer statement.
// vim: set cino=^1s:			// First indent 8 (add 1 shift width).

#ifdef OS_WINDOWS
inline size_t CchLen( char const* psz ) { return strlen( psz ); }
inline int LStrCmp( char const* pszA, char const* pszB ) { return lstrcmpA( pszA, pszB ); }
inline int LStrCmpI( char const* pszA, char const* pszB ) { return lstrcmpiA( pszA, pszB ); }
#ifdef UNICODE
inline size_t CchLen( WCHAR const* psz ) { return lstrlenW( psz ); }
inline int LStrCmp( WCHAR const* pszA, WCHAR const* pszB ) { return lstrcmpW( pszA, pszB ); }
inline int LStrCmpI( WCHAR const* pszA, WCHAR const* pszB ) { return lstrcmpiW( pszA, pszB ); }
#endif
#else
inline size_t CchLen( char const* psz ) { return strlen( psz ); }
inline int LStrCmp( char const* pszA, char const* pszB ) { return strcmp( pszA, pszB ); }
inline int LStrCmpI( char const* pszA, char const* pszB ) { return _stricmp( pszA, pszB ); }
#ifdef UNICODE
inline size_t CchLen( WCHAR const* psz ) { return wcslen( psz ); }
inline int LStrCmp( WCHAR const* pszA, WCHAR const* pszB ) { return wcscmp( pszA, pszB ); }
inline int LStrCmpI( WCHAR const* pszA, WCHAR const* pszB ) { return _wcsicmp( pszA, pszB ); }
#endif
#endif

template <class T>
size_t
CchCopy( T const* pszFrom, T* pszTo, size_t cchMax )
{
	size_t cchCopied = 0;

	if( cchMax )
	{
	    while( --cchMax && *pszFrom )
	    {
		*pszTo = *pszFrom;

		pszTo++;
		pszFrom++;
		cchCopied++;
	    }

	    *pszTo = 0;
	}

	return cchCopied;
}

#ifdef UNICODE

size_t WideToNarrowString( WCHAR const* pwzFrom, char* pszTo, size_t cchMax );
size_t NarrowToWideString( char const* pszFrom, WCHAR* pwzTo, size_t cchMax );

#if defined( OS_WINDOWS )

inline size_t
WideToNarrow( WCHAR const* pwzFrom, size_t cchFrom, char* pszTo, size_t cchMax )
{
	Assert( size_t( -1 ) != cchFrom );
	return WideCharToMultiByte( CP_ACP, 0, pwzFrom, cchFrom, pszTo, cchMax, 0, 0 );
}

inline size_t
NarrowToWide( char const* pszFrom, size_t cchFrom, WCHAR* pwzTo, size_t cchMax )
{
	Assert( size_t( -1 ) != cchFrom );
	return MultiByteToWideChar( CP_ACP, 0, pszFrom, cchFrom, pwzTo, cchMax );
}

#elif defined( OS_CRT )

size_t WideToNarrow( WCHAR const* pwzFrom, size_t cchFrom, char* pszTo, size_t cchMax );
size_t NarrowToWide( char const* pszFrom, size_t cchFrom, WCHAR* pwzTo, size_t cchMax );

#endif

inline size_t TCharToNarrow( TCHAR const* pszFrom, size_t cchFrom, char* pszTo, size_t cchMax ) { return WideToNarrow( pszFrom, cchFrom, pszTo, cchMax ); }
inline size_t NarrowToTChar( char const* pszFrom, size_t cchFrom, TCHAR* pszTo, size_t cchMax ) { return NarrowToWide( pszFrom, cchFrom, pszTo, cchMax ); }
inline size_t TCharToNarrowString( TCHAR const* pszFrom, char* pszTo, size_t cchMax ) { return WideToNarrowString( pszFrom, pszTo, cchMax ); }
inline size_t NarrowToTCharString( char const* pszFrom, TCHAR* pszTo, size_t cchMax ) { return NarrowToWideString( pszFrom, pszTo, cchMax ); }

#else // !UNICODE

size_t CharToCharN( char const* pszFrom, size_t cchFrom, char* pszTo, size_t cchMax );
inline size_t TCharToNarrow( TCHAR const* pszFrom, size_t cchFrom, char* pszTo, size_t cchMax ) { return CharToCharN( pszFrom, cchFrom, pszTo, cchMax ); }
inline size_t NarrowToTChar( char const* pszFrom, size_t cchFrom, TCHAR* pszTo, size_t cchMax ) { return CharToCharN( pszFrom, cchFrom, pszTo, cchMax ); }
inline size_t TCharToNarrowString( TCHAR const* pszFrom, char* pszTo, size_t cchMax ) { return CchCopy( pszFrom, pszTo, cchMax ); }
inline size_t NarrowToTCharString( char const* pszFrom, TCHAR* pszTo, size_t cchMax ) { return CchCopy( pszFrom, pszTo, cchMax ); }

#endif // !UNICODE

enum AppendQuoteFlags
{
	AQF_ESCAPEQUOTES		= 0x01,
	AQF_QUOTESPACES			= 0x02,
	AQF_SPREADSHEETCOMPATIBILITY	= 0x04,
};

inline AppendQuoteFlags
operator|( AppendQuoteFlags const& a, AppendQuoteFlags const& b )
{
	return AppendQuoteFlags( int( a ) | int( b ) );
}

inline AppendQuoteFlags
operator&( AppendQuoteFlags const& a, AppendQuoteFlags const& b )
{
	return AppendQuoteFlags( int( a ) & int( b ) );
}

template <class T>
class Str
{
    public:
			Str<T>() { Init(); }
			~Str<T>();

	T const*	Text() const { return m_psz; }
	size_t		Length() const { return m_cchLen; }
	size_t		AllocSize() const { return m_cchAlloc; }

	void		Clear();
	void		Free();
	void		Init();

	bool		Equal( T const* psz ) const { return !LStrCmp( psz, Text() ); }
	bool		Equal( Str<T> const* s ) const;
	bool		Equal( Str<T> const& s ) const { return Equal( &s ); }
	bool		EqualI( T const* psz ) const { return !LStrCmpI( psz, Text() ); }
	bool		EqualI( Str<T> const* s ) const;
	bool		EqualI( Str<T> const& s ) const { return EqualI( &s ); }

	void		SetEnd( T const* pszEnd );
	void		SetLength( size_t cch );

	void		Set( T const* psz, size_t cch = -1 );
	void		Set( Str<T> const& s ) { Set( s.Text(), s.Length() ); }

	T*		Alloc( size_t cch );

	void		SetAt( T const* psz, T ch );

	void		Append( TCHAR ch );
	void		Append( T const* psz, size_t cch = -1 );
	void		Append( Str<T> const& s ) { Append( s.Text(), s.Length() ); }
	void		AppendQuote( T const* psz, AppendQuoteFlags flags = AQF_ESCAPEQUOTES );

	void		PrintfV( T const* pszFormat, va_list args );
	void		Printf( T const* pszFormat, ... );
#ifdef OS_WINDOWS
	void		Printf( UINT idsFormat, ... );
	size_t		LoadString( UINT ids );
	size_t		GetWindowText( HWND hwnd );
#endif

	void		TrimLeft();
	void		TrimRight();
	void		Trim();

	void		ToLower();

	void		Unquote();

	void		SanitizeFileName();

	size_t		GetCbXferPalm() const;

	bool		GetToken( Str<T>& out );

	void		Attach( T* psz, size_t cchLen, size_t cchAlloc );
	void		Swap( Str<T>& other );

	void		FormatForSingleLine();

	void		ConvertToCrtLineEndings();
	void		ConvertToNativeLineEndings();
	void		AppendNativeLineEnding();

    protected:
	T*		AllocInternal( size_t cch );

    protected:
	T*		m_psz;
	size_t		m_cchLen;
	size_t		m_cchAlloc;

	static T const* const s_pszEmpty;
};

template <class T>
T const* const Str<T>::s_pszEmpty = reinterpret_cast<T const*>( reinterpret_cast<void const*>( L"" ) );

template <class T>
Str<T>::~Str()
{
	if( s_pszEmpty != m_psz )
	    MemFree( m_psz );
}

template <class T>
void
Str<T>::Clear()
{
	if( m_cchLen )
	{
	    m_psz[0] = 0;
	    m_cchLen = 0;
	}
}

template <class T>
void
Str<T>::Free()
{
	if( s_pszEmpty != m_psz )
	    MemFree( m_psz );

	Init();
}

template <class T>
void
Str<T>::Init()
{
	m_psz = const_cast<T*>( s_pszEmpty );
	m_cchLen = 0;
	m_cchAlloc = 0;
}

template <class T>
bool
Str<T>::Equal( Str<T> const* s ) const
{
	Assert( s );

	if( Length() != s->Length() )
	    return false;

	if( LStrCmp( Text(), s->Text() ) )
	    return false;

	return true;
}

template <class T>
bool
Str<T>::EqualI( Str<T> const* s ) const
{
	Assert( s );

	if( Length() != s->Length() )
	    return false;

	if( LStrCmpI( Text(), s->Text() ) )
	    return false;

	return true;
}

template <class T>
void
Str<T>::SetEnd( T const* pszEnd )
{
	SetLength( pszEnd - m_psz );
}

template <class T>
void
Str<T>::SetLength( size_t cch )
{
	Assert( s_pszEmpty != m_psz );
	Assert( cch >= 0 );
	Assert( cch <= m_cchLen );

	m_cchLen = cch;
	m_psz[m_cchLen] = 0;
}

template <class T>
void
Str<T>::Set( T const* psz, size_t cch )
{
	Clear();
	Append( psz, cch );
}

template <class T>
void
Str<T>::SetAt( T const* psz, T ch )
{
	Assert( psz );
	Assert( s_pszEmpty != m_psz );
	Assert( m_psz <= psz );
	Assert( psz < m_psz + m_cchLen );

	*const_cast<T*>( psz ) = ch;
}

template <class T>
void
Str<T>::Append( TCHAR ch )
{
	Assert( ch );

	memcpy( AllocInternal( 1 ), &ch, sizeof( ch ) );
	m_cchLen++;
	m_psz[m_cchLen] = 0;
}

template <class T>
void
Str<T>::Append( T const* psz, size_t cch )
{
	if( psz )
	{
	    if( int( cch ) < 0 )
		cch = CchLen( psz );
	    memcpy( AllocInternal( cch ), psz, cch * sizeof( *psz ) );
	    m_cchLen += cch;
	    m_psz[m_cchLen] = 0;
	}
}

template <class T>
T*
Str<T>::Alloc( size_t cch )
{
	m_cchLen += cch;
	return AllocInternal( 0 ) - cch;
}

template <class T>
T*
Str<T>::AllocInternal( size_t cch )
{
	cch++;				// make room for null terminator

	if( m_cchLen + cch > m_cchAlloc )
	{
	    // Alloc exact length the first time, but use geometric growth if
	    // there is already a buffer allocated.

	    size_t cchAlloc = m_cchLen + cch;

	    if( m_cchAlloc > 31 )
		cchAlloc += ( cchAlloc / 3 ) * 2;
	    else
		cchAlloc *= 2;

	    // Allocate the new buffer.

	    m_psz = ( T* )MemReAlloc( ( s_pszEmpty == m_psz ) ? 0 : m_psz, cchAlloc * sizeof( T ) );
	    m_cchAlloc = cchAlloc;
	}

	return m_psz + m_cchLen;
}

inline char const* IsQuotingNeeded( char const* const psz, AppendQuoteFlags const flags ) { return strpbrk( psz, ( flags & AQF_QUOTESPACES ) ? " \r\n,\"" : "\r\n,\"" ); }
inline char const* FindQuote( char const* const psz ) { return strchr( psz, '\"' ); }
#ifdef UNICODE
inline WCHAR const* IsQuotingNeeded( WCHAR const* const psz, AppendQuoteFlags const flags ) { return wcspbrk( psz, ( flags & AQF_QUOTESPACES ) ? L" \r\n,\"" : L"\r\n,\"" ); }
inline WCHAR const* FindQuote( WCHAR const* const psz ) { return wcschr( psz, '\"' ); }
#endif

template <class T>
void
Str<T>::AppendQuote( T const* psz, AppendQuoteFlags const flags )
{
	if( psz )
	{
	    if( IsQuotingNeeded( psz, flags ) )
	    {
		T const szQuote[2] = { '\"', '\0' };

		Append( szQuote );

		if( flags & AQF_SPREADSHEETCOMPATIBILITY )
		    Append( '\'' );

		if( flags & AQF_ESCAPEQUOTES )
		{
		    while( T const* pszBrk = FindQuote( psz ) )
		    {
			Append( psz, pszBrk - psz );

			switch( *pszBrk )
			{
			case '\"':
			    Append( szQuote );	// One double quote.
			    Append( szQuote );	// Second double quote.
			    break;
			default:
			    Panic2( "unhandled char '%c' (0x%02x)", *pszBrk, *pszBrk );
			    break;
			}

			psz = pszBrk + 1;
		    }
		}

		Append( psz );
		Append( szQuote );
	    }
	    else
	    {
		Append( psz );
	    }
	}
}

inline int __vsprintf_s( char* pszBuffer, size_t cchBuffer, char const* pszFormat, va_list args ) { return vsprintf_s( pszBuffer, cchBuffer, pszFormat, args ); }
#ifdef UNICODE
inline int __vsprintf_s( WCHAR* pszBuffer, size_t cchBuffer, WCHAR const* pszFormat, va_list args ) { return vswprintf_s( pszBuffer, cchBuffer, pszFormat, args ); }
#endif

template <class T>
void
Str<T>::PrintfV( T const* pszFormat, va_list args )
{
	size_t cch = size_t( -1 );
	size_t cchAlloc = 100;	// start with 100

	while( int( cch ) < 0 )
	{
	    // Subtract 1 from the max character count to ensure room for null
	    // terminator; see MSDN for idiosyncracy of the snprintf family of
	    // functions.

	    cch = __vsprintf_s( AllocInternal( cchAlloc ), cchAlloc - 1, pszFormat, args );
	    cchAlloc *= 2;
	}

	m_cchLen += cch;
}

template <class T>
void
Str<T>::Printf( T const* pszFormat, ... )
{
	va_list args;
	va_start( args, pszFormat );

	PrintfV( pszFormat, args );

	va_end( args );
}

#ifdef OS_WINDOWS

template <class T>
void
Str<T>::Printf( UINT idsFormat, ... )
{
	va_list args;
	va_start( args, idsFormat );

	TCHAR szFormat[100];
	LPCTSTR pszFormat = szFormat;
	StrT sFormat;

	if( LoadStringU( idsFormat, szFormat, DimensionOf( szFormat ) ) + 1 == DimensionOf( szFormat ) )
	{
	    sFormat.LoadString( idsFormat );
	    pszFormat = sFormat.Text();
	}

	PrintfV( pszFormat, args );

	va_end( args );
}

template <class T>
size_t
Str<T>::LoadString( UINT ids )
{
	for( ;; )
	{
	    Clear();

	    if( m_cchAlloc )
		m_cchLen = LoadStringU( ids, m_psz, m_cchAlloc );

	    // When loading a resource string in Windows, the only way to
	    // detect that the buffer may not have been large enough is to
	    // check if the string exactly fills the buffer.

	    if( m_cchLen + 1 < m_cchAlloc )
		break;

	    AllocInternal( Max( size_t( 31 ), m_cchAlloc ) );
	}

	return m_cchLen;
}

template <class T>
size_t
Str<T>::GetWindowText( HWND hwnd )
{
	int const cch = ::GetWindowTextLength( hwnd );

	Clear();

	if( cch )
	{
	    AllocInternal( cch );
	    m_cchLen = ::GetWindowText( hwnd, m_psz, m_cchAlloc );
	}

	return m_cchLen;
}

#endif

template <class T>
void
Str<T>::Unquote()
{
	if( s_pszEmpty != m_psz && m_psz[0] == '\"' )
	{
	    T* pszHare = m_psz + 1;
	    T* pszTortoise = m_psz;

	    while( pszHare < m_psz + m_cchLen )
	    {
		*pszTortoise = *pszHare;

		if( pszHare[0] == '\"' )
		{
		    if( pszHare[1] == '\"' )
			pszHare++;
		    else
			// intentionally truncate the string when we find a
			// double quote by itself (vs two double quotes).
			break;
		}

		pszHare++;
		pszTortoise++;
	    }

	    SetEnd( pszTortoise );
	}
}

template <class T>
void
Str<T>::SanitizeFileName()
{
	for( LPCTSTR psz = Text(); *psz; psz++ )
	{
	    // Since the general case will be that the character is a letter
	    // or number, first test for that.

	    if( isalnum( *psz ) )
		continue;

	    // Otherwise compare against a list of invalid characters and
	    // replace them with underscores.

	    if( *psz == '?' ||
		*psz == '*' ||
		*psz == '<' ||
		*psz == '>' ||
		*psz == '|' ||
		*psz == ':' ||
		*psz == '/' ||
		*psz == '\\' )
		SetAt( psz, '_' );
	}
}

template <class T>
size_t
Str<T>::GetCbXferPalm() const
{
	size_t cbSize;

#ifdef UNICODE
	cbSize = WideToNarrow( Text(), Length(), 0, 0 );
#else
	cbSize = Length();
#endif

#ifdef OS_CRLF
	for( const T* pszWalk = m_psz; *pszWalk; pszWalk++ )
	    if( '\r' == *pszWalk )
		cbSize--;
#endif

	return cbSize;
}

template <class T>
void
Str<T>::TrimLeft()
{
	if( Length() )
	{
	    T const* psz = Text();

	    while( *psz && isspace( *psz ) )
		psz++;

	    if( psz > Text() )
	    {
		m_cchLen -= int( psz - m_psz );
		memmove( m_psz, psz, ( m_cchLen + 1 ) * sizeof( *psz ) );
	    }
	}
}

template <class T>
void
Str<T>::TrimRight()
{
	if( Length() )
	{
	    T const* pszLast = Text() + Length() - 1;

	    while( pszLast >= Text() && isspace( *pszLast ) )
		pszLast--;

	    SetEnd( pszLast + 1 );
	}
}

template <class T>
void
Str<T>::Trim()
{
	TrimRight();
	TrimLeft();
}

template <class T>
void
Str<T>::ToLower()
{
	for( T* pszWalk = m_psz; pszWalk < m_psz + m_cchLen; pszWalk++ )
	    *pszWalk = _totlower( *pszWalk );
}

template <class T>
bool
Str<T>::GetToken( Str<T> &out )
{
	// Clear out token string.

	out.Clear();

	if( s_pszEmpty == m_psz )
	    return 0;

	// Parse the string for a comma separated token.  Newline and EOS are
	// also considered separators.

	bool fQuote = false;
	T const* psz = Text();

	for( ; *psz; psz++ )
	{
	    if( '\"' == *psz )
	    {
		if( !fQuote && !out.Length() )
		{
		    fQuote = true;
		    continue;
		}
		else if( fQuote )
		{
		    fQuote = false;

		    psz++;

		    if( !*psz )
		    {
			// End of string.
			break;
		    }

		    if( ',' == *psz || '\n' == *psz )
		    {
			// End of token.
			psz++;
			break;
		    }

		    // Two double quotes in a row are interpreted as one
		    // double quote and do not leave quote mode.

		    if( '\"' == *psz )
		    {
			fQuote = true;
			out.Append( psz, 1 );
			continue;
		    }

		    // For poorly terminated tokens, skip up to the next
		    // delimeter.

		    while( *psz )
		    {
			if( '\n' == *psz || ',' == *psz )
			{
			    psz++;
			    break;
			}
		    }

		    break;
		}
	    }

	    if( !fQuote && ( '\n' == *psz || ',' == *psz ) )
	    {
		psz++;
		break;
	    }

	    out.Append( psz, 1 );
	}

	// Chop off the parsed token from the beginning of the string.  This
	// is where an iterator like StrRef would come in handy.

	m_cchLen -= int( psz - m_psz );
	memmove( m_psz, psz, ( m_cchLen + 1 ) * sizeof( *psz ) );

	return !!out.Length();
}

template <class T>
void
Str<T>::Attach( T* psz, size_t cchLen, size_t cchAlloc )
{
	if( s_pszEmpty != m_psz )
	    MemFree( m_psz );

	if( psz )
	{
	    Assert( cchLen );
	    Assert( cchAlloc );
	    Assert( cchAlloc > cchLen );

	    m_cchLen = cchLen;
	    m_cchAlloc = cchAlloc;

	    m_psz = psz;
	    m_psz[cchLen] = 0;
	}
	else
	{
	    m_cchLen = 0;
	    m_cchAlloc = 0;
	    m_psz = const_cast<T*>( s_pszEmpty );
	}
}

template <class T>
void
Str<T>::Swap( Str<T>& other )
{
	T* const psz = m_psz;
	m_psz = other.m_psz;
	other.m_psz = psz;

	size_t const cchLen = m_cchLen;
	m_cchLen = other.m_cchLen;
	other.m_cchLen = cchLen;

	size_t const cchAlloc = m_cchAlloc;
	m_cchAlloc = other.m_cchAlloc;
	other.m_cchAlloc = cchAlloc;
}

inline bool
IsReplaceableWhitespace( TCHAR const ch )
{
	return ch == ' ' || ch == '\n' || ch == '\r' || ch == '\t';
}

template <class T>
void
Str<T>::FormatForSingleLine()
{
	if( Length() )
	{
	    LPTSTR pszAt = m_psz;
	    LPCTSTR pszWalk = m_psz;

	    while( *pszWalk )
	    {
		if( pszWalk[0] == ' ' && pszWalk[1] == ' ' )
		{
		    *( pszAt++ ) = *( pszWalk++ );
		    *( pszAt++ ) = *( pszWalk++ );
		    goto LSkipRun;
		}
		else if( IsReplaceableWhitespace( *pszWalk ) )
		{
		    // Replace a run of whitespace with a single space.

		    *pszAt = ' ';
		    pszAt++;

		    // Skip forward past the run of whitespace.

LSkipRun:
		    while( IsReplaceableWhitespace( *pszWalk ) )
			++pszWalk;
		}
		else
		{
		    // Keep other characters as-is.

		    if( pszAt != pszWalk )
			*pszAt = *pszWalk;
		    pszAt++;
		    pszWalk++;
		}
	    }

	    SetEnd( pszAt );

	    Trim();
	}
}

template <class T>
void
Str<T>::ConvertToCrtLineEndings()
{
#ifndef OS_LF
	if( Length() )
	{
	    T* pszAt = m_psz;
	    T const* pszWalk = m_psz;

	    while( *pszWalk )
	    {
#if defined( OS_CRLF )
		if( *pszWalk == '\r' )
		{
		    // Don't copy.
		}
#elif defined( OS_CR )
		if( *pszWalk == '\r' )
		{
		    *pszAt = '\n';
		    pszAt++;
		}
#endif
		else
		{
		    if( pszAt != pszWalk )
			*pszAt = *pszWalk;
		    pszAt++;
		}

		pszWalk++;
	    }

	    SetEnd( pszAt );
	}
#endif
}

template <class T>
void
Str<T>::ConvertToNativeLineEndings()
{
#ifndef OS_LF
	if( Length() )
	{
#ifdef OS_CRLF
	    Str<T> tmp;
#else
	    T* pszAt = m_psz;
#endif
	    T const* pszWalk = m_psz;

	    while( *pszWalk )
	    {
#if defined( OS_CRLF )
		T const* const pszBegin = pszWalk;
		while( *pszWalk && *pszWalk != '\n' )
		    pszWalk++;
		if( pszWalk - pszBegin )
		    tmp.Append( pszBegin, pszWalk - pszBegin );
		if( *pszWalk == '\n' )
		{
		    tmp.Append( '\r' );
		    tmp.Append( '\n' );
		    pszWalk++;
		}
#elif defined( OS_CR )
		if( *pszWalk == '\r' )
		    *pszAt = '\n';
		pszAt++;
		pszWalk++;
#endif
	    }

#ifdef OS_CRLF
	    Swap( tmp );
#else
	    SetEnd( pszAt );
#endif
	}
#endif
}

template <class T>
void
Str<T>::AppendNativeLineEnding()
{
#if defined( OS_CRLF )
	Append( '\r' );
	Append( '\n' );
#elif defined( OS_LF )
	Append( '\n' );
#elif defined( OS_CR )
	Append( '\r' );
#endif
}

/*
 * StrA and StrW classes.
 */

class StrA;
class StrW;

#ifdef UNICODE
typedef StrW		StrT;
#else
typedef StrA		StrT;
#endif

class StrA : public Str<char>
{
    public:
	void		SetA( char const* psz, size_t cch = -1 ) { Set( psz, cch ); }
	void		SetT( TCHAR const* psz, size_t cch = -1 );
#ifdef UNICODE
	void		SetW( WCHAR const* psz, size_t cch = -1 );
#endif

	void		SetA( StrA const& s ) { SetA( s.Text(), s.Length() ); }
	void		SetT( StrT const& s );
#ifdef UNICODE
	void		SetW( StrW const& s );
#endif
};

#ifdef UNICODE
class StrW : public Str<WCHAR>
{
    public:
	void		SetA( char const* psz, size_t cch = -1 );
	void		SetT( TCHAR const* psz, size_t cch = -1 );
	void		SetW( WCHAR const* psz, size_t cch = -1 ) { Set( psz, cch ); }

	void		SetA( StrA const& s );
	void		SetT( StrT const& s );
	void		SetW( StrW const& s ) { SetW( s.Text(), s.Length() ); }
};
#endif

#ifdef UNICODE
inline void
StrA::SetW( StrW const& s )
{
	SetW( s.Text(), s.Length() );
}

inline void
StrW::SetA( StrA const& s )
{
	SetA( s.Text(), s.Length() );
}
#endif

#ifdef UNICODE
inline void StrA::SetT( TCHAR const* psz, size_t cch ) { SetW( psz, cch ); }
inline void StrA::SetT( StrT const& s ) { SetW( s ); }
inline void StrW::SetT( TCHAR const* psz, size_t cch ) { SetW( psz, cch ); }
inline void StrW::SetT( StrT const& s ) { SetW( s ); }
#else
inline void StrA::SetT( TCHAR const* psz, size_t cch ) { SetA( psz, cch ); }
inline void StrA::SetT( StrT const& s ) { SetA( s ); }
#if 0
inline void StrW::SetT( TCHAR const* psz, size_t cch ) { SetA( psz, cch ); }
inline void StrW::SetT( StrT const& s ) { SetA( s ); }
#endif
#endif

