/*
 * Sys.cpp - System abstractions.
 */

// vim: set noet:			// Hard tabs.
// vim: set ts=8:			// Tab width 8.
// vim: set sw=4:			// Shift (indent) width 4.
// vim: set cino={0s:			// Braces flush with outer statement.
// vim: set cino=^1s:			// First indent 8 (add 1 shift width).

#include "precomp.h"

#if defined( OS_CRT )
#include <io.h>
#include <errno.h>
#include <sys\stat.h>
#endif

#include "std.h"
#include "str.h"
#include "sys.h"

/*
 * SysError - Simple general error codes.
 */

#if defined( OS_WINDOWS )

static SysError
SysErrorFromWinError( DWORD const dwErr )
{
	switch( dwErr )
	{
	case ERROR_SUCCESS:
	    return SE_OK;

	case ERROR_NOT_ENOUGH_MEMORY:
	case ERROR_OUTOFMEMORY:
	    return SE_OUTOFMEMORY;

	case ERROR_FILE_NOT_FOUND:
	case ERROR_PATH_NOT_FOUND:
	    return SE_FILENOTFOUND;

	case ERROR_ACCESS_DENIED:
	case ERROR_WRITE_PROTECT:
	    return SE_ACCESSDENIED;

	case ERROR_FILE_EXISTS:
	    return SE_FILEEXISTS;

	case ERROR_HANDLE_DISK_FULL:
	case ERROR_DISK_FULL:
	    return SE_DISKFULL;

	case ERROR_SHARING_VIOLATION:
	case ERROR_LOCK_VIOLATION:
	    return SE_FILEINUSE;

	default:
	    // Windows Mobile emulator can claim to not support deleting files
	    // on storage cards (which are really folders in the host
	    // computer's file system), and can get confused trying to delete
	    // a file that does not exist.
	    Assert1( g_fEmulator && ( dwErr == ERROR_CALL_NOT_IMPLEMENTED || dwErr == ERROR_INVALID_PARAMETER ), "Error %d is unknown.", dwErr );
	    return SE_UNKNOWN;
	}
}

inline SysError
GetLastSysError()
{
	return SysErrorFromWinError( GetLastError() );
}

#elif defined( OS_CRT )

static SysError
SysErrorFromErrno( errno_t const err )
{
	switch( err )
	{
	case 0:
	    return SE_OK;

	case ENOMEM:
	    return SE_OUTOFMEMORY;

	case ENOENT:
	case ENODEV:
	case ENOTDIR:
	case EISDIR:
	    return SE_FILENOTFOUND;

	case EPERM:
	case EACCES:
	case EROFS:
	    return SE_ACCESSDENIED;

	case EEXIST:
	    return SE_FILEEXISTS;

	case ENOSPC:
	    return SE_DISKFULL;

	    //$ note: (chrisant) No errno value corresponds to SE_FILEINUSE.

	default:
	    Assert1( false, "Error %d is unknown.", err );
	    return SE_UNKNOWN;
	}
}

inline SysError
GetLastSysError()
{
	return SysErrorFromErrno( errno );
}

#else
#error GetLastSysError not implemented for this platform.
#endif

/*
 * Filename helpers.
 */

bool
FileExists( TCHAR const* pszFile )
{
	Assert( pszFile );

#if defined( OS_WINDOWS )
	DWORD const dw = GetFileAttributes( pszFile );
	if( dw == DWORD( -1 ) )
	    return false;
	if( dw & FILE_ATTRIBUTE_DIRECTORY )
	    return false;
#elif defined( OS_CRT )
	struct _stat info;
	if( _tstat( pszFile, &info ) )
	    return false;
	if( info.st_mode & _S_IFDIR )
	    return false;
#else
#error FileExists not implemented for this platform.
#endif

	return true;
}

TCHAR const*
GetFileName( TCHAR const* pszFile )
{
	// GetFileName is defined as returning a pointer to the filename
	// component of the argument WITHIN the argument.  Callers rely on
	// this and perform pointer arithmetic.  An implementation of this
	// must not return a pointer to any other buffer.

#if defined( USE_DOS_PATH_PARSING )
	LPCTSTR pszName;

	for( pszName = pszFile; *pszFile; pszFile++ )
	    if( '/' == *pszFile || '\\' == *pszFile )
		pszName = pszFile + 1;

	return *pszName ? pszName : 0;
#else
#error GetFileName not implemented for this platform.
#endif
}

TCHAR const*
GetExtension( TCHAR const* pszFile )
{
	LPCTSTR pszDot;

	for( pszDot = 0; *pszFile; pszFile++ )
	    switch( *pszFile )
	    {
	    case '.':
		pszDot = pszFile;
		break;
	    case ' ':
	    case ':':
	    case '\\':
	    case '/':
		pszDot = 0;
		break;
	    }

	return pszDot;
}

TCHAR const*
StripFilename( TCHAR const* pszPath, StrT& sPathOnly, StrT* psFileOnly )
{
	sPathOnly.Set( pszPath );

	TCHAR const* const pszName = GetFileName( sPathOnly.Text() );

	if( pszName )
	{
	    if( psFileOnly )
		psFileOnly->Set( pszName );

	    sPathOnly.SetLength( pszName - sPathOnly.Text() );
	}
	else
	{
	    if( psFileOnly )
		psFileOnly->Set( sPathOnly );

	    sPathOnly.Clear();
	}

	if( sPathOnly.Length() && IsDirectorySeparator( sPathOnly.Text()[sPathOnly.Length() - 1] ) )
	    sPathOnly.SetLength( sPathOnly.Length() - 1 );

	return sPathOnly.Length() ? sPathOnly.Text() : 0;
}

SysError
__DeleteFile( TCHAR const* pszFile )
{
#if defined( OS_WINDOWS )
	bool const fSuccess = !!DeleteFile( pszFile );
#elif defined( OS_CRT )
	bool const fSuccess = !_tunlink( pszFile );
#else
#error __DeleteFile not implemented for this platform.
#endif

	if( !fSuccess )
	{
#ifdef OS_WINDOWS
	    DWORD const dwErr = GetLastError();
	    SysError e = SysErrorFromWinError( dwErr );
# ifdef OS_WINDOWSMOBILE
	    if( g_fEmulator && e == SE_UNKNOWN && !FileExists( pszFile ) )
		e = SE_FILENOTFOUND;
# endif
	    Assert3( e == SE_OK || e == SE_FILENOTFOUND, "Delete failed (SysError %d, WinError %d).\nFile '%s'.", e, dwErr, pszFile );
#else
	    SysError const e = GetLastSysError();
	    Assert2( e == SE_OK || e == SE_FILENOTFOUND, "Delete failed (SysError %d).\nFile '%s'.", e, pszFile );
#endif
	    return e;
	}

	return SE_OK;
}

SysError
__RenameFile( TCHAR const* pszFrom, TCHAR const* pszTo )
{
#if defined( OS_WINDOWS )
	bool const fSuccess = !!MoveFile( pszFrom, pszTo );
#elif defined( OS_CRT )
	bool const fSuccess = !_trename( pszFrom, pszTo );
#else
#error __RenameFile not implemented for this platform.
#endif

	if( !fSuccess )
	{
	    SysError const e = GetLastSysError();
#ifdef OS_WINDOWS
	    Assert4( false, "Rename failed (SysError %d, WinError %d).\nFrom '%s' to '%s'.", e, GetLastError(), pszFrom, pszTo );
#else
	    Assert3( false, "Rename failed (SysError %d).\nFrom '%s' to '%s'.", e, pszFrom, pszTo );
#endif
	    return e;
	}

	return SE_OK;
}

/*
 * FileIO - File IO abstraction.
 */

#if defined( OS_CRT )

static unsigned
__fsize( FILE* const pfile )
{
	// Remember current position.

	size_t const posCur = ftell( pfile );
	if( posCur == 0xffffffff )
	    return 0xffffffff;

	// Go to end.

	if( fseek( pfile, 0, SEEK_END ) )
	    return 0xffffffff;

	// Get end position (file size).

	size_t const posRet = ftell( pfile );
	if( posRet == 0xffffffff )
	    return 0xffffffff;

	// Restore current position.

	if( fseek( pfile, posCur, SEEK_SET ) )
	    return 0xffffffff;

	return posRet;
}

#endif

FileIO*
FileIO::Create()
{
	return new FileIO();
}

FileIO::~FileIO()
{
	Close();
}

void
FileIO::Set( TCHAR const* const pszFile )
{
	Assert( !IsSet() );

	m_sName.Set( pszFile );
}

bool
FileIO::Open( FileMode const mode, SysError& e )
{
	Assert( IsSet() );
	Assert( !IsOpen() );

#if defined( OS_WINDOWS )
	static DWORD const c_rgdwAccess[] = { GENERIC_READ, GENERIC_READ|GENERIC_WRITE, GENERIC_WRITE };
	static DWORD const c_rgdwShare[] = { FILE_SHARE_READ, 0, 0 };
	static DWORD const c_rgdwCreate[] = { OPEN_EXISTING, OPEN_EXISTING, CREATE_ALWAYS };
	CASSERT( DimensionOf( c_rgdwAccess ) == FM_MAX );
	CASSERT( DimensionOf( c_rgdwShare ) == FM_MAX );
	CASSERT( DimensionOf( c_rgdwCreate ) == FM_MAX );
	HANDLE const hFile = CreateFile( Name(), c_rgdwAccess[mode], c_rgdwShare[mode], 0, c_rgdwCreate[mode], 0, 0 );
	if( hFile == INVALID_HANDLE_VALUE )
	{
	    e = GetLastSysError();
	    return false;
	}
	m_shFile = hFile;
#elif defined( OS_CRT )
	static TCHAR const* const c_rgszMode[] = { TEXT( "rb" ), TEXT( "r+b" ), TEXT( "wb" ) };
	CASSERT( DimensionOf( c_rgszMode ) == FM_MAX );
	FILE* pFile = 0;
	if( _tfopen_s( &pFile, Name(), c_rgszMode[mode] ) || !pFile )
	{
	    e = GetLastSysError();
	    return false;
	}
	m_shFile = pFile;
#else
#error FileIO::Open not implemented for this platform.
#endif

	m_mode = mode;
	return true;
}

bool
FileIO::Seek( unsigned const offset, SysError& e )
{
	Assert( IsSet() );
	Assert( IsOpen() );

#if defined( OS_WINDOWS )
	bool const fSuccess = ( 0xffffffff != SetFilePointer( m_shFile, offset, 0, FILE_BEGIN ) );
#elif defined( OS_CRT )
	bool const fSuccess = !fseek( m_shFile, offset, SEEK_SET );
#else
#error FileIO::Seek not implemented for this platform.
#endif

	if( !fSuccess )
	{
	    e = GetLastSysError();
	    return false;
	}

	return true;
}

bool
FileIO::Read( void* const pvBuffer, unsigned const cbToRead, SysError& e, unsigned* const pcbDidRead )
{
	Assert( IsSet() );
	Assert( IsOpen() );
	Assert( m_mode == FM_READWRITE || m_mode == FM_READ );

#if defined( OS_WINDOWS )
	DWORD dwDidRead;
	bool const fSuccess = !!ReadFile( m_shFile, pvBuffer, cbToRead, &dwDidRead, 0 );
	if( pcbDidRead )
	    *pcbDidRead = dwDidRead;
#elif defined( OS_CRT )
	unsigned const cbDidRead = fread( pvBuffer, 1, cbToRead, m_shFile );
	bool const fSuccess = ( cbDidRead == cbToRead || ( pcbDidRead && feof( m_shFile ) ) );
#else
#error FileIO::Read not implemented for this platform.
#endif

	if( !fSuccess )
	{
	    e = GetLastSysError();
	    return false;
	}

	return true;
}

bool
FileIO::Write( void const* const pvBuffer, size_t const cbToWrite, SysError& e )
{
	Assert( IsSet() );
	Assert( IsOpen() );
	Assert( m_mode == FM_READWRITE || m_mode == FM_WRITE );

#if defined( OS_WINDOWS )
	DWORD dwDidWrite;
	bool const fSuccess = ( WriteFile( m_shFile, pvBuffer, cbToWrite, &dwDidWrite, 0 ) && cbToWrite == dwDidWrite );
#elif defined( OS_CRT )
	size_t const cbDidWrite = fwrite( pvBuffer, 1, cbToWrite, m_shFile );
	bool const fSuccess = ( cbDidWrite == cbToWrite );
#else
#error FileIO::Write not implemented for this platform.
#endif

	if( !fSuccess )
	{
	    e = GetLastSysError();
	    return false;
	}

	return true;
}

unsigned
FileIO::Offset( SysError& e )
{
	Assert( IsSet() );
	Assert( IsOpen() );

#if defined( OS_WINDOWS )
	unsigned const offset = SetFilePointer( m_shFile, 0, 0, FILE_CURRENT );
#elif defined( OS_CRT )
	unsigned const offset = ftell( m_shFile );
#else
#error FileIO::Offset not implemented for this platform.
#endif

	if( offset == 0xffffffff )
	    e = GetLastSysError();

	return offset;
}

unsigned
FileIO::Size( SysError& e )
{
	Assert( IsSet() );
	Assert( IsOpen() );

#if defined( OS_WINDOWS )
	unsigned const offset = GetFileSize( m_shFile, 0 );
#elif defined( OS_CRT )
	unsigned const offset = __fsize( m_shFile );
#else
#error FileIO::Offset not implemented for this platform.
#endif

	if( offset == 0xffffffff )
	    e = GetLastSysError();

	return offset;
}

bool
FileIO::Truncate( SysError& e )
{
	Assert( IsSet() );
	Assert( IsOpen() );
	Assert( m_mode == FM_READWRITE || m_mode == FM_WRITE );

#if defined( OS_WINDOWS )
	bool const fSuccess = !!SetEndOfFile( m_shFile );
#elif defined( OS_CRT )
	unsigned const offset = Offset( e );
	if( e )
	    return false;
	int const fd = _fileno( m_shFile );
	bool const fSuccess = ( fd != -1 && !_chsize( fd, offset ) );
#else
#error FileIO::Truncate not implemented for this platform.
#endif

	if( !fSuccess )
	{
	    e = GetLastSysError();
	    return false;
	}

	return true;
}

void
FileIO::Close()
{
	if( IsOpen() )
	    m_shFile.Free();
}

bool
FileIO::Delete( SysError& e )
{
	Assert( IsSet() );
	Assert( !IsOpen() );

	e = __DeleteFile( Name() );

	return !e;
}

bool
FileIO::Rename( TCHAR const* pszNewName, SysError& e )
{
	Assert( IsSet() );
	Assert( !IsOpen() );

	e = __RenameFile( Name(), pszNewName );

	return !e;
}

